--[[=============================================================================
    Basic Template for PROJECTOR Driver

    Copyright 2015 Control4 Corporation. All Rights Reserved.
===============================================================================]]------------
-------------------------------------------------------------
--===========================================================
----------------------- TEMPLATE INCLUDES ---------------------------
--[[=============================================================================
    Driver Declarations used to call startup routines, teardown routines, and 
    other basic functions of the drivers operation

    Copyright 2016 Control4 Corporation. All Rights Reserved.
===============================================================================]]

-- Template Version Table
TEMPLATE_VERSION = {}
TEMPLATE_VERSION.c4_driver_declarations = "2016.01.08"

-- Command Handler Tables
EX_CMD = {}
PRX_CMD = {}
UI_REQ = {}
NOTIFY = {}
DEV_MSG = {}
LUA_ACTION = {}

--[[=============================================================================
    Tables of functions
    The following tables are function containers that are called within the
    following functions:

    OnDriverInit()
        First calls all functions contained within ON_DRIVER_EARLY_INIT table
        then calls all functions contained within ON_DRIVER_INIT table

    OnDriverLateInit()
        Calls all functions contained within ON_DRIVER_LATEINIT table

    OnDriverDestroyed()
        Calls all functions contained within ON_DRIVER_DESTROYED table

    OnPropertyChanged()
        Calls all functions contained within ON_PROPERTY_CHANGED table
===============================================================================]]
ON_DRIVER_INIT = {}
ON_DRIVER_EARLY_INIT = {}
ON_DRIVER_LATEINIT = {}
ON_DRIVER_DESTROYED = {}
ON_PROPERTY_CHANGED = {}

-- Constants
DEFAULT_PROXY_BINDINGID = 5001
--[[=============================================================================
    c4_object Class

    Copyright 2016 Control4 Corporation. All Rights Reserved.
===============================================================================]]

-- Set template version for this file
if (TEMPLATE_VERSION ~= nil) then
	TEMPLATE_VERSION.c4_object = "2016.01.08"
end

function inheritsFrom( baseClass )
	local new_class = {}
	local class_mt = { __index = new_class }

	function new_class:create(...)
		local newinst = {}

		setmetatable( newinst, class_mt )

		-- Call the constructor when we create this class
		if newinst.construct then
			-- Allow returning a different obj than self. This allows for readonly tables etc...
			newinst = newinst:construct(...) or newinst
		end

		return newinst
	end

	if nil ~= baseClass then
		setmetatable( new_class, { __index = baseClass } )
	end

	--[[=============================================================================
	    Implementation of additional OO properties starts here 
	===============================================================================]]
	
	-- Return the class object of the instance
	function new_class:class()
		return new_class
	end

	--[[=============================================================================
	    Return the super class object of the instance. 

	    Note Calling methods on the base class itself will modify
	    the base table's static properties. In order to have call
	    the base class' methods and have them modify the current object
	    use super() or superAsSelf(). 
	===============================================================================]]
	function new_class:superClass()
		return baseClass
	end

	--[[=============================================================================
	    Returns a table that allows calling of the base class's method
	    while maintaining the objects state as the modified state of the base
	    class' methods. For example consider the following statements (order matters):

	    -- The child sees the parents property if the child hasn't overriden the property
	       obj:superClass().id = "parent"
	       obj.id == "parent" -- true

	    -- Setting the property on the child overrides (hides) the parents property
	       obj.id = "child"
	       obj.id == "child" -- true

	    -- The super() method pass
	       obj:super().id == "parent" -- true
	       obj:super().id = "child"
	       obj:super().id == "parent" -- still true
	       obj.id == "child" -- still true
	===============================================================================]]
	function new_class:super()
		local holder = {}

		holder.child = self
		holder.parent = baseClass

		local mt = {}
		mt.__index = function(table, index)
			if table.parent[index] then
				return table.parent[index]
			else 
				return table.child[index]
			end
		end

		-- Only set the new values to the child.
		mt.__newindex = function(table, key, value)
			table.child[key] = value
		end

		mt.__tostring = function(table)
			return tostring(table.child)
		end

		setmetatable(holder, mt)
		return holder
	end

	new_class.new = new_class.create

	--[[=============================================================================
	    Return true if the caller is an instance of theClass
	===============================================================================]]
	function new_class:isa( theClass )
		local b_isa = false
		local cur_class = new_class

		while ( nil ~= cur_class ) and ( false == b_isa ) do
			if cur_class == theClass then
				b_isa = true
			else
				cur_class = cur_class:superClass()
			end
		end

		return b_isa
	end

	return new_class
end

--[[=============================================================================
    Inheritance unit tests
===============================================================================]]
function __test_inheritance()
	local b = inheritsFrom(nil)

	b.construct = function(self, msg)
		self._msg = msg
	end

	local t = inheritsFrom(b)
	t.construct = function(self, msg)
		self:super():construct(msg)
	end

	t1 = t:new("t1")
	t2 = t:new("t2")
	assert(t1._msg == "t1", "t1 message is not equal to 't1' it''s: " .. t1._msg)
	assert(t2._msg == "t2", "t2 message is not equal to 't2' it''s: " .. t2._msg)    
	assert(tostring(t1:super()) ==  tostring(t1), "tostrings don't match");
	assert(t1:superClass() == b, "superClass and baseClass should be the same. They are not.")

	t1:superClass().id = "parent"
	assert(t1.id == "parent", "obect''s super class has invalid property value: ", t1.id)

	-- Setting the property on the child overrides (hides) the parents property
	t1.id = "child"
	assert(t1.id == "child", "object instance variable has invalid property value: " .. t1.id)

	-- The super() method maintains the self pointer to the child and not to the base
	assert(t1:super().id == "parent", "superAsSelf found invalid value for base class variable")
	t1:super().id = "child1"
	assert(t1:super().id == "parent", "Setting of instance variable hid base classes variable from itself");
	assert(t1.id == "child1", "Settings of instance variable did not change child instance variable")
end
--[[=============================================================================
    c4_log Class

    Copyright 2016 Control4 Corporation. All Rights Reserved.
===============================================================================]]

-- Set template version for this file
if (TEMPLATE_VERSION ~= nil) then
	TEMPLATE_VERSION.c4_log = "2016.01.08"
end

c4_log = inheritsFrom(nil)

function c4_log:construct(logName)
	self._logLevel = tonumber(string.sub(Properties['Log Level'] or "", 1, 1)) or 5
	self._outputPrint = Properties['Log Mode']:find("Print") ~= nil
	self._outputC4Log = Properties['Log Mode']:find("Log") ~= nil
	self._logName = logName or ""

	-- make sure Property is up to date (no harm if absent)
	C4:UpdateProperty("Log Level", Properties['Log Level'])
end

function c4_log:SetLogLevel(level)
	self._logLevel = tonumber(string.sub(level or "", 1, 1)) or self._logLevel
end

function c4_log:LogLevel()
	return self._logLevel
end

function c4_log:OutputPrint(value)
	self._outputPrint = value
end

function c4_log:OutputC4Log(value)
	self._outputC4Log = value
end

function c4_log:SetLogName(logName)

	if (logName == nil or logName == "") then
		logName = ""
	else
		logName = logName .. ": "
	end

	self._logName = logName
end

function c4_log:LogName()
	return self._logName
end

function c4_log:Enabled()
	return (self._outputPrint or self._outputC4Log)
end

function c4_log:PrintEnabled()
	return self._outputPrint
end

function c4_log:C4LogEnabled()
	return self._outputC4Log
end

function c4_log:CreateTableText(tValue, tableText)
	tableText = tableText or ""

	if (type(tValue) == "table") then

		tableText = tableText .. "{"
		for k, v in pairs(tValue) do

			-- add key
			if (type(k) == "number") then
				tableText = tableText .. "[" .. tostring(k) .. "]="
			elseif (type(k) == "string") then
				tableText = tableText .. k .. "="
			else
				print (tostring(k) .. ": " .. tostring (v))
			end

			-- add value
			if (type(v) == "number") then
				tableText = tableText .. tostring(v) .. ","
			elseif (type(v) == "string") then
				tableText = tableText .. "'" .. v .. "',"
			elseif (type(v) == "table") then
				tableText = c4_log:CreateTableText(v, tableText)
				tableText = tableText .. ","
			elseif (type(v) == "boolean") then
				tableText = tableText .. tostring(v) .. ","
			end
		end

		tableText = tableText .. "}"
	end

	return tableText
end

function InsertIndent(indentLevel)
	local indentStr = ""

	for i=1, indentLevel do
		indentStr = indentStr .. "\t"
	end

	return indentStr
end

function c4_log:CreateTableTextFormatted(tValue, tableText, indentLevel)
	tableText = tableText or ""
	indentLevel = indentLevel or 0

	if (type(tValue) == "table") then

		indentLevel = indentLevel + 1
		tableText = tableText .. "{\n"
		for k, v in pairs(tValue) do

			-- add key
			if (type(k) == "number") then
				tableText = tableText .. InsertIndent(indentLevel) .. "[" .. tostring(k) .. "]="
			elseif (type(k) == "string") then
				tableText = tableText .. InsertIndent(indentLevel) .. k .. "="
			else
				print (tostring(k) .. ": " .. tostring (v))
			end

			-- add value
			if (type(v) == "number") then
				tableText = tableText .. tostring(v) .. ",\n"
			elseif (type(v) == "string") then
				tableText = tableText .. "'" .. v .. "',\n"
			elseif (type(v) == "table") then
				tableText = c4_log:CreateTableTextFormatted(v, tableText, indentLevel)
				tableText = tableText .. ",\n"
			elseif (type(v) == "boolean") then
				tableText = tableText .. tostring(v) .. ",\n"
			end
		end

		indentLevel = indentLevel - 1
		tableText = tableText .. InsertIndent(indentLevel) .. "}"
	end

	return tableText
end

MAX_TABLE_LEVELS = 10
function c4_log:PrintTable(tValue, tableText, sIndent, level)
	tableText = tableText or ""
	level = level + 1
	
	if (level <= MAX_TABLE_LEVELS) then
		if (type(tValue) == "table") then
			for k,v in pairs(tValue) do
				if (tableText == "") then
					tableText = sIndent .. tostring(k) .. ":  " .. tostring(v)
					if (sIndent == ".   ") then sIndent = "    " end
				else
					tableText = tableText .. "\n" .. sIndent .. tostring(k) .. ":  " .. tostring(v)
				end
				if (type(v) == "table") then
					tableText = self:PrintTable(v, tableText, sIndent .. "   ", level)
				end
			end
		else
			tableText = tableText .. "\n" .. sIndent .. tostring(tValue)
		end
	end
	
	return tableText
end

function c4_log:LogTable(tValue, sIndent, level)
	level = level + 1
	
	if (level <= MAX_TABLE_LEVELS) then
		if (type(tValue) == "table") then
			for k,v in pairs(tValue) do
				C4:ErrorLog(self._logName .. sIndent .. tostring(k) .. ":  " .. tostring(v))
				if (type(v) == "table") then
					self:LogTable(v, sIndent .. "   ", level)
				end
			end
		else
			C4:ErrorLog(self._logName .. sIndent .. tValue)
		end
	end
end

function c4_log:Print(logLevel, sLogText)

	if (self._logLevel >= logLevel) then
		if (type(sLogText) == "table") then
			if (self._outputPrint) then
				print (self:PrintTable(sLogText, tableText, ".   ", 0))
			end

			if (self._outputC4Log) then
				self:LogTable(sLogText, "   ", 0)
			end

			return
		end

		if (self._outputPrint) then
			print (sLogText)
		end

		if (self._outputC4Log) then
			C4:ErrorLog(self._logName .. tostring(sLogText))
		end
	end
end

function c4_log:Fatal(sLogText, ...)
	self:LogOutput(0, sLogText, ...)
end

function c4_log:Error(sLogText, ...)
	self:LogOutput(1, sLogText, ...)
end

function c4_log:Warn(sLogText, ...)
	self:LogOutput(2, sLogText, ...)
end

function c4_log:Info(sLogText, ...)
	self:LogOutput(3, sLogText, ...)
end

function c4_log:Debug(sLogText, ...)
	self:LogOutput(4, sLogText, ...)
end

function c4_log:Trace(sLogText, ...)
	self:LogOutput(5, sLogText, ...)
end

function c4_log:LogOutput(level, sLogText, ...)
	if (LogEnabled()) then
		if (type(sLogText) == "string") then
			sLogText = string.format(sLogText, ...)
		end

		self:Print(level, sLogText)
	end
end

--[[=============================================================================
    c4_log wrapper functions
===============================================================================]]
function TryLog(level, sLogText, ...)
	LOG:LogOutput(level, sLogText, ...)
end

--[[=============================================================================
    SetLogLevel(level)

    Description: 
    Sets the desired log level to view

    Parameters:
    level(int) - The logging level to set the message to
                 0 = Fatal
                 1 = Error
                 2 = Warn
                 3 = Info
                 4 = Debug
                 5 = Trace

    Returns:
    None
===============================================================================]]
function SetLogLevel(level)
	LOG:SetLogLevel(level)
end

--[[=============================================================================
    LogLevel()

    Description: 
    Returns the currently set log level

    Parameters:
    None

    Returns:
    The current log level
        0 = Fatal
        1 = Error
        2 = Warn
        3 = Info
        4 = Debug
        5 = Trace
===============================================================================]]
function LogLevel()
	return LOG:LogLevel()
end

--[[=============================================================================
    OutputPrint(value)

    Description: 
    Specifies whether to output log messages or not

    Parameters:
    value(bool) - true to enable logging output, false otherwise

    Returns:
    None
===============================================================================]]
function OutputPrint(value)
	LOG:OutputPrint(value)
end

--[[=============================================================================
    OutputPrint(value)

    Description: 
    Specifies whether to output log messages to file or not

    Parameters:
    value(bool) - true to enable logging output, false otherwise

    Returns:
    None
===============================================================================]]
function OutputC4Log(value)
	LOG:OutputC4Log(value)
end

--[[=============================================================================
    SetLogName(logName)

    Description: 
    Sets the name of the log file where the messages will be written to

    Parameters:
    logName(string) - Sets the name of the log to write messages to

    Returns:
    None
===============================================================================]]
function SetLogName(logName)
	LOG:SetLogName(logName)
end

--[[=============================================================================
    LogName(logName)

    Description: 
    Gets the name of the log file where the messages will be written to

    Parameters:
    None

    Returns:
    The value of the log file that has been set
===============================================================================]]
function LogName()
	return LOG:LogName()
end

--[[=============================================================================
    LogEnabled()

    Description: 
    Identifies if logging or print has been enabled

    Parameters:
    None

    Returns:
    true if either logging or print has been enabled, false otherwise
===============================================================================]]
function LogEnabled()
	return LOG:Enabled()
end

--[[=============================================================================
    PrintEnabled()

    Description: 
    Gets the state of print output

    Parameters:
    None

    Returns:
    true if print has been enabled, false otherwise
===============================================================================]]
function PrintEnabled()
	return LOG:PrintEnabled()
end

--[[=============================================================================
    C4LogEnabled()

    Description: 
    Gets the state of logging

    Parameters:
    None

    Returns:
    true if logging has been enabled, false otherwise
===============================================================================]]
function C4LogEnabled()
	return LOG:C4LogEnabled()
end

--[[=============================================================================
    LogFatal(sLogText, ...)

    Description: 
    Formats and prints a series of characters and values to the enabled outputs
    when the set logging level is Fatal(0) or higher

    Parameters:
    sLogText(string) - Format control string
    ...              - Optional arguments which will replace all the format
                       specifiers contained in the format string

    Returns:
    None
===============================================================================]]
function LogFatal(sLogText, ...)
	local status, err = pcall(TryLog, 0, sLogText, ...)
	if (not status) then
		LOG:Print(1, "LUA_ERROR - LogFatal failed: " .. err)
	end
end

--[[=============================================================================
    LogError(sLogText, ...)

    Description: 
    Formats and prints a series of characters and values to the enabled outputs
    when the set logging level is Error(1) or higher

    Parameters:
    sLogText(string) - Format control string
    ...              - Optional arguments which will replace all the format
                       specifiers contained in the format string

    Returns:
    None
===============================================================================]]
function LogError(sLogText, ...)
	local status, err = pcall(TryLog, 1, sLogText, ...)
	if (not status) then
		LOG:Print(1, "LUA_ERROR - LogError failed: " .. err)
	end
end

--[[=============================================================================
    LogWarn(sLogText, ...)

    Description: 
    Formats and prints a series of characters and values to the enabled outputs
    when the set logging level is Warn(2) or higher

    Parameters:
    sLogText(string) - Format control string
    ...              - Optional arguments which will replace all the format
                       specifiers contained in the format string

    Returns:
    None
===============================================================================]]
function LogWarn(sLogText, ...)
	local status, err = pcall(TryLog, 2, sLogText, ...)
	if (not status) then
		LOG:Print(1, "LUA_ERROR - LogWarn failed: " .. err)
	end
end

--[[=============================================================================
    LogInfo(sLogText, ...)

    Description: 
    Formats and prints a series of characters and values to the enabled outputs
    when the set logging level is Info(3) or higher

    Parameters:
    sLogText(string) - Format control string
    ...              - Optional arguments which will replace all the format
                       specifiers contained in the format string

    Returns:
    None
===============================================================================]]
function LogInfo(sLogText, ...)
	local status, err = pcall(TryLog, 3, sLogText, ...)
	if (not status) then
		LOG:Print(1, "LUA_ERROR - LogInfo failed: " .. err)
	end
end

--[[=============================================================================
    LogDebug(sLogText, ...)

    Description: 
    Formats and prints a series of characters and values to the enabled outputs
    when the set logging level is Debug(4) or higher

    Parameters:
    sLogText(string) - Format control string
    ...              - Optional arguments which will replace all the format
                       specifiers contained in the format string

    Returns:
    None
===============================================================================]]
function LogDebug(sLogText, ...)
	local status, err = pcall(TryLog, 4, sLogText, ...)
	if (not status) then
		LOG:Print(1, "LUA_ERROR - LogDebug failed: " .. err)
	end
end

--[[=============================================================================
    LogTrace(sLogText, ...)

    Description: 
    Formats and prints a series of characters and values to the enabled outputs
    when the set logging level is Trace(5) or higher

    Parameters:
    sLogText(string) - Format control string
    ...              - Optional arguments which will replace all the format
                       specifiers contained in the format string

    Returns:
    None
===============================================================================]]
function LogTrace(sLogText, ...)
	local status, err = pcall(TryLog, 5, sLogText, ...)
	if (not status) then
		LOG:Print(1, "LUA_ERROR - LogTrace failed: " .. err)
	end
end

function dbgPrint(buf)
	if (LOG:PrintEnabled()) then
		print (buf)
	end
end

function dbgHexdump(buf)
	hexdump(buf, dbgPrint)
end

--[[=============================================================================
    c4_log unit tests
===============================================================================]]
function __test_c4_log()
		
	local LOG = c4_log:new("test_c4_log")
	assert(LOG:LogName() == "test_c4_log", "_logName is not equal to 'test_c4_log' it is: " .. LOG:LogName())

	-- Test setting log level
	LOG:SetLogLevel("2 - Warning")
	assert(LOG:LogLevel() == 2, "_logLevel is not equal to '2' it is: " .. LOG:LogLevel())

	LOG:SetLogLevel(3)
	assert(LOG:LogLevel() == 3, "_logLevel is not equal to '3' it is: " .. LOG:LogLevel())

	-- Test enabling logs
	LOG:OutputPrint(false)
	assert(LOG:PrintEnabled() == false, "_outputPrint is not equal to 'false' it is: " .. tostring(LOG:PrintEnabled()))

	LOG:OutputC4Log(true)
	assert(LOG:C4LogEnabled() == true, "_outputC4Log is not equal to 'true' it is: " .. tostring(LOG:C4LogEnabled()))

	LOG:SetLogLevel(4)
	LogTrace("***** This is a test *****")
end

function __test_CreatTableText()
	local tTest = {}

	tTest[1] = {}
	tTest[2] = {}
	tTest[3] = 30
	tTest[4] = "Forty"
	
	LogTrace("----- tText -----")
	LogTrace(tTest)

	local tTest2 = { One = {},
					 Two = {},
					 Three = 30,
					 Four = "Forty" }
	LogTrace("----- tText2 -----")
	LogTrace(tTest2)
	
	local tTest3 = { [1] = {},
					 [2] = {},
					 [3] = 30,
					 [4] = "Forty" }
	LogTrace("----- tText3 -----")
	LogTrace(tTest3)

	local tTest4 = { [1] = {},
					 Two = {},
					 [3] = 30,
					 [4] = "Forty",
					 Five = "Fifty" }
	LogTrace("----- tText4 -----")
	LogTrace(tTest4)

	local tableText = LOG:CreateTableText(tTest4)
	LogTrace("----- tableText -----")
	LogTrace(tableText)
	
	--local tNew = {[1] = {},[3] = 30,[4] = 'Forty',Five = 'Fifty',Two = {},}
	--LogTrace(tNew)
end

function __TestCreateTableTextFormatted()
		
	local LOG = c4_log:new("test_c4_log")
	local tButtons = {
		Name = 'heat',
		Attributes = {},
		ChildNodes = {
			[1] = {
				Name = 'button',
				Attributes = {},
				ChildNodes = {
					[1] = {
						Value = '51',
						Attributes = {},
						Name = 'id',
						ChildNodes = {},
					},
					[2] = {
						Value = 'Pool Heater',
						Attributes = {},
						Name = 'button_text',
						ChildNodes = {},
					},
					[3] = {
						Value = 'POOLHT',
						Attributes = {},
						Name = 'button_name',
						ChildNodes = {},
					},
				},
			},
			[2] = {
				Name = 'button',
				Attributes = {},
				ChildNodes = {
					[1] = {
						Value = '53',
						Attributes = {},
						Name = 'id',
						ChildNodes = {},
					},
					[2] = {
						Value = 'Spa Heater',
						Attributes = {},
						Name = 'button_text',
						ChildNodes = {},
					},
					[3] = {
						Value = 'SPAHT',
						Attributes = {},
						Name = 'button_name',
						ChildNodes = {},
					},
				},
			},
			[3] = {
				Name = 'button',
				Attributes = {},
				ChildNodes = {
					[1] = {
						Value = '54',
						Attributes = {},
						Name = 'id',
						ChildNodes = {},
					},
					[2] = {Value = 'Pool Solar Heater',
						Attributes = {},
						Name = 'button_text',
						ChildNodes = {}
					},
					[3] = {
						Value = 'SOLHT',
						Attributes = {},
						Name = 'button_name',
						ChildNodes = {},
					},
				}
			}
		}
	}

	print(LOG:CreateTableTextFormatted(tButtons))
end
--[[=============================================================================
    c4_queue Class

    Copyright 2016 Control4 Corporation. All Rights Reserved.
===============================================================================]]

-- Set template version for this file
if (TEMPLATE_VERSION ~= nil) then
	TEMPLATE_VERSION.c4_queue = "2016.01.08"
end

c4_queue = inheritsFrom(nil)

function c4_queue:construct()
	-- entry table
	self._et = {first = 0, last = -1}
	self._maxSize = 0	-- no size limit
	self._name = ""

	local mt = getmetatable(self)
	if (mt ~= nil) then
		mt.__tostring = self.__tostring
	end
end

function c4_queue:__tostring()
	local tOutputString = {}
	table.insert(tOutputString, "--- Queue ---")
	if (not IsEmpty(self._name)) then
		table.insert(tOutputString, "  name = " .. tostring(self._name))
	end
	table.insert(tOutputString, "  first = " .. tostring(self._et.first))
	table.insert(tOutputString, "  last = " .. tostring(self._et.last))
	table.insert(tOutputString, "  number in queue = " .. tostring(self._et.last - self._et.first + 1))
	table.insert(tOutputString, "  maximum size = " .. self._maxSize)
	table.insert(tOutputString, "  next value = " .. tostring(self:value()))
	return table.concat(tOutputString, "\n")
end


-- push a value on the queue
function c4_queue:push(value, ...)
	local numItems = self._et.last - self._et.first + 1

	if ((self._maxSize <= 0) or (numItems < self._maxSize) ) then
		local last = self._et.last + 1
		self._et.last = last
		
		local interval = select(1, ...)
		local units = select(2, ...)
		local command_name = select(3, ...)
		self._et[last] = {["command"] = value, ["command_delay"] = interval, ["delay_units"] = units, ["command_name"] = command_name}		
		--LogTrace ("Queue:push(), first = " .. tostring(self._et.first) .. ", last = " .. tostring(self._et.last) .. ", number in queue = " .. tostring(self._et.last - self._et.first + 1) .. ", value = " .. value)
	else
		-- if addToQueue == true then push value to queue
		if (self:OnMaxSizeReached()) then
			local last = self._et.last + 1
			self._et.last = last
			self._et[last] = {["command"] = value, ["command_delay"] = interval, ["delay_units"] = units, ["command_name"] = command_name}
			--LogTrace ("Queue:push(), first = " .. tostring(self._et.first) .. ", last = " .. tostring(self._et.last) .. ", number in queue = " .. tostring(self._et.last - self._et.first + 1) .. ", value = " .. value)
		end
	end
end

function c4_queue:OnMaxSizeReached()
	--LogTrace ("Max Size Reached - clear queue and push value to the queue (default).")
	local addToQueue = true

	self:clear()
	return (addToQueue)
end

-- pop a value from the queue
function c4_queue:pop()
	local first = self._et.first

	if first > self._et.last then
		--LogTrace("Queue:pop(), queue is empty")
		return ""
	end

	local value = self._et[first]
	self._et[first] = nil        -- to allow garbage collection
	self._et.first = first + 1
	--LogTrace ("Queue:pop(), first = " .. tostring(self._et.first) .. ", last = " .. tostring(self._et.last) .. ", number in queue = " .. tostring(self._et.last - self._et.first + 1) .. ", value = " .. value)

	return value
end

-- clear queue
function c4_queue:clear()
	local first = self._et.first

	if first > self._et.last then
		--LogTrace ("Queue:clear(), queue is empty")
		return ""
	end

	self._et = {first = 0, last = -1}
	--LogTrace ("Queue:clear(), first = " .. tostring(self._et.first) .. ", last = " .. tostring(self._et.last) .. ", number in queue = " .. tostring(self._et.last - self._et.first + 1))
	--LogTrace (self._et)
end

-- return value of first item
function c4_queue:value()
	local first = self._et.first

	if (first > self._et.last) then
		return ""
	else
		return self._et[first]
	end
end

-- return queue's maximum size
function c4_queue:MaxSize()
	return self._maxSize
end

-- return queue's maximum size
function c4_queue:SetMaxSize(size)
	self._maxSize = size
end

function c4_queue:SetName(name)
	self._name = name
end

-- return the queue's current size
function c4_queue:Size()
	return self._et.last - self._et.first + 1
end

-- is queue empty?
function c4_queue:empty()
	-- print ("self._et.first = " .. tostring(self._et.first) .. ", self._et.last = " .. tostring(self._et.last))
	if (self._et.first > self._et.last) then
		return true
	else
		return false
	end
end

--[[
    c4_queue unit tests
--]]
function __test_c4_queue()
		
	local LOG = c4_log:new("test_c4_queue")
	LOG:SetLogLevel(5)
	LOG:OutputPrint(true)

	-- create an instance of the queue
	local c4Queue = c4_queue:new()

	c4Queue:SetMaxSize(3)
	assert(c4Queue:MaxSize() == 3, "_maxSize is not equal to '3' it is: " .. c4Queue:MaxSize())

	c4Queue:push("Item #1 in Queue")
	c4Queue:push("Item #2 in Queue")
	c4Queue:push("Item #3 in Queue")
	c4Queue:push("Item #4 in Queue")  -- this should cause OnMaxSizeReached() to be called and clear the queue
	assert(c4Queue:Size() == 1, "queue size is not equal to '1' it is: " .. c4Queue:Size())

	print (c4Queue)

	-- Test inheritance overriding OnMaxSizeReached
	-- Create a new class
	c4_queue_new = inheritsFrom(c4_queue)

	-- override construct()
	function c4_queue_new:construct()
		self.superClass():construct()  -- call base class
		self._maxSizeOption = 1

		local mt = getmetatable(self)
		if (mt ~= nil) then
			mt.__tostring = self.__tostring
		end
	end

	-- override OnMaxSizeReached()
	function c4_queue_new:OnMaxSizeReached()
		--Default: clear queue and push value to the queue. (No need to overload,

		-- Option 1: Do Nothing, new item is not added to queue
		if (self._maxSizeOption == 1) then
			
			LogInfo("Max Size Reached - do nothing, new item not added to queue (option 1)")
			return (false)
		-- Option 2: pop value, and push new value on queue
		elseif(self._maxSizeOption == 2) then
			LogInfo("Max Size Reached - pop value, and push new value on queue (option 2)")
			self:pop()
			return (true)
		-- Option 3: clear queue and DO NOT push new value onto queue
		elseif(self._maxSizeOption == 3) then
			LogInfo("Max Size Reached - clear queue and DO NOT push new value onto queue")
			self:clear()
			return (false)
		end
	end

	-- create an instance of the new queue
	local c4QueueNew = c4_queue_new:new()
	c4QueueNew:SetMaxSize(3)
	c4QueueNew:push("Item #1 in Queue")
	c4QueueNew:push("Item #2 in Queue")
	c4QueueNew:push("Item #3 in Queue")
	c4QueueNew:push("Item #4 in Queue")  -- this should cause OnMaxSizeReached() to be called and clear the queue
	assert(c4QueueNew:Size() == 3, "queue size is not equal to '3' it is: " .. c4QueueNew:Size())

	print(c4QueueNew)
	print ("done...")
end
--[[=============================================================================
    c4_timer Class

    Copyright 2016 Control4 Corporation. All Rights Reserved.
===============================================================================]]

-- Set template version for this file
if (TEMPLATE_VERSION ~= nil) then
	TEMPLATE_VERSION.c4_timer = "2016.01.08"
end

c4_timer = inheritsFrom(nil)

function c4_timer:construct(name, interval, units, Callback, repeating, CallbackParam)
	self._name = name
	self._timerID = TimerLibGetNextTimerID()
	self._interval = interval
	self._units = units
	self._repeating = repeating or false
	self._Callback = Callback
	self._CallbackParam = CallbackParam or ""
	self._id = 0
	gTimerLibTimers = gTimerLibTimers or {}
	for k,v in ipairs(gTimerLibTimers) do
	if(v._name == self._name and v._id ~= 0) then
			local timer = v
			timer:KillTimer()
			LogTrace("Killed old '"..timer._name.."' timer...")
		end
	end

	gTimerLibTimers[self._timerID] = self
	if (LOG ~= nil and type(LOG) == "table") then
		LogTrace("Created timer " .. self._name)
	end
end

function c4_timer:StartTimer(...)
	self:KillTimer()--timer lib fix

	-- optional parameters (interval, units, repeating)
	if ... then
		local interval = select(1, ...)
		local units = select(2, ...)
		local repeating = select(3, ...)

		self._interval = interval or self._interval
		self._units = units or self._units
		self._repeating = repeating or self._repeating
	end

	if (tonumber(self._interval) > 0) then
		if (LOG ~= nil and type(LOG) == "table") then
			LogTrace("Starting Timer: " .. self._name)
		end

		self._id = C4:AddTimer(self._interval, self._units, self._repeating)
	end

end

function c4_timer:KillTimer()
	if (self._id) then
		self._id = C4:KillTimer(self._id)


	end
end

function c4_timer:TimerStarted()
	return (self._id ~= 0)
end

function c4_timer:TimerStopped()
	return (self._id == 0)
end

function c4_timer:GetTimerInterval()
	return (self._interval)
end

function TimerLibGetNextTimerID()
	gTimerLibTimerCurID = gTimerLibTimerCurID + 1
	return gTimerLibTimerCurID
end

function ON_DRIVER_EARLY_INIT.c4_timer()
	gTimerLibTimers = {}
	gTimerLibTimerCurID = 0
end

function ON_DRIVER_DESTROYED.c4_timer()
	-- Kill open timers
	for k,v in pairs(gTimerLibTimers) do
		v:KillTimer()
	end
end

--[[=============================================================================
    OnTimerExpired(idTimer)

    Description:
    Function called by Director when the specified Control4 timer expires.

    Parameters:
    idTimer(string) - Timer ID of expired timer.
===============================================================================]]
function OnTimerExpired(idTimer)
	for k,v in pairs(gTimerLibTimers) do
		if (idTimer == v._id) then
			if (v._Callback) then
				v._Callback(v._CallbackParam)
			end
		end
	end
end

--[[=============================================================================
    CreateTimer(name, interval, units, callback, repeating, callbackParam)

    Description: 
    Creates a named timer with the given attributes

    Parameters:
    name(string)       - The name of the timer being created
    interval(int)      - The amount of the given time between calls to the
                         timers callback function
    units(string)      - The time of time interval used (e.g. SECONDS, MINUTES, ...)
    callback(string)   - The function to call when the timer expires
    repeating(bool)    - Parameter indicating whether the timer should be
                         called repeatedly until cancelled
    callbackParam(...) - Parameters to be passed to the callback function

    Returns:
    A handle to the timer
===============================================================================]]
function CreateTimer(name, interval, units, callback, repeating, callbackParam)
	timer = c4_timer:new(name, interval, units, callback, repeating, callbackParam)
	return timer
end

--[[=============================================================================
    StartTimer(handle, ...)

    Description: 
    Starts the timer created by calling the CreateTimer functions

    Parameters:
    handle(timer)   - Handle to a created timer object
    interval(int)   - The amount of the given time between calls to the
                      timers callback function
    units(string)   - The time of time interval used (e.g. SECONDS, MINUTES, ...)
    repeating(bool) - Parameter indicating whether the timer should be
                      called repeatedly until cancelled

    Returns:
    None
===============================================================================]]
function StartTimer(handle, ...)
	handle:StartTimer(...)
end

--[[=============================================================================
    KillTimer(handle)

    Description: 
    Starts the timer created by calling the CreateTimer functions

    Parameters:
    handle(timer) - Handle to a created timer object

    Returns:
    None
===============================================================================]]
function KillTimer(handle)
	handle:KillTimer()
end

--[[=============================================================================
    TimerStarted(handle)

    Description: 
    Identifies whether a timer has been started or not

    Parameters:
    handle(timer) - Handle to a created timer object

    Returns:
    Returns true if a the given timer handle has been started, or false otherwise
===============================================================================]]
function TimerStarted(handle)
	return handle:TimerStarted()
end

--[[=============================================================================
    TimerStopped(handle)

    Description: 
    Identifies whether a timer has been stopped or not

    Parameters:
    handle(timer) - Handle to a created timer object

    Returns:
    Returns true if a the given timer handle has been stopped, or false otherwise
===============================================================================]]
function TimerStopped(handle)
	return handle:TimerStopped()
end

--[[=============================================================================
    GetTimerInterval(handle)

    Description: 
    Gets the interval setting of the given timer

    Parameters:
    handle(timer) - Handle to a created timer object

    Returns:
    Returns the interval setting of the given timer
===============================================================================]]
function GetTimerInterval(handle)
	return handle:GetTimerInterval()
end

--[[=============================================================================
    c4_timer Unit Tests
===============================================================================]]
function __test_c4_timer()
			
	OnDriverInit()

	local LOG = c4_log:new("test_c4_timer")
	LOG:SetLogLevel(5)
	LOG:OutputPrint(true)

	function OnTestTimerExpired()
		c4Timer:KillTimer()
	end

	-- create an instance of the timer
	c4Timer = c4_timer:new("Test", 45, "MINUTES", OnTestTimerExpired)

	assert(c4Timer._id == 0, "_id is not equal to '0' it is: " .. c4Timer._id)
	c4Timer:StartTimer()
	assert(c4Timer._id == 10001, "_id is not equal to '10001' it is: " .. c4Timer._id)
	assert(c4Timer:TimerStarted() == true, "TimerStarted is not equal to true it is: " .. tostring(c4Timer:TimerStarted()))
	assert(c4Timer:TimerStopped() == false, "TimerStopped is not equal to false it is: " .. tostring(c4Timer:TimerStopped()))
	OnTimerExpired(c4Timer._id)
	assert(c4Timer:TimerStarted() == false, "TimerStarted is not equal to false it is: " .. tostring(c4Timer:TimerStarted()))
	assert(c4Timer:TimerStopped() == true, "TimerStopped is not equal to true it is: " .. tostring(c4Timer:TimerStopped()))
end
--[[=============================================================================
    Functions for parsing and managing xml

    Copyright 2016 Control4 Corporation. All Rights Reserved.
===============================================================================]]

if (TEMPLATE_VERSION ~= nil) then
	TEMPLATE_VERSION.c4_xml = "2016.01.08"
end

--[[=============================================================================
    GetParsedXmlNode(tXml, node)

    Description: 
    Find the specified node within the given table

    Parameters:
    tXml(table)  - Xml fragment containing the node we are looking for
    node(string) - The name of the node

    Returns:
    nil or the specified node within the table
===============================================================================]]
function GetParsedXmlNode(tXml, node)
	for k, v in pairs(tXml["ChildNodes"]) do
		if (v["Name"] == node) then
			return v["ChildNodes"]
		end
	end

	return nil
end

--[[=============================================================================
    GetParsedXmlValuesByKey(tXml, node, key, keyIsNumber)

    Description: 
    Find the specified node element within the given table

    Parameters:
    tXml(table)       - Xml fragment to find the value in
    node(string)      - The name of the node
    key(string)       - The name of the key
    keyIsNumber(bool) - Indicates whether the table index is a number or a string

    Returns:
    nil or a table of the found values within the Xml
===============================================================================]]
function GetParsedXmlValuesByKey(tXml, node, key, keyIsNumber)
	local tParams = {}

	keyIsNumber = keyIsNumber or false
	for k,v in pairs(tXml) do
		if (v["Name"] == node) then
			local keyValue
			
			-- get the key
			for nodeKey, nodeValue in pairs(v["ChildNodes"]) do
				if (nodeValue["Name"] == key) then
					if (keyIsNumber == true) then
						keyValue = tonumber(nodeValue.Value)
					else
						keyValue = tostring(nodeValue.Value)
					end
					break
				end
			end

			-- get other tags
			tParams[keyValue] = {}
			for nodeKey, nodeValue in pairs(v["ChildNodes"]) do
				if (nodeValue["Name"] ~= key) then
					tParams[keyValue][nodeValue.Name] = nodeValue.Value
				end
			end
		end
	end

	return tParams
end

--[[=============================================================================
    GetParsedXmlVaulesByKeyAttribute(tXml, node, key, keyIsNumber)

    Description: 
    Find the specified node attribute within the given table

    Parameters:
    tXml(table)       - Xml fragment to find the value in
    node(string)      - The name of the node
    key(string)       - The name of the key
    keyIsNumber(bool) - Indicates whether the table index is a number or a string

    Returns:
    nil or a table of the found values within the Xml
===============================================================================]]
function GetParsedXmlVaulesByKeyAttribute(tXml, node, key, keyIsNumber)
	local tParams = {}

	keyIsNumber = keyIsNumber or false
	for k,v in pairs(tXml["ChildNodes"]) do
		if (v["Name"] == node) then
			local keyValue

			if (keyIsNumber == true) then
				keyValue = tonumber(v["Attributes"][key])
			else
				keyValue = v["Attributes"][key]
			end
			
			tParams[keyValue] = v["Value"]
		end
	end

	return tParams
end

--[[=============================================================================
    BuildSimpleXml(tag, tData, escapeValue)

    Description: 
    Find the specified node within the given table

    Parameters:
    tag(string)       - Xml tag name to create
    tData(table)      - key value pairs that will be added as elements under tag
    escapeValue(bool) - Indicates whether the values should be escaped or not

    Returns:
    nil or an Xml fragment the specified node within the table
===============================================================================]]
function BuildSimpleXml(tag, tData, escapeValue)
	local xml = ""
	
	if (tag ~= nil) then
		xml = "<" .. tag .. ">"
	end
	
	for k,v in pairs(tData) do
		xml = xml .. "<" .. k
		if (type(v) == "table") then
			-- handle attributes
			for kAttrib, vAttrib in pairs(v.attributes) do
				xml = xml .. ' ' .. kAttrib .. '=\"' .. vAttrib .. '\"'
			end
			xml = xml .. ">" .. InsertValue(v.value, escapeValue) .. "</" .. k .. ">"
		else
			xml = xml .. ">" .. InsertValue(v, escapeValue) .. "</" .. k .. ">"
		end
	end
	
	if (tag ~= nil) then
		xml = xml .. "</" .. tag .. ">"
	end

	--DbgTrace("BuildSimpleXml(): " .. xml)

	return xml
end

--[[=============================================================================
    InsertValue(value, escapeValue)

    Description: 
    Return the given value if escapeValue is true it will escape any special
    characters in the value

    Parameters:
    value(string)     - value to be manipulated
    escapeValue(bool) - Indicates whether the values should be escaped or not

    Returns:
    The value given or an escaped value if specified
===============================================================================]]
function InsertValue(value, escapeValue)

	if (escapeValue) then
		value = C4:XmlEscapeString(tostring(value))
	end

	return value
end

--[[=============================================================================
    StartElement(tag)

    Description: 
    Wrap the given tag as an Xml element (i.e. <tag>)

    Parameters:
    tag(string) - The name of the item to be wrapped as a starting Xml element

    Returns:
    The value wrapped as Xml tag
===============================================================================]]
function StartElement(tag)
	return "<" .. tag .. ">"
end

--[[=============================================================================
    EndElement(tag)

    Description: 
    Wrap the given tag as an Xml end element (i.e. </tag>)

    Parameters:
    tag(string) - The name of the item to be wrapped as a ending Xml element

    Returns:
    The value wrapped as ending Xml tag
===============================================================================]]
function EndElement(tag)
	return "</" .. tag .. ">"
end

--[[=============================================================================
    AddElement(tag, data)

    Description: 
    Wrap the given tag and value as an Xml element (i.e. <tag>data</tag>)

    Parameters:
    tag(string)  - The name of the item to be wrapped as an Xml element
    data(string) - The value of the Xml element being created

    Returns:
    The value wrapped as Xml tag and value
===============================================================================]]
function AddElement(tag, data)
	LogTrace("tag = " .. tag)
	LogTrace("data = " .. data)
	
	return "<" .. tag .. ">" .. data .. "</" .. tag .. ">"
end

--[[=============================================================================
    Initial driver initialization and destruction functions

    Copyright 2016 Control4 Corporation. All Rights Reserved.
===============================================================================]]

-- Set template version for this file
if (TEMPLATE_VERSION ~= nil) then
	TEMPLATE_VERSION.c4_init = "2016.01.08"
end

--[[=============================================================================
    OnDriverInit()

    Description
    Invoked by director when a driver is loaded. This API is provided for the
    driver developer to contain all of the driver objects that will require
    initialization.

    Parameters
    None

    Returns
    Nothing
===============================================================================]]
function OnDriverInit()
	gInitializingDriver = true
	C4:ErrorLog("INIT_CODE: OnDriverInit()")

	-- Call all ON_DRIVER_EARLY_INIT functions.
	for k,v in pairs(ON_DRIVER_EARLY_INIT) do
		if (ON_DRIVER_EARLY_INIT[k] ~= nil and type(ON_DRIVER_EARLY_INIT[k]) == "function") then
			C4:ErrorLog("INIT_CODE: ON_DRIVER_EARLY_INIT." .. k .. "()")
			local status, err = pcall(ON_DRIVER_EARLY_INIT[k])
			if (not status) then
				C4:ErrorLog("LUA_ERROR: " .. err)
			end
		end
	end

	-- Call all ON_DRIVER_INIT functions
	for k,v in pairs(ON_DRIVER_INIT) do
		if (ON_DRIVER_INIT[k] ~= nil and type(ON_DRIVER_INIT[k]) == "function") then
			C4:ErrorLog("INIT_CODE: ON_DRIVER_INIT." .. k .. "()")
			local status, err = pcall(ON_DRIVER_INIT[k])
			if (not status) then
				C4:ErrorLog("LUA_ERROR: " .. err)
			end
		end
	end

	-- Fire OnPropertyChanged to set the initial Headers and other Property
	-- global sets, they'll change if Property is changed.
	for k,v in pairs(Properties) do
		C4:ErrorLog("INIT_CODE: Calling OnPropertyChanged - " .. k .. ": " .. v)
		local status, err = pcall(OnPropertyChanged, k)
		if (not status) then
			C4:ErrorLog("LUA_ERROR: " .. err)
		end
	end

	gInitializingDriver = false
end

--[[=============================================================================
    OnDriverLateInit()

    Description
    Invoked by director after all drivers in the project have been loaded. This
    API is provided for the driver developer to contain all of the driver
    objects that will require initialization after all drivers in the project
    have been loaded.

    Parameters
    None

    Returns
    Nothing
===============================================================================]]
function OnDriverLateInit()
	C4:ErrorLog("INIT_CODE: OnDriverLateInit()")
	
	-- Call all ON_DRIVER_LATEINIT functions
	for k,v in pairs(ON_DRIVER_LATEINIT) do
		if (ON_DRIVER_LATEINIT[k] ~= nil and type(ON_DRIVER_LATEINIT[k]) == "function") then
			C4:ErrorLog("INIT_CODE: ON_DRIVER_LATEINIT." .. k .. "()")
			ON_DRIVER_LATEINIT[k]()
		end
	end
end


--[[=============================================================================
    OnDriverDestroyed()
    Function called by Director when a driver is removed. Release things this
    driver has allocated such as timers.

    Parameters
    None

    Returns
    Nothing
===============================================================================]]
function OnDriverDestroyed()
	C4:ErrorLog("INIT_CODE: OnDriverDestroyed()")
	
	-- Call all ON_DRIVER_DESTROYED functions
	for k, v in pairs(ON_DRIVER_DESTROYED) do
		if (ON_DRIVER_DESTROYED[k] ~= nil and type(ON_DRIVER_DESTROYED[k]) == "function") then
			C4:ErrorLog("INIT_CODE: ON_DRIVER_DESTROYED." .. k .. "()")
			ON_DRIVER_DESTROYED[k]()
		end
	end
end
--[[=============================================================================
    Functions for handling and executing commands and actions

    Copyright 2016 Control4 Corporation. All Rights Reserved.
===============================================================================]]

-- Set template version for this file
if (TEMPLATE_VERSION ~= nil) then
	TEMPLATE_VERSION.c4_command = "2016.01.08"
end

--[[=============================================================================
    ExecuteCommand(sCommand, tParams)

    Description
    Function called by Director when a command is received for this DriverWorks
    driver. This includes commands created in Composer programming.

    Parameters
    sCommand(string) - Command to be sent
    tParams(table)   - Lua table of parameters for the sent command

    Returns
    Nothing
===============================================================================]]
function ExecuteCommand(sCommand, tParams)
	LogTrace("ExecuteCommand(" .. sCommand .. ")")
	LogInfo(tParams)

	-- Remove any spaces (trim the command)
	local trimmedCommand = string.gsub(sCommand, " ", "")
	local status, ret

	-- if function exists then execute (non-stripped)
	if (EX_CMD[sCommand] ~= nil and type(EX_CMD[sCommand]) == "function") then
		status, ret = pcall(EX_CMD[sCommand], tParams)
	-- elseif trimmed function exists then execute
	elseif (EX_CMD[trimmedCommand] ~= nil and type(EX_CMD[trimmedCommand]) == "function") then
		status, ret = pcall(EX_CMD[trimmedCommand], tParams)
	elseif (EX_CMD[sCommand] ~= nil) then
		QueueCommand(EX_CMD[sCommand])
		status = true
	else
		LogInfo("ExecuteCommand: Unhandled command = " .. sCommand)
		status = true
	end
	
	if (not status) then
		LogError("LUA_ERROR: " .. ret)
	end
	
	return ret -- Return whatever the function returns because it might be xml, a return code, and so on
end

--[[=============================================================================
    EX_CMD.LUA_ACTION(tParams)

    Description
    Function called for any actions executed by the user from the Actions Tab
    in Composer.

    Parameters
    tParams(table) - Lua table of parameters for the command option

    Returns
    Nothing
===============================================================================]]
function EX_CMD.LUA_ACTION(tParams)
	if (tParams ~= nil) then
		for cmd, cmdv in pairs(tParams) do
			if (cmd == "ACTION" and cmdv ~= nil) then
				local status, err = pcall(LUA_ACTION[cmdv], tParams)
				if (not status) then
					LogError("LUA_ERROR: " .. err)
				end
				break
			end
		end
	end
end

--[[=============================================================================
    ReceivedFromProxy(idBinding, sCommand, tParams)

    Description
    Function called for any actions executed by the user from the Actions Tab
    in Composer.

    Parameters
    idBinding(int)   - Binding ID of the proxy that sent a BindMessage to the
                       DriverWorks driver.
    sCommand(string) - Command that was sent
    tParams(table)   - Lua table of received command parameters

    Returns
    Nothing
===============================================================================]]
function ReceivedFromProxy(idBinding, sCommand, tParams)

	if (sCommand ~= nil) then

		-- initial table variable if nil
		if (tParams == nil) then
			tParams = {}
		end
		
		LogTrace("ReceivedFromProxy(): " .. sCommand .. " on binding " .. idBinding .. "; Call Function PRX_CMD." .. sCommand .. "()")
		LogInfo(tParams)

		if ((PRX_CMD[sCommand]) ~= nil) then
			local status, err = pcall(PRX_CMD[sCommand], idBinding, tParams)
			if (not status) then
				LogError("LUA_ERROR: " .. err)
			end
		else
			LogInfo("ReceivedFromProxy: Unhandled command = " .. sCommand)
		end
	end
end

--[[
	This function is called when a UI (Navigator) requests data, and
	calls the function requested.
--]]
function UIRequest(sRequest, tParams)
	local ret = ""

	if (sRequest ~= nil) then
		tParams = tParams or {}   -- initial table variable if nil
		LogTrace("UIRequest(): " .. sRequest .. "; Call Function UI_REQ." .. sRequest .. "()")
		LogInfo(tParams)

		if (UI_REQ[sRequest]) ~= nil then
			ret = UI_REQ[sRequest](tParams)
		else
			LogWarn("UIRequest: Unhandled request = " .. sRequest)
		end
	end

	return ret
end

--[[=============================================================================
    ON_INIT, Timer,s and Property management functions

    Copyright 2016 Control4 Corporation. All Rights Reserved.
===============================================================================]]

-- Set template version for this file
if (TEMPLATE_VERSION ~= nil) then
	TEMPLATE_VERSION.c4_common = "2016.01.08"
end

--[[=============================================================================
    Create and Initialize Logging
===============================================================================]]
function ON_DRIVER_EARLY_INIT.c4_common()
	-- Create a logger
	LOG = c4_log:new("Template_c4z Change Name")
end

function ON_DRIVER_INIT.c4_common()
	-- Create Log Timer
	gC4LogTimer = c4_timer:new("Log Timer", 45, "MINUTES", OnLogTimerExpired)
end

--[[=============================================================================
    Log timer callback function
===============================================================================]]
function OnLogTimerExpired()
	LogWarn("Turning Log Mode Off (timer expired)")
	gC4LogTimer:KillTimer()
	
	C4:UpdateProperty("Log Mode", "Off")
	OnPropertyChanged("Log Mode")
end

gForceLogging = false
function ON_PROPERTY_CHANGED.LogMode(propertyValue)
	gC4LogTimer:KillTimer()
	
	if (gForceLogging) then
		LOG:OutputPrint(true)
		LOG:OutputC4Log(true)
	else
		LOG:OutputPrint(propertyValue:find("Print") ~= nil)
		LOG:OutputC4Log(propertyValue:find("Log") ~= nil)
		if (propertyValue == "Off") then
			return
		end
		
		gC4LogTimer:StartTimer()
	end
end

function ON_PROPERTY_CHANGED.LogLevel(propertyValue)
	if (gForceLogging) then
		LOG:SetLogLevel("5 - Debug")
	else
		LOG:SetLogLevel(propertyValue)
	end
end

--[[=============================================================================
    Print Template Versions
===============================================================================]]
function TemplateVersion()
	print ("\nTemplate Versions")
	print ("-----------------------")
	for k, v in pairs(TEMPLATE_VERSION) do
		print (k .. " = " .. v)
	end
	
	print ("")
end
--[[=============================================================================
    DeviceConnectionBase Class

    Copyright 2016 Control4 Corporation. All Rights Reserved.
===============================================================================]]

-- Set template version for this file
if (TEMPLATE_VERSION ~= nil) then
	TEMPLATE_VERSION.c4_device_connection_base = "2016.01.08"
end

COMMAND_QUEUE_SIZE = 100
DEFAULT_COMMAND_DELAY_INTERVAL = 100            -- Don't send consecutive commands faster than this many milliseconds
DEFAULT_COMMAND_RESPONSE_INTERVAL = 3           -- If we haven't received and ACK after this many seconds, try again
DEFAULT_RETRY_COUNT_MAX = 3

function ON_DRIVER_EARLY_INIT.c4_device_connection_base()
	gReceiveBuffer = ""
	gIsUrlConnected = false
end

DeviceConnectionBase = inheritsFrom(nil)

function DeviceConnectionBase:construct()

	self._IsConnected = false
	self._SendTimer = nil
	self._WaitResponseTimer = nil
	self._CommandQueue = nil
	self._Priority1CommandQueue = nil
	self._Priority2CommandQueue = nil
	self._LastCommand = nil
	self._ExpectAck = false
	self._CommandRetryCount = 0
	self._RetryCountMax = DEFAULT_RETRY_COUNT_MAX

	self._SendCommandDelayMS = DEFAULT_COMMAND_DELAY_INTERVAL
	self._CommandResponseWaitS = DEFAULT_COMMAND_RESPONSE_INTERVAL
	
	-- Polling
	self._PollingInterval = 0
	self._PollingUnits = "SECONDS"
	self._PollingTimer = nil
end

function DeviceConnectionBase:Initialize(ExpectAck, CommandDelayInterval, CommandResponseInterval, CallbackParam)

	if (ExpectAck ~= nil) then
		self._ExpectAck = ExpectAck
	end

	if (CommandDelayInterval ~= nil) then
		self._SendCommandDelayMS = CommandDelayInterval
	end

	if (CommandResponseInterval ~= nil) then
		self._CommandResponseWaitS = CommandResponseInterval
	end

	self._CommandQueue = c4_queue:new()
	self._CommandQueue:SetMaxSize(COMMAND_QUEUE_SIZE)
	self._CommandQueue:SetName("Command Queue")

	self._Priority1CommandQueue = c4_queue:new()
	self._Priority1CommandQueue:SetMaxSize(COMMAND_QUEUE_SIZE)
	self._Priority1CommandQueue:SetName("P1 Queue")

	self._Priority2CommandQueue = c4_queue:new()
	self._Priority2CommandQueue:SetMaxSize(COMMAND_QUEUE_SIZE)
	self._Priority2CommandQueue:SetName("P2 Queue")

	-- usually only one of these timers will be used, but it's pretty low overhead to instantiate both of them
	self._SendTimer = c4_timer:new("SendCommand", self._SendCommandDelayMS, "MILLISECONDS", DeviceConnectionBase.OnSendTimeExpired, false, CallbackParam)
	self._WaitResponseTimer = c4_timer:new("WaitResponse", self._CommandResponseWaitS, "SECONDS", DeviceConnectionBase.OnWaitTimeExpired, false, CallbackParam)
end

function DeviceConnectionBase:InitPolling(PollingInterval, PollingUnits, CallbackParam)
	LogFatal("DeviceConnectionBase:InitPolling()")
	if (PollingInterval ~= nil) then
		self._PollingInterval = PollingInterval
	end
	
	self._PollingUnits = PollingUnits or self._PollingUnits

	LogFatal("self._PollingInterval: %s, self._PollingUnits: %s", tostring(self._PollingInterval), tostring(self._PollingUnits))
	
	-- create polling timer
	self._PollingTimer = c4_timer:new("Polling", self._PollingInterval, self._PollingUnits, DeviceConnectionBase.OnPollingTimerExpired, false, CallbackParam)
end

function DeviceConnectionBase:StartPolling(interval, units)
	LogFatal("DeviceConnectionBase:StartPolling()")
	LogFatal("self._PollingTimer: %s", tostring(self._PollingTimer))
	
	if (self._PollingTimer ~= nil) then
		self._PollingTimer:KillTimer()
		
		local timer_units = units or self._PollingTimer._units
		local timer_interval = interval or self._PollingInterval

		self._PollingTimer:StartTimer(timer_interval, timer_units)
	end
end

function DeviceConnectionBase:StopPolling()
	LogFatal("DeviceConnectionBase:StopPolling()")
	self._PollingTimer:KillTimer()
end

function DeviceConnectionBase:SetExpectACK(ExpectACK)
	self._ExpectAck = ExpectACK
end

function DeviceConnectionBase:SetCommandDelayInterval(DelayInterval)
	self._SendCommandDelayMS = DelayInterval
end

function DeviceConnectionBase:SetResponseWaitInterval(WaitInterval)
	self._CommandResponseWaitS = WaitInterval
end

function DeviceConnectionBase:ReceivedFromCom(sData)

	gReceiveBuffer = gReceiveBuffer .. sData
	LogTrace("ReceivedFromCom  ReceiveBuffer is now {{{%s}}}", gReceiveBuffer)

	message = self:GetMessage()
	while (message ~= nil and message ~= "") do
		status, err = pcall(HandleMessage, message)
		if (status) then
			message = self:GetMessage()
		else
			LogError("LUA_ERROR: " .. err)
			message = ""
			gReceiveBuffer = ""
		end
	end
end

function DeviceConnectionBase:SetConnection(IsConnected, method)
	self._IsConnected = IsConnected
	gControlMethod = method
	C4:PersistSetValue("gControlMethod", gControlMethod)
end

function DeviceConnectionBase:ControlMethod()
	-- Override in derived class
	print("WARNING: Need to override ControlMethod - should never be called")
	
	return ""
end

function DeviceConnectionBase:StartCommandTimer(...)
	local value = select(1, ...)
	local units = select(2, ...)
	local command_name = select(3, ...) or ""

	self._WaitResponseTimer:KillTimer()
	self._SendTimer:KillTimer()

	if (self._ExpectAck) then
		-- expecting an ACK set the Response Wait timer
		local timer_units = units or self._WaitResponseTimer._units
		local timer_interval = value or self._CommandResponseWaitS

		self._WaitResponseTimer:StartTimer(timer_interval, timer_units)
		LogTrace(string.format("Starting wait Timer:  %d", self._WaitResponseTimer._timerID) .. " for " .. command_name)
	else
		-- no ACK expected, just wait the designated amount of time and send another command
		local timer_units = units or self._SendTimer._units
		local timer_interval = value or self._SendCommandDelayMS

		self._SendTimer:StartTimer(timer_interval, timer_units)
		LogTrace(string.format("Starting Send Timer:  %d for %s (timer_interval = %d, timer_units = %s)", self._SendTimer._timerID, command_name, timer_interval, timer_units))
	end
end

-- Note the '.' instead of the ':'
function DeviceConnectionBase.OnSendTimeExpired(Instance)
	LogTrace("Send Timer expired")
	Instance._SendTimer:KillTimer()

	local tCommand = Instance._LastCommand
	if (tCommand.command_name ~= nil) then
		LogTrace("Send Timer expired - Last Command: %s, Send Next Command", tostring(tCommand.command_name))
	elseif (type(tCommand) == "string") then
		LogTrace("Send Timer expired - Last Command: %s, Send Next Command", tostring(tCommand))
	end
	
	Instance._LastCommand = nil
	Instance:SendNextCommand()

	if (DoEvents ~= nil and type(DoEvents) == "function") then
		DoEvents()
	end
end

function DeviceConnectionBase.OnWaitTimeExpired(Instance)
	LogTrace("Wait Timer expired")
	Instance._WaitResponseTimer:KillTimer()
	
     --if a response is not received from the TV, the queue will stall, so resending the last command will not rememdy the situation
     --the best remedy is to advance the queue
	gCon:HandleACK()
	
	--[[
	Instance._CommandRetryCount = Instance._CommandRetryCount + 1

	if (Instance._CommandRetryCount >= Instance._RetryCountMax) then
		-- To many retries, pop the current command and try the next one
		Instance._CommandRetryCount = 0
		Instance:SendNextCommand()
		
	     if (DoEvents ~= nil and type(DoEvents) == "function") then
		    DoEvents()
	     end		
	else
		Instance:SendLastCommand()
		
		--if a response is not received from the TV, the queue will stall, so resending the last command will not rememdy the situation
		--the best remedy is to advance the queue
		--Instance._CommandRetryCount = 0
		--gCon:HandleACK()

	end
	--]]
end

function DeviceConnectionBase.OnPollingTimerExpired(Instance)
	LogTrace("Polling Timer expired")
	Instance._PollingTimer:KillTimer()

	OnPollingTimerExpired()
	
	Instance._PollingTimer:StartTimer(Instance._PollingInterval)
end

function DeviceConnectionBase:HandleACK()
	self._LastCommand = nil
	
	self._WaitResponseTimer:KillTimer()
	self._CommandRetryCount = 0
	self:SendNextCommand()
	
     if (DoEvents ~= nil and type(DoEvents) == "function") then
		DoEvents()
     end
end

function DeviceConnectionBase:QueueEmpty()
	return (self._CommandQueue:empty() and self._Priority1CommandQueue:empty() and self._Priority2CommandQueue:empty())
end

function DeviceConnectionBase:QueueCommand(sCommand, ...)
--	LogTrace("QueueCommand(%s)", sCommand)
	local command_delay = select(1, ...)
	local delay_units = select(2, ...)
	local command_name = select(3, ...)

	if (sCommand == nil) or (sCommand == "") then
		return
	end

	if (self._LastCommand == nil) then
		self._CommandQueue:push(sCommand, command_delay, delay_units, command_name)
		self._LastCommand = self._CommandQueue:pop()
		self:SendCommand(sCommand, command_delay, delay_units, command_name)
	else
		self._CommandQueue:push(sCommand, command_delay, delay_units, command_name)
	end
end

function DeviceConnectionBase:QueuePriority1Command(sCommand, ...)
	LogTrace("QueuePriority1Command(%s)", sCommand)
	local command_delay = select(1, ...)
	local delay_units = select(2, ...)
	local command_name = select(3, ...)

	if (sCommand == nil) or (sCommand == "") then
		return
	end

	if (self._LastCommand == nil) then
		self._Priority1CommandQueue:push(sCommand, command_delay, delay_units, command_name)
		self._LastCommand = self._Priority1CommandQueue:pop()
		self:SendCommand(sCommand, command_delay, delay_units, command_name)
	else
		self._Priority1CommandQueue:push(sCommand, command_delay, delay_units, command_name)
	end
end

function DeviceConnectionBase:QueuePriority2Command(sCommand, ...)
	LogTrace("QueuePriority2Command(%s)", sCommand)
	local command_delay = select(1, ...)
	local delay_units = select(2, ...)
	local command_name = select(3, ...)

	if (sCommand == nil) or (sCommand == "") then
		return
	end

	if (self._LastCommand == nil) then
		self._Priority2CommandQueue:push(sCommand, command_delay, delay_units, command_name)
		self._LastCommand = self._Priority2CommandQueue:pop()
		self:SendCommand(sCommand, command_delay, delay_units, command_name)
	else
		self._Priority2CommandQueue:push(sCommand, command_delay, delay_units, command_name)
	end
end

function DeviceConnectionBase:SendNextCommand()
	LogTrace("DeviceConnectionBase:SendNextCommand")

	local tCommand = nil
	if (not self._Priority1CommandQueue:empty()) then
		tCommand = self._Priority1CommandQueue:pop()
		LogTrace(tostring(gCon._Priority1CommandQueue))
	elseif (not self._Priority2CommandQueue:empty()) then
		tCommand = self._Priority2CommandQueue:pop()
		LogTrace(tostring(gCon._Priority2CommandQueue))
	elseif (not self._CommandQueue:empty()) then
		tCommand = self._CommandQueue:pop()
		LogTrace(tostring(gCon._CommandQueue))
	end
	
	if (tCommand ~= nil) then
		self._LastCommand = tCommand
		local sCommand = tCommand.command
		local command_delay = tCommand.command_delay
		local delay_units = tCommand.delay_units
		local command_name = tCommand.command_name

		if (sCommand == nil or sCommand == "") then
			self._SendTimer:KillTimer()
			self._WaitResponseTimer:KillTimer()
		else
			LogTrace("SendCommand: %s", sCommand)
			self:SendCommand(sCommand, command_delay, delay_units, command_name)
		end
	end
end

function DeviceConnectionBase:SendLastCommand()
--	LogTrace("DeviceConnectionBase:SendLastCommand")

	local tCommand = self._LastCommand
	if (tCommand ~= nil) then
		local sCommand = tCommand.command
		local command_delay = tCommand.command_delay
		local delay_units = tCommand.delay_units
		local command_name = tCommand.command_name

		if (sCommand == nil or sCommand == "") then
			self._SendTimer:KillTimer()
			self._WaitResponseTimer:KillTimer()
		else
			LogTrace("SendCommand: %s", sCommand)
			self:SendCommand(sCommand, command_delay, delay_units, command_name)
		end
	end
end

function DeviceConnectionBase:SendCommand()
	-- Dummy routine.  Override in derived class
	print("Need to override SendCommand - should never be called")
end


function DeviceConnectionBase:GetMessage()
	
	-- Brain dead version of this routine. Just return the current receive buffer.
	-- It's very likely that a GetMessage() function will need to be created
	if (GetMessage ~= nil and type(GetMessage) == "function") then
		return GetMessage()
	else
		local ComMessage = gReceiveBuffer
		gReceiveBuffer = ""

		return ComMessage
	end
end

--[[=============================================================================
    Other Connection Functions
===============================================================================]]

function ReceivedFromSerial(idBinding, sData)
	if (gCon.ReceivedFromSerial == nil) then return end --serial is bound but not the current control method
	gCon:ReceivedFromSerial(idBinding, sData)
end

function ReceivedFromNetwork(idBinding, nPort, sData)
	gCon:ReceivedFromNetwork(idBinding, nPort, sData)
end

function OnServerDataIn(nHandle, strData)
--	LogTrace("Received Data on Handle: " .. nHandle .. ": " .. strData)
--	LogTrace("Data Is: %s", HexToString(strData))
	gCon:ReceivedFromNetworkServer(nHandle, strData)
end


--[[=============================================================================
    The ReceivedAsync function is called in response to 'url_get_request'. 
    The ticketId is the number returned from the request.
===============================================================================]]
function ReceivedAsync(ticketId, strData, responseCode, tHeaders)
	strData = strData or ""
	responseCode = responseCode or 0
	tHeaders = tHeaders or {}

--	LogTrace("ReceivedAsync[" .. ticketId .. "]: Response Code: " .. responseCode .. " Length: " .. string.len(strData))
--	LogTrace(tHeaders)

	gCon:ReceivedAsync(ticketId, strData, responseCode, tHeaders)
end
	
--[[=============================================================================
    OnBindingChanged(idBinding, class, bIsBound)
  
    Description:
    Function called by Director when a binding changes state(bound or unbound).
  
    Parameters:
    idBinding(int) - ID of the binding whose state has changed.
    class(string)  - Class of binding that has changed.
                     A single binding can have multiple classes(i.e. COMPONENT,
                     STEREO, RS_232, etc).
                     This indicates which has been bound or unbound.
    bIsBound(bool) - Whether the binding has been bound or unbound.
  
    Returns:
    None
===============================================================================]]
function OnBindingChanged(idBinding, class, bIsBound)
	
	LogTrace("OnBindingChanged(): idBinding = " .. tostring(idBinding) .. ", class = " .. class .. ", bIsBound = " .. tostring(bIsBound))
	if (idBinding == SERIAL_BINDING_ID) then
		gIsSerialConnected = bIsBound
		SetControlMethod()
		OnSerialConnectionChanged(idBinding, class, bIsBound)
	elseif (idBinding == IR_BINDING_ID) then
		gIsIRConnected = bIsBound
		SetControlMethod()
		OnIRConnectionChanged(idBinding, class, bIsBound)
	elseif(OnConnectionChanged ~= nil and type(OnConnectionChanged) == "function") then
		OnConnectionChanged(idBinding, class, bIsBound)
	end
end

--[[=============================================================================
    OnNetworkBindingChanged(idBinding, bIsBound)
  
    Description:
    Function called by Director when a network binding changes state(bound or unbound).
  
    Parameters:
    idBinding(int) - ID of the binding whose state has changed.
    bIsBound(bool) - Whether the binding has been bound or unbound.
  
    Returns:
    None
===============================================================================]]
function OnNetworkBindingChanged(idBinding, bIsBound)
	LogTrace('OnNetworkBindingChanged(): idBinding = ' .. tostring(idBinding) .. ' bIsBound = ' .. tostring(bIsBound))

	gIsNetworkConnected = bIsBound
	SetControlMethod()
	OnNetworkConnectionChanged(idBinding, bIsBound)
	if (bIsBound) then
		-- Start a special instance of reconnect timer to eventually do NetConnect if not done automatically
		gCon._NetworkReconnectTimer:StartTimer(gNetworkReconnectInterval) 	
	end	
end

--[[=============================================================================
    OnConnectionStatusChanged(idBinding, nPort, sStatus)
  
    Description:
    Sets the updated status of the specified binding
  
    Parameters:
    idBinding(int)  - ID of the binding whose status has changed
    nPort(int)      - The communication port of the specified bindings connection
    sStatus(string) - "ONLINE" if the connection status is to be set to Online,
                      any other value will set the status to Offline
  
    Returns:
    None
===============================================================================]]
function OnConnectionStatusChanged(idBinding, nPort, sStatus)
	LogTrace("OnConnectionStatusChanged[" .. idBinding .. " (" .. tostring(nPort) .. ")]: " .. sStatus)

	local isOnline = false

	gNetworkStatus = sStatus
	if (sStatus == "ONLINE") then
		isOnline = true
	end
	if(gCon == nil) then SetControlMethod() end
	gCon:SetOnlineStatus(isOnline)
	OnNetworkStatusChanged(idBinding, nPort, sStatus)
end

--[[=============================================================================
    SetControlMethod()
  
    Description:
    Sets the control method type for the drivers internal infrastructure
  
    Parameters:
    None
  
    Returns:
    The type of control method for the drivers connection(i.e. Network, Serial,
    IR, or (none))
===============================================================================]]
function SetControlMethodFlags()
	local controlMethod = gControlMethod or "NOT  DEFINED"
	print("Control Method is ".. controlMethod)
	if(gControlMethod == "NetworkServer") then gIsNetworkServerConnected = true
	elseif(gControlMethod == "Network") then gIsNetworkConnected = true
	elseif(gControlMethod == "URL") then gIsUrlConnected = true
	elseif(gControlMethod == "Serial") then gIsSerialConnected = true
	elseif(gControlMethod == "IR") then gIsIRConnected = true
	end
end

function SetControlMethod()
	if (gIsNetworkConnected == false) and (gCon._NetworkReconnectTimer ~= nil) then
		--housekeeping when changing from network control to serial or IR control
		gCon._NetworkReconnectTimer:KillTimer() 
	end


	if( gIsNetworkServerConnected) then
		-- connect to NetworkServer communicator if not already connected
		if (gCon == nil or gCon.ControlMethod() ~= "NetworkServer") then
			gCon = NetworkServerConnectionBase:new()
			gCon:Initialize()
		end
		gCon:SetConnection(true, "NetworkServer")
	elseif (gIsNetworkConnected) then
		-- connect to Network communicator if not already connected
		if (gCon == nil or gCon.ControlMethod() ~= "Network") then
			gCon = NetworkConnectionBase:new(NETWORK_BINDING_ID, NETWORK_PORT)
			gCon:Initialize(COM_USE_ACK, COM_COMMAND_DELAY_MILLISECONDS, COM_COMMAND_RESPONSE_TIMEOUT_SECONDS)
		end
		gCon:SetConnection(true, "Network")
	elseif (gIsUrlConnected) then
		-- connect to URL communicator if not already connected
		if (gCon == nil or gCon.ControlMethod() ~= "URL") then
			gCon = UrlConnectionBase:new()
			gCon:Initialize(COM_USE_ACK, COM_COMMAND_DELAY_MILLISECONDS, COM_COMMAND_RESPONSE_TIMEOUT_SECONDS)
		end
		gCon:SetConnection(true, "URL")
	elseif (gIsSerialConnected) then
		-- connect to Serial communicator if not already connected
		if (gCon == nil or gCon.ControlMethod() ~= "Serial") then
			gCon = SerialConnectionBase:new(SERIAL_BINDING_ID)
			gCon:Initialize(COM_USE_ACK, COM_COMMAND_DELAY_MILLISECONDS, COM_COMMAND_RESPONSE_TIMEOUT_SECONDS)
			gCon:InitPolling(tonumber(gPollingTimerInterval), "MINUTES", gCon)
		end
		gCon:SetConnection(true, "Serial")
	elseif (gIsIRConnected) then
		-- connect to IR communicator if not already connected
		if (gCon == nil or gCon.ControlMethod() ~= "IR") then
			gCon = IRConnectionBase:new(IR_BINDING_ID)
			gCon:Initialize(COM_USE_ACK, COM_COMMAND_DELAY_MILLISECONDS, COM_COMMAND_RESPONSE_TIMEOUT_SECONDS)
		end
		gCon:SetConnection(true, "IR")
	else
		if (gCon ~= nil) then
			gCon:SetConnection(false, "(none)")
		end
		-- gCon = nil
	end

	gCon._CommandQueue:clear()
	gCon._Priority1CommandQueue:clear()
	gCon._Priority2CommandQueue:clear()
end

--[[=============================================================================
    ValidateControlMethod(controlMethod)
  
    Description:
    Identifies whether the specified control method has a valid connection
  
    Parameters:
    controlMethod(string) - The communication we are validating against
                            Valid types are (Network, Serial, and IR)
  
    Returns:
    true if the controlMethod specified has been connected, false otherwise.
===============================================================================]]
function ValidateControlMethod(controlMethod)
	local isValid = false

	if (controlMethod == "Network") and (gIsNetworkConnected) then
		isValid = true
	elseif (controlMethod == "URL") and (gIsUrlConnected) then
		isValid = true
	elseif (controlMethod == "Serial") and (gIsSerialConnected) then
		isValid = true
	elseif (controlMethod == "IR") and (gIsIRConnected) then
		isValid = true
	end

	return isValid
end

--[[=============================================================================
    Functions for Testing different aspects of the environment

    Copyright 2016 Control4 Corporation. All Rights Reserved.
===============================================================================]]

-- Set template version for this file
if (TEMPLATE_VERSION ~= nil) then
	TEMPLATE_VERSION.c4_diagnostics = "2016.01.08"
end

function DisplayGlobals()

	print ("Global Variables")
	print ("----------------------------")
	for k,v in pairs(_G) do                             -- globals
		if not (type(v) == "function") then
			if (string.find(k, "^g%L")  == 1) then
				print(k .. ":  " .. tostring(v))
				if (type(v) == "table") then
					C4PrintTable(v, "   ")
				end
			end
		end
	end

	print ("")
end

function C4PrintTable(tValue, sIndent)

	sIndent = sIndent or "   "
	for k,v in pairs(tValue) do

		print(sIndent .. tostring(k) .. ":  " .. tostring(v))
		if (type(v) == "table") then
			C4PrintTable(v, sIndent .. "   ")
		end
	end
end
--[[=============================================================================
    Base for an IR connection driver

    Copyright 2016 Control4 Corporation. All Rights Reserved.
===============================================================================]]

-- Set template version for this file
if (TEMPLATE_VERSION ~= nil) then
	TEMPLATE_VERSION.c4_ir_connection = "2016.01.08"
end

IRConnectionBase = inheritsFrom(DeviceConnectionBase)

function IRConnectionBase:construct(BindingID)
	self.superClass():construct()
	self._BindingID = BindingID
end

function IRConnectionBase:Initialize(ExpectAck, DelayInterval, WaitInterval)
	print("tSerConBase:Initialize")
	gControlMethod = "IR"
	self:superClass():Initialize(ExpectAck, DelayInterval, WaitInterval, self)
end

function IRConnectionBase:ControlMethod()
	return "IR"
end

function IRConnectionBase:SendCommand(sCommand, ...)
	if(self._IsConnected) then
		local command_delay = select(1, ...)
		local delay_units = select(2, ...)
		local command_name = select(3, ...)

		C4:SendIR(self._BindingID, sCommand)
		self:StartCommandTimer(command_delay, delay_units, command_name)
	else
		LogWarn("IR connection is not bound. Command not sent.")
	end
end

--[[=============================================================================
    Base for a network connection driver

    Copyright 2016 Control4 Corporation. All Rights Reserved.
===============================================================================]]

-- Set template version for this file
if (TEMPLATE_VERSION ~= nil) then
	TEMPLATE_VERSION.c4_network_connection = "2016.01.08"
end

DEFAULT_POLLING_INTERVAL_SECONDS = 30
DEFAULT_RECONNECT_INTERVAL_SECONDS = 5

gNetworkKeepAliveInterval = DEFAULT_POLLING_INTERVAL_SECONDS
gNetworkReconnectInterval = DEFAULT_RECONNECT_INTERVAL_SECONDS

NetworkConnectionBase = inheritsFrom(DeviceConnectionBase)

function NetworkConnectionBase:construct(BindingID, Port)
	self.superClass():construct()

	self._BindingID = BindingID
	self._Port = Port
	self._LastCheckin = 0
	self._IsOnline = false
	self._KeepAliveTimer = nil
end

function NetworkConnectionBase:Initialize(ExpectAck, DelayInterval, WaitInterval)
	print("NetConBase:Initialize")
	gControlMethod = "Network"
	self:superClass():Initialize(ExpectAck, DelayInterval, WaitInterval, self)
	self._KeepAliveTimer = c4_timer:new("PollingTimer", gNetworkKeepAliveInterval, "SECONDS", NetworkConnectionBase.OnKeepAliveTimerExpired, false, self)
	self._NetworkReconnectTimer = c4_timer:new("NetworkReconnectTimer", gNetworkReconnectInterval, "SECONDS", NetworkConnectionBase.OnNetworkReconnectTimerExpired, false, self)
end

function NetworkConnectionBase:ControlMethod()
	return "Network"
end

function NetworkConnectionBase:SendCommand(sCommand, ...)
	if(self._IsConnected) then
		if(self._IsOnline) then
			local command_delay = select(1, ...)
			local delay_units = select(2, ...)
			local command_name = select(3, ...)

			C4:SendToNetwork(self._BindingID, self._Port, sCommand)
			self:StartCommandTimer(command_delay, delay_units, command_name)
			LogInfo("NetworkConnectionBase:SendCommand(sCommand: " .. sCommand)
		else
			self:CheckNetworkConnectionStatus()
		end
	else
		LogWarn("Not connected to network. Command not sent.")
	end
end

function NetworkConnectionBase:ReceivedFromNetwork(idBinding, nPort, sData)
	self._LastCheckin = 0
	self:ReceivedFromCom(sData)
end

function NetworkConnectionBase:CheckNetworkConnectionStatus()
	if (self._IsConnected and (not self._IsOnline)) then
		LogWarn("Network status is OFFLINE. Trying to reconnect to the device's Control port...")
		C4:NetDisconnect(self._BindingID, self._Port)
		--C4:NetConnect(self._BindingID, self._Port)
		self._NetworkReconnectTimer:StartTimer(gNetworkReconnectInterval)
	end
end

function NetworkConnectionBase.OnKeepAliveTimerExpired(Instance)
	Instance._LastCheckin = Instance._LastCheckin + 1

	if(Instance._LastCheckin == 3) then
		LogWarn("Failed to receive poll responses... initiating network recovery mode...")
		C4:NetDisconnect(Instance._BindingID, Instance._Port)
		Instance._NetworkReconnectTimer:StartTimer(gNetworkReconnectInterval)
		return
	elseif(Instance._LastCheckin > 4) then	
		Instance._LastCheckin = 4
	end

	if (SendKeepAlivePollingCommand ~= nil and type(SendKeepAlivePollingCommand) == "function" and Instance._IsOnline) then
		SendKeepAlivePollingCommand()
	end

	Instance._KeepAliveTimer:StartTimer(gNetworkKeepAliveInterval)
end

function NetworkConnectionBase.OnNetworkReconnectTimerExpired(Instance)
	if (Instance._IsConnected) then
		LogWarn("OnNetworkReconnectTimerExpired: Attempting to reactivate network connection...")
		C4:NetDisconnect(Instance._BindingID, Instance._Port)
		C4:NetConnect(Instance._BindingID, Instance._Port)
		Instance._NetworkReconnectTimer:StartTimer(gNetworkReconnectInterval)
	else
		LogWarn("Cannot attempt to reactivate, the network connection is not bound")
	end
end

function NetworkConnectionBase:SetOnlineStatus(IsOnline)
	self._IsOnline = IsOnline

	if(IsOnline) then
		self._KeepAliveTimer:StartTimer(gNetworkKeepAliveInterval)
		self._NetworkReconnectTimer:KillTimer()
		self._LastCheckin = 0
		if (UpdateProperty ~= nil and type(UpdateProperty) == "function") then
			UpdateProperty("Connected To Network", "true")
		end

		self:SendNextCommand()
	else
		self._KeepAliveTimer:KillTimer()
		self._NetworkReconnectTimer:StartTimer(gNetworkReconnectInterval)
		if (UpdateProperty ~= nil and type(UpdateProperty) == "function") then
			UpdateProperty("Connected To Network", "false")
		end
	end
end

function ON_DRIVER_LATEINIT.c4_network_connection()
	-- Ensure existing connection is taken into consideration (useful on Driver Update)
	if (gCon ~= nil) then
		local tmp = C4:GetBindingAddress(gCon._BindingID)
		if (tmp ~= nil and string.len(tmp) > 0) then OnNetworkBindingChanged(gCon._BindingID, true) end
	end
end

--[[=============================================================================
    Base for a network server connection driver

    Copyright 2016 Control4 Corporation. All Rights Reserved.
===============================================================================]]

-- Set template version for this file
if (TEMPLATE_VERSION ~= nil) then
	TEMPLATE_VERSION.c4_networkserver_connection = "2016.01.08"
end

DEFAULT_POLLING_INTERVAL_SECONDS = 30

gNetworkKeepAliveInterval = DEFAULT_POLLING_INTERVAL_SECONDS

JK_NETWORK_BINDING_ID = 6001
JK_IP_ADDRESS = "192.168.0.169"
JK_PORT = 0x0C00


NetworkServerConnectionBase = inheritsFrom(DeviceConnectionBase)

function NetworkServerConnectionBase:construct()
	self.superClass():construct()

	self._Port = JK_PORT
	self._Handle = 0
end

function NetworkServerConnectionBase:Initialize(ExpectAck, DelayInterval, WaitInterval)
	print("NetworkServerConnectionBase:Initialize")
	gControlMethod = "NetworkServer"
	self:superClass():Initialize(ExpectAck, DelayInterval, WaitInterval, self)

end

function NetworkServerConnectionBase:ControlMethod()
	return "NetworkServer"
end

function NetworkServerConnectionBase:SendCommand(sCommand, ...)
	if(self._IsConnected) then
		if(self._IsOnline) then
			local command_delay = select(1, ...)
			local delay_units = select(2, ...)
			local command_name = select(3, ...)

			C4:SendToNetwork(self._BindingID, self._Port, sCommand)
			self:StartCommandTimer(command_delay, delay_units, command_name)
		else
			self:CheckNetworkConnectionStatus()
		end
	else
		LogWarn("Not connected to network. Command not sent.")
	end
end


function NetworkServerConnectionBase:SendRaw(sData)
--	LogTrace("Sending raw: %s", HexToString(sData))
	C4:ServerSend(self._Handle, sData, #sData)
end


function NetworkServerConnectionBase:ReceivedFromNetworkServer(nHandle, sData)
	self._Handle = nHandle
	self:ReceivedFromCom(sData)
end


function NetworkServerConnectionBase:StartListening()
	LogTrace("Creating Listener on Port %d", self._Port)
	C4:CreateServer(self._Port)
end


function NetworkServerConnectionBase:StopListening()
	LogTrace("Closing Listener on Port %d", self._Port)
	C4:DestroyServer()
end



-- function NetworkServerConnectionBase:CheckNetworkConnectionStatus()
	-- if (self._IsConnected and (not self._IsOnline)) then
		-- LogWarn("Network status is OFFLINE. Trying to reconnect to the device's Control port...")
		-- C4:NetDisconnect(self._BindingID, self._Port)
		-- C4:NetConnect(self._BindingID, self._Port)
	-- end
-- end

-- function NetworkServerConnectionBase.OnKeepAliveTimerExpired(Instance)
	-- Instance._LastCheckin = Instance._LastCheckin + 1

	-- if(Instance._LastCheckin > 2) then
		-- if(not Instance._IsOnline) then
			-- C4:NetDisconnect(Instance._BindingID, Instance._Port)
			-- C4:NetConnect(Instance._BindingID, Instance._Port)
		-- else
			-- C4:NetDisconnect(Instance._BindingID, Instance._Port)
			-- LogWarn("Failed to receive poll responses... Disconnecting...")
		-- end
	-- end

	-- if (SendKeepAlivePollingCommand ~= nil and type(SendKeepAlivePollingCommand) == "function") then
		-- SendKeepAlivePollingCommand()
	-- end

	-- Instance._KeepAliveTimer:StartTimer(gNetworkKeepAliveInterval)
-- end

-- function NetworkServerConnectionBase:SetOnlineStatus(IsOnline)
	-- self._IsOnline = IsOnline

	-- if(IsOnline) then
		-- self._KeepAliveTimer:StartTimer()
		-- self._LastCheckin = 0
		-- if (UpdateProperty ~= nil and type(UpdateProperty) == "function") then
			-- UpdateProperty("Connected To Network", "true")
		-- end

		-- self:SendNextCommand()
	-- else
		-- self._KeepAliveTimer:KillTimer()
		-- if (UpdateProperty ~= nil and type(UpdateProperty) == "function") then
			-- UpdateProperty("Connected To Network", "false")
		-- end
	-- end
-- end


--[[=============================================================================
    Notification Functions

    Copyright 2016 Control4 Corporation. All Rights Reserved.
===============================================================================]]

-- Set template version for this file
if (TEMPLATE_VERSION ~= nil) then
	TEMPLATE_VERSION.c4_notify = "2016.01.08"
end

--[[=============================================================================
    SendNotify(notifyText, tParams, bindingID)

    Description
    Forwards a notification to the proxy with a list of parameters

    Parameters
    notifyText(string) - The function identifier for the proxy
    tParams(table)     - Table of key value pairs that hold the the parameters
                         and their values used in the proxy function
    bindingID(int)     - The requests binding id

    Returns
    Nothing
===============================================================================]]
function SendNotify(notifyText, tParams, bindingID)
	C4:SendToProxy(bindingID, notifyText, tParams, "NOTIFY")
end

--[[=============================================================================
    SendSimpleNotify(notifyText, ...)

    Description
    Forwards a notification to the proxy with no parameters

    Parameters
    notifyText(string) - The function identifier for the proxy
    bindingID(int)     - Optional parameter containing the requests binding id,
                         if not specified then the DEFAULT_PROXY_ID is given.

    Returns
    Nothing
===============================================================================]]
function SendSimpleNotify(notifyText, ...)
	bindingID = select(1, ...) or DEFAULT_PROXY_BINDINGID
	C4:SendToProxy(bindingID, notifyText, {}, "NOTIFY")
end
--[[=============================================================================
    Function for changing properties

    Copyright 2016 Control4 Corporation. All Rights Reserved.
===============================================================================]]

-- Set template version for this file
if (TEMPLATE_VERSION ~= nil) then
	TEMPLATE_VERSION.c4_property = "2016.01.08"
end

--[[=============================================================================
    OnPropertyChanged(sProperty)

    Description
    Function called by Director when a property changes value. The value of the
    property that has changed can be found with: Properties[sName]. Note that
    OnPropertyChanged is not called when the Property has been changed by the
    driver calling the UpdateProperty command, only when the Property is changed
    by the user from the Properties Page. This function is called by Director
    when a property changes value.

    Parameters
    sProperty(string) - Name of property that has changed.

    Returns
    Nothing
===============================================================================]]
function OnPropertyChanged(sProperty)
	local propertyValue = Properties[sProperty]

	if (LOG ~= nil and type(LOG) == "table") then
		LogTrace("OnPropertyChanged(" .. sProperty .. ") changed to: " .. Properties[sProperty])
	end

	-- Remove any spaces (trim the property)
	local trimmedProperty = string.gsub(sProperty, " ", "")
	local status = true
	local err = ""

	if (ON_PROPERTY_CHANGED[sProperty] ~= nil and type(ON_PROPERTY_CHANGED[sProperty]) == "function") then
		status, err = pcall(ON_PROPERTY_CHANGED[sProperty], propertyValue)
	elseif (ON_PROPERTY_CHANGED[trimmedProperty] ~= nil and type(ON_PROPERTY_CHANGED[trimmedProperty]) == "function") then
		status, err = pcall(ON_PROPERTY_CHANGED[trimmedProperty], propertyValue)
	end

	if (not status) then
		LogError("LUA_ERROR: " .. err)
	end
end

--[[=============================================================================
    UpdateProperty(propertyName, propertyValue)
  
    Description:
    Sets the value of the given property in the driver
  
    Parameters:
    propertyName(string)  - The name of the property to change
    propertyValue(string) - The value of the property being changed
  
    Returns:
    None
===============================================================================]]
function UpdateProperty(propertyName, propertyValue)
	if (Properties[propertyName] ~= nil) then
		C4:UpdateProperty(propertyName, propertyValue)
	end
end

--[[=============================================================================
    Base for a serial connection driver

    Copyright 2016 Control4 Corporation. All Rights Reserved.
===============================================================================]]

-- Set template version for this file
if (TEMPLATE_VERSION ~= nil) then
	TEMPLATE_VERSION.c4_serial_connection = "2016.01.08"
end

SerialConnectionBase = inheritsFrom(DeviceConnectionBase)

function SerialConnectionBase:construct(BindingID)
	self.superClass():construct()
	self._BindingID = BindingID
end

function SerialConnectionBase:Initialize(ExpectAck, DelayInterval, WaitInterval)
	gControlMethod = "Serial"
	self:superClass():Initialize(ExpectAck, DelayInterval, WaitInterval, self)
end

function SerialConnectionBase:ControlMethod()
	return "Serial"
end

function SerialConnectionBase:SendCommand(sCommand, ...)
	if(self._IsConnected) then
		local command_delay = select(1, ...)
		local delay_units = select(2, ...)
		local command_name = select(3, ...)

		C4:SendToSerial(self._BindingID, sCommand)
		self:StartCommandTimer(command_delay, delay_units, command_name)
		LogInfo("SerialConnectionBase:SendCommand(sCommand: " .. sCommand)
	else
		LogWarn("Not connected to serial. Command not sent.")
	end
end

function SerialConnectionBase:SendRaw(sData)
	C4:SendToSerial(self._BindingID, sData)
end


function SerialConnectionBase:ReceivedFromSerial(idBinding, sData)
	self:ReceivedFromCom(sData)
end

--[[=============================================================================
    Base for a url connection driver

    Copyright 2016 Control4 Corporation. All Rights Reserved.
===============================================================================]]

-- Set template version for this file
if (TEMPLATE_VERSION ~= nil) then
	TEMPLATE_VERSION.c4_url_connection = "2016.01.08"
end

UrlConnectionBase = inheritsFrom(DeviceConnectionBase)

function UrlConnectionBase:construct(Url)
	self.superClass():construct()
	self._Url = Url
end

function UrlConnectionBase:Initialize(ExpectAck, DelayInterval, WaitInterval)
	gControlMethod = "URL"
	self:superClass():Initialize(ExpectAck, DelayInterval, WaitInterval, self)
	OnURLConnectionChanged()
end

function UrlConnectionBase:ControlMethod()
	return "URL"
end

function UrlConnectionBase:SetUrl(Url)
	self._Url = Url
end

function UrlConnectionBase:SendCommand(sCommand, sHeader, ignoreConnect)
	ignoreConnect = ignoreConnect or false

	local ticketId
	if(self._IsConnected or ignoreConnect) then
		if (sHeader ~= nil) then
			ticketId = C4:urlPost(self._Url, sCommand, sHeader)
		else
			ticketId = C4:urlPost(self._Url, sCommand)
		end
	else
		LogWarn("Not connected. Command not sent.")
	end
	
	return ticketId
end

function UrlConnectionBase:SendCommandUrl(sCommand, url, sHeader, ignoreConnect)
	ignoreConnect = ignoreConnect or false

	local ticketId
	if(self._IsConnected or ignoreConnect) then
		if (sHeader ~= nil) then
			ticketId = C4:urlPost(url, sCommand, sHeader)
		else
			ticketId = C4:urlPost(url, sCommand)
		end
	else
		LogWarn("Not connected. Command not sent.")
	end
	
	return ticketId
end

function UrlConnectionBase:UrlPost(sCommand, url, sHeader, ignoreConnect)
	ignoreConnect = ignoreConnect or false

	local ticketId
	if(self._IsConnected or ignoreConnect) then
		if (sHeader ~= nil) then
			ticketId = C4:urlPost(url, sCommand, sHeader)
		else
			ticketId = C4:urlPost(url, sCommand)
		end
	else
		LogWarn("Not connected. Command not sent.")
	end
	
	return ticketId
end

function UrlConnectionBase:UrlGet(url, sHeader, ignoreConnect)
	ignoreConnect = ignoreConnect or false

	local ticketId
	if(self._IsConnected or ignoreConnect) then
		if (sHeader ~= nil) then
			ticketId = C4:urlGet(url, sHeader)
		else
			ticketId = C4:urlGet(url)
		end
	else
		LogWarn("Not connected. Command not sent.")
	end
	
	return ticketId
end

function UrlConnectionBase:ReceivedAsync(ticketId, sData, responseCode, tHeaders)
	LogTrace("ReceivedAsync[" .. ticketId .. "]: Response Code: " .. responseCode .. " Length: " .. string.len(sData))
	LogTrace(tHeaders)
	local tMessage = {
		["ticketId"] = ticketId,
		["sData"] = sData,
		["responseCode"] = responseCode,
		["tHeaders"] = tHeaders
	}
	
	status, err = pcall(HandleMessage, tMessage)
	if (not status) then
		LogError("LUA_ERROR: " .. err)
	end
end

function ConnectURL()
	gIsUrlConnected = true
	SetControlMethod()
end

function DisconnectURL()
	gIsUrlConnected = false
	SetControlMethod()
end

--[[=============================================================================
    Helper functions

    Copyright 2016 Control4 Corporation. All Rights Reserved.
===============================================================================]]

-- Set template version for this file
if (TEMPLATE_VERSION ~= nil) then
	TEMPLATE_VERSION.c4_utils = "2016.01.08"
end

--[[=============================================================================
    AsciiToBCD(InString)

    Description
    Convert an ascii string to a binary coded decimal. Each decimal digit is
    stored in one byte, with the lower four bits encoding the digit in BCD form.

    Parameters
    InString(string) - Ascii string that is to be converted into bcd

    Returns
    The binary coded decimal
===============================================================================]]
function AsciiToBCD(InString)
	local WorkVal = 0
	local RetValStr = ""
	local DoingHighNybble = false
	local WorkStr = ((#InString % 2) == 0) and (InString) or ("0" .. InString)	-- make sure length is an even number

	for CharCount = 1, #WorkStr do
		local NumVal = tonumber(WorkStr:sub(CharCount, CharCount))

		WorkVal = bit.lshift(WorkVal, 4) + NumVal
		if (DoingHighNybble) then
			RetValStr = RetValStr .. string.char(WorkVal)
			WorkVal = 0
		end

		DoingHighNybble = (not DoingHighNybble)
	end

	return RetValStr
end

--[[=============================================================================
    BCDToAscii(InByte)

    Description
    Convert an BCD string to an ascii string.

    Parameters
    InByte(string) - Binary coded decimal that is to be converted into ascii

    Returns
    The ascii string
===============================================================================]]
function BCDToAscii(InByte)
	return tostring(bit.rshift(InByte, 4)) .. tostring(bit.band(InByte, 0x0F))
end

--[[=============================================================================
    MakeXMLNode(Tag, Value)

    Description
    Create an Xml element

    Parameters
    Tag(string)   - The Xml elements name
    Value(string) - The Xml elements value

    Returns
    The xml element created for the specified value
===============================================================================]]
function MakeXMLNode(Tag, Value)
	return "<" .. Tag .. ">" .. Value .. "</" .. Tag .. ">"
end

--[[=============================================================================
    MakeXMLAttrNode(Tag, Value, Attribute, AttrValue)

    Description
    Create an Xml element with an attribute

    Parameters
    Tag(string)       - The Xml elements name
    Value(string)     - The Xml elements value
    Attribute(string) - The attribute to be added to the Xml element
    AttrValue(string) - The value of the attribute to be added

    Returns
    The xml element created for the specified value
===============================================================================]]
function MakeXMLAttrNode(Tag, Value, Attribute, AttrValue)
    return "<" .. Tag .. " " .. Attribute .. "=\"" .. AttrValue .. "\">" .. Value .. "</" .. Tag .. ">"
end

--[[=============================================================================
    StringFromUnicode(UnicodeString)

    Description
    Convert a unicode string

    Parameters
    UnicodeString(string) - The unicode string to be converted to ascii

    Returns
    The ascii representation of the unicode string
===============================================================================]]
function StringFromUnicode(UnicodeString)
	local RetVal = ""

	-- extract every other byte from the unicode string
	for Index = 2, #UnicodeString, 2 do
		RetVal = RetVal .. string.sub(UnicodeString, Index, Index)
	end

	return RetVal
end

--[[=============================================================================
    StringSplit(s)

    Description
    Splits a string into multiple strings at every space character

    Parameters
    s(string) - The string that is to be split into several strings at the
                space character

    Returns
    A table of strings containing all the seperate values in the given string
===============================================================================]]
function StringSplit(s)
	local delim = " "
	local result = {}

	if s == nil or s == "" then
		return result
	end

	for match in (s..delim):gmatch("(.-)"..delim) do
		table.insert(result, match)
	end

	return result
end

--[[=============================================================================
    toboolean(s)

    Description
    Returns a boolean representation of the given string

    Parameters
    s(string) - The string to evaluate the boolean representation from

    Returns
    The value true or false based on the given string
===============================================================================]]
function toboolean(val)
	rval = false;

	if type(val) == "string" and (string.lower(val) == "true" or val == "1") then
		rval = true
	elseif type(val) == "number" and val ~= 0 then
		rval =  true
	elseif type(val) == "boolean" and val == true then
		rval = true
	end

	return rval
end

--[[=============================================================================
    Go(to, err, ...)

    Description
    Call a function with the given arguments if it exists or report the error

    Parameters
    to(string)  - The string to evaluate the boolean representation from
    err(string) - The error to report if the function does not exist
    ...         - Additional optional parameters for the function specified by
                  the "to" parameter

    Returns
    Nothing
===============================================================================]]
function Go(to, err, ...)
	if (type(to) == "function") then
		return to(...)
	else
		LogTrace(err)
	end
end

--[[=============================================================================
    IsEmpty(str)

    Description
    Identifies if the string given is nil or empty

    Parameters
    str(string) - The string to evaluate for the empty condition

    Returns
    True if the given value is empty, false otherwise
===============================================================================]]
function IsEmpty(str)
	return str == nil or str == ""
end

--[[=============================================================================
    ReverseTable(a)

    Description
    Reverse table entries (key=value, value=key)

    Parameters
    a(table) - The table to reverse

    Returns
    new reversed table
===============================================================================]]
function ReverseTable(a)
	local b = {}
	for k,v in pairs(a) do b[v] = k end
	return b
end

function tonumber_loc(str, base)
  local s = str:gsub(",", ".") -- Assume US Locale decimal separator
  local num = tonumber(s, base)
  if (num == nil) then
    s = str:gsub("%.", ",") -- Non-US Locale decimal separator
    num = tonumber(s, base)
  end
  return num
end

--[[=============================================================================
    HexToString(InString)

    Description
    Converts a string of Hex characters to a readable string of ASCII characters

    Parameters
    InString(string) - The string to be converted

    Returns
    A string showing the hex bytes of the InString
===============================================================================]]
function HexToString(InString)
	local RetVal = ""

	for Index = 1, #InString do
		RetVal = RetVal .. string.format("%02X ", InString:byte(Index))
	end
	return RetVal
end


--[[=============================================================================
    StringToHex(InString)

    Description
    Converts a string of ASCII characters to as string with the actual Hex bytes in them.
	Basically an array of hex bytes.

    Parameters
    InString(string) - The string to be converted

    Returns
    A string of hex bytes (really an array of hex values) 
===============================================================================]]
function StringToHex(InString)
	local RetVal = ""

	for HexByteString in string.gfind(InString, "%x%x") do
		RetVal = RetVal .. string.char(tonumber(HexByteString, 16))
	end
	return RetVal
end

function RecordHistory(severity, eventType, category, subcategory, description)
	C4:RecordHistory(severity, eventType, category, subcategory, description)
end

function RecordCriticalHistory(eventType, category, subcategory, description)
	RecordHistory("Critical", eventType, category, subcategory, description)
end

function RecordWarningHistory(eventType, category, subcategory, description)
	RecordHistory("Warning", eventType, category, subcategory, description)
end

function RecordInfoHistory(eventType, category, subcategory, description)
	RecordHistory("Info", eventType, category, subcategory, description)
end



--[[=============================================================================
    Lua Action Code

    Copyright 2015 Control4 Corporation. All Rights Reserved.
===============================================================================]]

-- This macro is utilized to identify the version string of the driver template version used.
if (TEMPLATE_VERSION ~= nil) then
	TEMPLATE_VERSION.actions = "2015.03.02"
end

-- TODO: Create a function for each action defined in the driver

function LUA_ACTION.TemplateVersion()
	TemplateVersion()
end

function LUA_ACTION.GetLampHours()
	GetLampHours()
end

function LUA_ACTION.SendConnectRequest()
	SendConnectRequest()
end

function LUA_ACTION.GetGeometryInfo()
	local command = "CORRECTMET?"
	PackAndQueueCommand("Querry Geometry Info", command)
end

--[[=============================================================================
    Functions for managing the status of the drivers bindings and connection state
 
    Copyright 2015 Control4 Corporation. All Rights Reserved.
===============================================================================]]

-- This macro is utilized to identify the version string of the driver template version used.
if (TEMPLATE_VERSION ~= nil) then
	TEMPLATE_VERSION.connections = "2015.03.31"
end

-- constants
COM_USE_ACK = true
COM_COMMAND_DELAY_MILLISECONDS = 250
COM_COMMAND_RESPONSE_TIMEOUT_SECONDS = 4

NETWORK_PORT = 3629
IR_BINDING_ID = 2
SERIAL_BINDING_ID = 1
NETWORK_BINDING_ID = 6001

--[[=============================================================================
    OnSerialConnectionChanged(idBinding, class, bIsBound)
  
    Description:
    Function called when a serial binding changes state(bound or unbound).
  
    Parameters:
    idBinding(int) - ID of the binding whose state has changed (SERIAL_BINDING_ID).
    class(string)  - Class of binding that has changed.
                     A single binding can have multiple classes(i.e. COMPONENT,
                     STEREO, RS_232, etc).
                     This indicates which has been bound or unbound.
    bIsBound(bool) - Whether the binding has been bound or unbound.
  
    Returns:
    None
===============================================================================]]
function OnSerialConnectionChanged(idBinding, class, bIsBound)
	
end

--[[=============================================================================
    OnIRConnectionChanged(idBinding, class, bIsBound)
  
    Description:
    Function called when an IR binding changes state(bound or unbound).
  
    Parameters:
    idBinding(int) - ID of the binding whose state has changed (SERIAL_BINDING_ID).
    class(string)  - Class of binding that has changed.
                     A single binding can have multiple classes(i.e. COMPONENT,
                     STEREO, RS_232, etc).
                     This indicates which has been bound or unbound.
    bIsBound(bool) - Whether the binding has been bound or unbound.
  
    Returns:
    None
===============================================================================]]
function OnIRConnectionChanged(idBinding, class, bIsBound)
	
end

--[[=============================================================================
    OnNetworkConnectionChanged(idBinding, bIsBound)
  
    Description:
    Function called when a network binding changes state(bound or unbound).
  
    Parameters:
    idBinding(int) - ID of the binding whose state has changed.
    bIsBound(bool) - Whether the binding has been bound or unbound.
  
    Returns:
    None
===============================================================================]]
function OnNetworkConnectionChanged(idBinding, bIsBound)
	
end

--[[=============================================================================
    OnNetworkStatusChanged(idBinding, nPort, sStatus)
  
    Description:
    Called when the network connection status changes. Sets the updated status of the specified binding
  
    Parameters:
    idBinding(int)  - ID of the binding whose status has changed
    nPort(int)      - The communication port of the specified bindings connection
    sStatus(string) - "ONLINE" if the connection status is to be set to Online,
                      any other value will set the status to Offline
  
    Returns:
    None
===============================================================================]]
function OnNetworkStatusChanged(idBinding, nPort, sStatus)
	LogTrace("OnConnectionStatusChanged[" .. idBinding .. " (" .. nPort .. ")]: " .. sStatus)
	if (idBinding == NETWORK_BINDING_ID) then
		if (sStatus	== "ONLINE") then
			SendConnectRequest()
		end				 
	end
end

--[[=============================================================================
    OnURLConnectionChanged(url)
  
    Description:
    Function called when the c4_url_connection is created.
  
    Parameters:
    url - url used by the url connection.
  
    Returns:
    None
===============================================================================]]
function OnURLConnectionChanged(url)
	
end

--[[=============================================================================
    DoEvents()
  
    Description:
    Called from OnSendTimeExpired in the DeviceConnectionBase.
  
    Parameters:
    None
  
    Returns:
    None
===============================================================================]]
gWasVolumeRamping = false
function DoEvents()
     LogTrace("DoEvents()")
	if (gProjectorProxy._VolumeIsRamping == true) then
	   gDoEventsTimer:StartTimer(tonumber(Properties["Volume Ramp Delay Milliseconds"]), "MILLISECONDS")
	end
	
	if (gProjectorProxy._MenuNavigationInProgress == true) then
	   gDoEventsTimer:StartTimer(tonumber(Properties["Command Delay Milliseconds"]), "MILLISECONDS")
	end		
end

function OnDoEventsTimerExpired()
     LogTrace("OnDoEventsTimerExpired()")
     gDoEventsTimer:KillTimer()
	if (gProjectorProxy._VolumeIsRamping == true) then
		if (gProjectorProxy._VolumeRamping.state == true) then gProjectorProxy:ContinueVolumeRamping() end
		gWasVolumeRamping = true
     else
	     if (gWasVolumeRamping) then
		  --note from Peter: not sure I like this approach but this is a test to clear any remaining volume commands in the queue
		  --there is an outside chance that other non-volume commands may have been queued and these will be lost. not sure of the likelihood though...
		  gCon._CommandQueue:clear()
		  gWasVolumeRamping = false
		end		
	end
	
	if (gProjectorProxy._MenuNavigationInProgress == true) then
	   SEND_COMMAND_FROM_COMMAND_TABLE(gProjectorProxy._MenuNavigationProxyID, gProjectorProxy._MenuNavigationMode)
	end	    
end

--[[=============================================================================
    SendKeepAlivePollingCommand()
  
    Description:
    Sends a driver specific polling command to the connected system
  
    Parameters:
    None
  
    Returns:
    None
===============================================================================]]
function SendKeepAlivePollingCommand()
    LogTrace("SendKeepAlivePollingCommand()")
	
	--TODO: set keep alive command for the network connected system if required.
    local command = "\r \r"
    LogTrace("command = " .. command)
    PackAndQueueCommand("SendKeepAlivePollingCommand", command)		
end

--[[=============================================================================
    Get, Handle and Dispatch message functions

    Copyright 2015 Control4 Corporation. All Rights Reserved.
===============================================================================]]

-- This macro is utilized to identify the version string of the driver template version used.
if (TEMPLATE_VERSION ~= nil) then
	TEMPLATE_VERSION.device_messages = "2015.03.31"
end

--[[=============================================================================
    GetMessage()
  
    Description:
    Used to retrieve a message from the communication buffer. Each driver is
    responsible for parsing that communication from the buffer.
  
    Parameters:
    None
  
    Returns:
    A single message from the communication buffer
===============================================================================]]
function GetMessage()
	local message, pos
    
	if (gReceiveBuffer:find(":") ~= nil) then
	   LogTrace("Handling ACK in GetMessage()")
	   gCon:HandleACK()
	   gReceiveBuffer = gReceiveBuffer:gsub(":", "") -- Remove any colons (ACK).
	end

	local pattern = "(.-)\r()"

	if (gReceiveBuffer:len() > 0) then
		message, pos = string.match(gReceiveBuffer, pattern)
		if (message == nil) then
			--LOG:Info("Do not have a complete message")
			return ""
		end
		gReceiveBuffer = gReceiveBuffer:sub(pos)		
	end

	return message
	
end

--[[=============================================================================
    HandleMessage(message)]

    Description
    This is where we parse the messages returned from the GetMessage()
    function into a command and data. The call to 'DispatchMessage' will use the
    'name' variable as a key to determine which handler routine, function, should
    be called in the DEV_MSG table. The 'value' variable will then be passed as
    a string parameter to that routine.

    Parameters
    message(string) - Message string containing the function and value to be sent to
                      DispatchMessage

    Returns
    Nothing
===============================================================================]]
function HandleMessage(message)
	
	--message = message:gsub(":", "") -- Remove any colons. Colons are left over ACKs that need to be removed
	
	if (message:find("ESC/VP") ~= nil ) then --Unit responds with ESC/VP when first connects, discarding this feedback
		message = message:sub(17) --Discarding thhe first 16 characters/handshake response 
	end	

    LogTrace(string.format("HandleMessage. Message is ==>%s<==", message))
    
	--TODO: Implement a string using Lua captures and patterns which 
	--		will be used by string.match to parse the message
	--      into a name / value pair.
	--		The example shown here will return all alpha characters 
	--		up to the first non-alpha character and store them in the "name" variable
	--		the remaining characters will be returned and stored in the "value" variable.
	
	
	local name, value = string.match(message, "(.*)=(.+)")
	
	name = name or message
	name = name:gsub(" ","")

	value = value or ""
	LogTrace("name = " .. name)
	LogTrace("value = " .. value)

	DispatchMessage(name, value)
	
end

--[[=============================================================================
    DispatchMessage(MsgKey, MsgData)

    Description
    Parse routine that will call the routines to handle the information returned
    by the connected system.

    Parameters
    MsgKey(string)  - The function to be called from within DispatchMessage
    MsgData(string) - The parameters to be passed to the function found in MsgKey

    Returns
    Nothing
===============================================================================]]
function DispatchMessage(MsgKey, MsgData)
	if (DEV_MSG[MsgKey] ~= nil and (type(DEV_MSG[MsgKey]) == "function")) then
		LogInfo("DEV_MSG." .. tostring(MsgKey) .. ":  " .. tostring(MsgData))
		local status, err = pcall(DEV_MSG[MsgKey], MsgData)
		if (not status) then
			LogError("LUA_ERROR: " .. err)
		end
	else
		LogTrace("HandleMessage: Unhanded command = " .. MsgKey)
	end
end

--[[
TODO: Create DEV_MSG functions for all messages to call Notifies.
	  Sample functions are included below for all applicable notifications.
--]]

function DEV_MSG.SOURCE(value)
	local value = "SOURCE " .. value
	LogTrace("DEV_MSG.SOURCE(), value = " .. value)
	local input = tInputConnMapByName[tInputResponseMap[value]].ID 	--mod 1000 value of Input Connection ID
	local bindingID = tInputConnMapByID[input].BindingID
	LogInfo("INPUT_CHANGED, input = " .. tInputResponseMap[value])
	gCurrentInput = gCurrentInput or -1
	if(input ~= gCurrentInput) then
		--all connections
		gProjectorProxy:dev_InputChanged(bindingID, input + 3000)        --audio

		--video only connections
		if (input > 20) then
			gProjectorProxy:dev_InputChanged(bindingID, input + 1000)        --video
		end

	end
	gCurrentInput = input
	--[[
	NOTE: Updated v2.10 (bug has existed for a long time but this is my current understanding...)
	There is TV proxy bug that affects Current Input in Composer Programming. 
	If connection has audio then it is considered audio, ONLY send audio connection notification.
	For video only connections (i.e. HDMI), BOTH audio and video connection notifications must be sent.
	--]]

end

function DEV_MSG.PWR(value)
	LogTrace("DEV_MSG.POWER(), value = " .. value)
	--[[
	TODO: "00" & "01" & "02" & "03" & "04" values will need to be edited based upon  
		the device protocol values, indicating if the device is "off", "on", 
		in a transitional state (i.e. Warming up, Cooling down) or in an error State.
		Then modify the if/then code block below to match the desired states.
	--]]
	if (value == "00") then
	   LogInfo ("Projector in standby mode with network off.  Need to enable the network stand by if using the network connection.")
		if (gProjector_Operating_Status ~= "OFF") then
			gProjector_Operating_Status = "OFF"
			C4:SetVariable('Projector_Operating_Status',"OFF")
			gProjectorProxy:dev_PowerOff()
		end

	elseif (value== "01") then -- 1 = Lamp On
		LogInfo ("Projector lamp turned on,  Return gProjector_Operating_Status ON")
		if (gProjector_Operating_Status ~= "ON") then
			gProjector_Operating_Status = "ON"
			C4:SetVariable('Projector_Operating_Status',"ON")
			gProjectorProxy:dev_PowerOn()
		end

		if (gPowerOffDelayTimer:TimerStarted()) then
		    --a power off was intiated after this power on
		    gPowerOffDelayTimer:StartTimer(1, "SECONDS")
		else
		    if gSetInputDelayTimer:TimerStarted() then
		      gSetInputDelayTimer:StartTimer(1, "SECONDS")
		    end	
		    
		    --send again in the event of err condition during power on sequence
		    gQueryVolumeLampHoursDelayTimer:StartTimer(1, "SECONDS")
		end
	elseif (value== "02") then	--2 = Warmup
		LogInfo ("Projector is warming up,  Return gProjector_Operating_Status WARMUP")
		if (gProjector_Operating_Status ~= "WARMING_UP") then
			gProjector_Operating_Status = "WARMING_UP"
			C4:SetVariable('Projector_Operating_Status',"WARMING_UP")
		end
		local command_delay = tonumber(Properties["Power Status Query Delay Seconds"])
		local delay_units = "SECONDS"
		gProjectorProxy._PowerStatusQueryDelayTimer:StartTimer(command_delay, delay_units)	
	elseif (value== "03") then	--3 = Cooldown
		LogInfo ("Projector is in cooldown,  Return Projector_Operating_Status COOLDOWN")
		if (gProjector_Operating_Status ~= "COOLING_DOWN") then
			gProjector_Operating_Status = "COOLING_DOWN"
			C4:SetVariable('Projector_Operating_Status',"COOLING_DOWN")
		end
		local command_delay = tonumber(Properties["Power Status Query Delay Seconds"])
		local delay_units = "SECONDS"
		gProjectorProxy._PowerStatusQueryDelayTimer:StartTimer(command_delay, delay_units)	
	elseif (value== "04") then	--4 = Standby with network enabled
		LogInfo ("Projector is in standby with network enabled,  Return Projector_Operating_Status OFF")
		if (gProjector_Operating_Status ~= "OFF") then
			gProjector_Operating_Status = "OFF"
			C4:SetVariable('Projector_Operating_Status',"OFF")
			gProjectorProxy:dev_PowerOff()
		end

		if (gPowerOnDelayTimer:TimerStarted()) then
		  --a power on was intiated after this power off
		  OnPowerOnDelayTimerExpired()
		end
	elseif (value == "05") then  --5 = Abnormality Standby
		LogInfo ("Projector Abnormality Standby,   Return Projector_Operating_Status ABNORMALITY_STANDBY")
		if (gProjector_Operating_Status ~= "ABNORMALITY_STANDBY") then
			gProjector_Operating_Status = "ABNORMALITY_STANDBY"
			C4:SetVariable('Projector_Operating_Status',"ABNORMALITY_STANDBY")
			gProjectorProxy:dev_PowerOff()
		end
	elseif (value == "07") then  --5 = WirelessHD Standby (Wireless Model Only)
		LogInfo ("Projector Abnormality Standby,   Return Projector_Operating_Status WIRELESSHD_STANDBY")
		if (gProjector_Operating_Status ~= "WIRELESSHD_STANDBY") then
			gProjector_Operating_Status = "WIRELESSHD_STANDBY"
			C4:SetVariable('Projector_Operating_Status',"WIRELESSHD_STANDBY")
			gProjectorProxy:dev_PowerOff()
		end
	elseif (value== "09") then	--9 = Standby with Audio enabled
		LogInfo ("Projector is in standby with Audio enabled,  Return Projector_Operating_Status OFF")
		if (gProjector_Operating_Status ~= "OFF") then
			gProjector_Operating_Status = "OFF"
			C4:SetVariable('Projector_Operating_Status',"OFF")
			gProjectorProxy:dev_PowerOff()
		end

	else
		LogWarn("DEV_MSG.POWER(): value not valid - " .. value)
	end
	C4:PersistSetValue("gProjector_Operating_Status",gProjector_Operating_Status)

end

function DEV_MSG.LAMP(value)
	LogTrace("Projector reports " .. value .. " Hours on the bulb")	
	C4:UpdateProperty('Lamp Hours', tostring(value))
	if (gLamp_Hours ~= value) then
		gLamp_Hours = value
		C4:SetVariable('Lamp_Hours',tonumber(value))
	end	
end
	
function DEV_MSG.VOL(value)
	LogTrace("DEV_MSG.VOLUME(), value = " .. value)
	
	-- TODO: derive and set  "output" from value or create separate DEV_MSG functions for each Output Connection
	local output = "0" --mod 1000 value of Output Connection ID
	
	-- TODO: edit "minDeviceLevel" & "maxDeviceLevel" values, based upon the protocol specification for volume range
	local deviceLevel = value
	local minDeviceLevel = 0
	local maxDeviceLevel = 243
	--C4 volume level uses a percentage scale: 0 - 100
	local c4Level = ConvertVolumeToC4(deviceLevel, minDeviceLevel, maxDeviceLevel)		
	
	gProjectorProxy:dev_VolumeLevelChanged(output, c4Level, deviceLevel)		
end

function DEV_MSG.MUTE(value)
	LogTrace("DEV_MSG.MUTE(), value = " .. value)
	
	-- TODO: derive and set  "output" from value or create separate DEV_MSG functions for each Output Connection
	local output = "0"    --mod 1000 value of Output Connection ID
	
	local state   	--Mute state represented as "True" or "False"
	-- TODO: values "01" & "00" will need to be modified based upon the device protocol specification
	if (value == "ON") then 
		state = "True"
		gMuteState = "ON"
	elseif (value == "OFF") then
		state = "False"
		gMuteState = "OFF"
	else
		LogWarn("DEV_MSG.MUTE(), value not valid, exiting...")
		return
	end
	gProjectorProxy:dev_MuteChanged(output, state)
end		

function DEV_MSG.ERR(value)
	if (value == "") then 
		LogInfo ("Misc ERR received")
		--likely occuring during a Power On sequence after a recent Power Off, so sending the power status query again...
		--local command_delay = tonumber(Properties["Power Off Delay Seconds"])
	     --gPowerQueryDelayTimer:StartTimer(command_delay, "SECONDS")
	elseif (value == "00") then
		LogInfo ("There is no error or the error is recovered")
	elseif (value == "01") then
		LogInfo ("Fan error")
	elseif (value == "03") then
		LogInfo ("Lamp failure at power on")
	elseif (value == "04") then
		LogInfo ("High internal temperature error")
	elseif (value == "06") then
		LogInfo ("Lamp error")	
	elseif (value == "07") then
		LogInfo ("Open Lamp cover door error")	
	elseif (value == "08") then
		LogInfo ("Cinema filter error")	
	elseif (value == "09") then
		LogInfo ("Electric dual-layered capacitor is disconnected")	
	elseif (value == "0A") then
		LogInfo ("Auto iris error")	
	elseif (value == "0B") then
		LogInfo ("Subsystem Error")			
	elseif (value == "0C") then
		LogInfo ("Low air flow error")
	elseif (value == "0D") then
		LogInfo ("Air filter air flow sensor error")
	elseif (value == "0E") then
		LogInfo ("Power supply unit error (Ballast)")
	elseif (value == "0F") then
		LogInfo ("Shutter error")
	elseif (value == "10") then
		LogInfo ("Cooling system error (Peltier element)")
	elseif (value == "11") then
		LogInfo ("Cooling system error (Pump)")
	elseif (value == "12") then
		LogInfo ("Static iris error")
	elseif (value == "13") then
		LogInfo ("Power supply unit error (Disagreement of Ballast)")
	elseif (value == "14") then
		LogInfo ("Exhaust shutter error")
	elseif (value == "15") then
		LogInfo ("Obstacle detection error")
	elseif (value == "16") then
		LogInfo ("IF board discernment error")
	else
		LogInfo ("Unknown error: " .. value)
	end
end

function DEV_MSG.IMEVENT(value)
    local pattern = "(%d+)%s(%w+)%s(%d+)%s(%d+)%s"
    local event, p1, p2, p3 = string.match(value, pattern)
    local status = getProjectorStatus(p1)
    local warning = getWarningType(p2)
    local event_msg = status .. ", " .. warning
    if (p1 == "FF") then
	   local alarm = getAlarmType(p3)
	   event_msg = event_msg .. ", " .. alarm
    end
    
    LogTrace("DEV_MSG.IMEVENT: " .. event_msg)
	
end	

function getProjectorStatus(p1)
    local msg
    if (p1 == "01") then
	   msg = "Standby"
    elseif (p1 == "02") then
	   msg = "Warmup"
    elseif (p1 == "03") then
	   msg = "Normal"
    elseif (p1 == "04") then
	   msg = "Cool down"
    elseif (p1 == "FF") then
	   msg = "Abnormal"
    else
	   msg = "Unhandled projector status: " .. p1
    end
    return msg
end

function getWarningType(p2)
    local n = tonumber(p2)
    local msg
    if (bit.band(n, 1) > 0) then
	   msg = "Lamp life"
    elseif (bit.band(n, 2) > 0) then
	   msg = "No signal"
    elseif (bit.band(n, 4) > 0) then
	   msg = "Unsupported signal"
    elseif (bit.band(n, 8) > 0) then
	   msg = "Air filter"
    elseif (bit.band(n, 16) > 0) then
	   msg = "High temperature"
    else
	   msg = "No warning"
    end
    return msg
end

function getAlarmType(p3)
    local n = tonumber(p3)
    local msg
    if (bit.band(n, 1) > 0) then
	   msg = "Lamp ON failure"
    elseif (bit.band(n, 2) > 0) then
	   msg = "Lamp lid"
    elseif (bit.band(n, 4) > 0) then
	   msg = "Lamp burnout"
    elseif (bit.band(n, 8) > 0) then
	   msg = "Fan"
    elseif (bit.band(n, 16) > 0) then
	   msg = "Temperature sensor"
	   
    elseif (bit.band(n, 32) > 0) then
	   msg = "High temperature"
    elseif (bit.band(n, 64) > 0) then
	   msg = "Interior (system)"
    else
	   msg = "No warning"
    end
    return msg
end

function DEV_MSG.BASS(value)
	LogTrace("DEV_MSG.BASS(), value = " .. value)
	
	-- TODO: derive and set  "output" from value or create separate DEV_MSG functions for each Output Connection
	local output  --mod 1000 value of Output Connection ID
	
	-- TODO: set "level", Bass level is represented as a percentage value
	local level  
	
	gProjectorProxy:dev_BassLevelChanged(output, level)
end	

function DEV_MSG.TREBLE(value)
	LogTrace("DEV_MSG.TREBLE(), value = " .. value)
	
	-- TODO: derive and set  "output" from value or create separate DEV_MSG functions for each Output Connection
	local output  --mod 1000 value of Output Connection ID
	
	-- TODO: set "level", Treble level is represented as a percentage value
	local level   
	
	gProjectorProxy:dev_TrebleLevelChanged(output, level)
end	

function DEV_MSG.BALANCE(value)
	LogTrace("DEV_MSG.BALANCE(), value = " .. value)
	
	-- TODO: derive and set  "output" from value or create separate DEV_MSG functions for each Output Connection
	local output  --mod 1000 value of Output Connection ID
	
	-- TODO: set "level", Bass level is represented as a percentage value
	local level  
	
	gProjectorProxy:dev_BalanceLevelChanged(output, level)
end	

function DEV_MSG.LOUDNESS(value)
	LogTrace("DEV_MSG.LOUDNESS(), value = " .. value)
	
	-- TODO: derive and set  "output" from value or create separate DEV_MSG functions for each Output Connection
	local output  --mod 1000 value of Output Connection ID
	
	-- TODO: set "state", Loudness state is represented as "True" or "False" (literal string, not boolean)
	local state  
	gProjectorProxy:dev_LoudnessChanged(output, state)
end
function DEV_MSG.UR(value)
	print("-----------------------------------------------\n\nTOP RIGHT (X Y) coordinate:		("..value..")\n\n-----------------------------------------------")
end

function DEV_MSG.UL(value)
	print("-----------------------------------------------\n\nTOP LEFT (X Y) coordinate:		("..value..")\n\n-----------------------------------------------")
end

function DEV_MSG.LR(value)
	print("-----------------------------------------------\n\nBOTTOM RIGHT (X Y) coordinate:	("..value..")\n\n-----------------------------------------------")
end

function DEV_MSG.LL(value)
	print("-----------------------------------------------\n\nBOTTOM LEFT (X Y) coordinate:	("..value..")\n\n-----------------------------------------------")
end

function DEV_MSG.CORRECTMET(value)
	if(value == "01") then
		CURRENT_GEOMETRY = "H/V KEYSTONE"
		local command = "VKEYSTONE?"
		PackAndQueueCommand("Query Vertical Keystone", command)
		command = "HKEYSTONE?"
		PackAndQueueCommand("Query Horizontal Keystone", command)
	end

	if(value == "02") then
		CURRENT_GEOMETRY = "QUICK CORNER"
		local command = "QC?"
		PackAndQueueCommand("Query Quick Corner Coordinates", command)
	end

	if(value == "03") then
		CURRENT_GEOMETRY = "POINT"--there are no commands for setting and querying point geometry
	end
	if(value == "06") then
		CURRENT_GEOMETRY = "CURVE"--there are no commands for setting and querying curve geometry
	end

	print("-------------CURRENT GEOMETRY: "..CURRENT_GEOMETRY)
end

function DEV_MSG.VKEYSTONE(value)
	print("-----------------------------------------------\n\nV KEYSTONE:	"..value.."\n\n-----------------------------------------------")
end
function DEV_MSG.HKEYSTONE(value)
	print("-----------------------------------------------\n\nH KEYSTONE:	"..value.."\n\n-----------------------------------------------")
end
--[[=============================================================================
    Copyright 2015 Control4 Corporation. All Rights Reserved.
===============================================================================]]

-- This macro is utilized to identify the version string of the driver template version used.
if (TEMPLATE_VERSION ~= nil) then
	TEMPLATE_VERSION.device_specific_commands = "2015.03.31"
end

--[[=============================================================================
    ExecuteCommand Code

    Define any functions for device specific commands (EX_CMD.<command>)
    received from ExecuteCommand that need to be handled by the driver.
===============================================================================]]
--function EX_CMD.NEW_COMMAND(tParams)
--	LogTrace("EX_CMD.NEW_COMMAND")
--	LogTrace(tParams)
--end

function EX_CMD.TeleWideZoom(tParams)
	local Level = tParams["Level"]
	local command = "ZOOM " .. Level
	LogTrace("command = " .. command)
	PackAndQueueCommand("TeleWideZoom", command)
	
	PackAndQueueCommand("Zoom Query", "ZOOM?")
end

function EX_CMD.Overscan(tParams)
	local Mode = tParams["Mode"]
	local command =(GAMMA_CMDS[Mode])
	LogTrace("command = " .. command)
	PackAndQueueCommand("Overscan", command)
end

function EX_CMD.Aspect(tParams)
	local Mode = tParams["Mode"]
	local command =(CMDS[Mode])
	LogTrace("command = " .. command)
	PackAndQueueCommand("Aspect", command)
end

function EX_CMD.ColorMode(tParams)
	local Mode = tParams["Mode"]
	local command =(Device_CMDS[Mode])
	LogTrace("command = " .. command)
	PackAndQueueCommand("ColorMode", command)
end

function EX_CMD.NoiseReduction(tParams)
	local Level = tParams["Level"]
	local command = "NRS " .. Level
	LogTrace("command = " .. command)
	PackAndQueueCommand("NoiseReduction", command)
	
	PackAndQueueCommand("Noise Reduction Query", "NRS?")
end

function EX_CMD.MPEGNoiseReduction(tParams)
	local Level = tParams["Level"]
	local command = "MPEGNRS " .. Level

	LogTrace("command = " .. command)
	PackAndQueueCommand("MPEGNoiseReduction", command)
	
	PackAndQueueCommand("MPEG Noise Reduction Query", "MPEGNRS?")
end

function EX_CMD.MosquitoNoise(tParams)
	local Mode = tParams["Mode"]
	local command =(Device_CMDS[Mode])
	LogTrace("command = " .. command)
	PackAndQueueCommand("MPEGNoiseReduction", command)
end

function EX_CMD.Gamma(tParams)
	local Mode = tParams["Mode"]
	local command =(GAMMA_CMDS[Mode])
	LogTrace("command = " .. command)
	PackAndQueueCommand("GAMMA", command)
end

function EX_CMD.FrameInterpolation(tParams)
	local Mode = tParams["Mode"]
	local command =(FI_CMDS[Mode])
	LogTrace("command = " .. command)
	PackAndQueueCommand("FrameInterpolation", command)
end

function EX_CMD.CallMemory(tParams)
	local Mode = tParams["Mode"]
	local command =(POPMEM_CMDS[Mode])
	LogTrace("command = " .. command)
	PackAndQueueCommand("CallMemory", command)
end

function EX_CMD.CallLensPositionMemory(tParams)
	local Mode = tParams["Mode"]
	local command =(POPLP_CMDS[Mode])
	LogTrace("command = " .. command)
	PackAndQueueCommand("CallLensPositionMemory", command)
end

function EX_CMD.SplitScreen(tParams)
	local Mode = tParams["Mode"]
	local command =(SplitScreen_CMDS[Mode])
	LogTrace("command = " .. command)
	PackAndQueueCommand("SplitScreen", command)
end

function EX_CMD.ClosedCaptionUSModelOnly(tParams)
	local Mode = tParams["Mode"]
	local command =(ClosedCaption_CMDS[Mode])
	LogTrace("command = " .. command)
	PackAndQueueCommand("ClosedCaption", command)
end

function EX_CMD.AVMute(tParams)
	local Mode = tParams["Mode"]
	local command =(AV_CMDS[Mode])
	LogTrace("command = " .. command)
	PackAndQueueCommand("AVMute", command)
end

function EX_CMD.FreezePicture(tParams)
	local Mode = tParams["Mode"]
	local command =(Freeze_CMDS[Mode])
	LogTrace("command = " .. command)
	PackAndQueueCommand("FreezePicture", command)
end

function EX_CMD.EdgeBlending(tParams)
	local Mode = tParams["Mode"]
	local command =(EdgeBlending_CMDS[Mode])
	LogTrace("command = " .. command)
	PackAndQueueCommand("EdgeBlending", command)
end

function EX_CMD.Luminance (tParams)
	local Mode = tParams["Mode"]
	local command =(LUM_CMDS[Mode])
	LogTrace("command = " .. command)
	PackAndQueueCommand("Luminance", command)
end

function EX_CMD.LuminanceLevel(tParams)
	local Mode = tParams["Level"]
	local command =(LUM_CMDS[Mode])
	LogTrace("command = " .. command)
	PackAndQueueCommand("LuminanceLevel", command)
end

function EX_CMD.ImageEnhancementPreset(tParams)
	local Mode = tParams["Mode"]
	local command =(IMGPRESET_CMDS[Mode])
	LogTrace("command = " .. command)
	PackAndQueueCommand("ImageEnhancementPreset", command)
end

function EX_CMD.Enhance4K(tParams)
	local Mode = tParams["Mode"]
	local command =(Enhance4K_CMDS[Mode])
	LogTrace("command = " .. command)
	PackAndQueueCommand("Enhance4K", command)
end

function EX_CMD.HDRDynamicRange(tParams)
	local Mode = tParams["Mode"]
	local command =(HDRDynamicRange_CMDS[Mode])
	LogTrace("command = " .. command)
	PackAndQueueCommand("HDRDynamicRange", command)
	PackAndQueueCommand("HDR Dynamic Range Query", "DYNRANGE?")
end

function EX_CMD.HDRColorSpace(tParams)
	local Mode = tParams["Mode"]
	command =(HDRColorSpace_CMDS[Mode])
	LogTrace("command = " .. command)
	PackAndQueueCommand("HDRColorSpace", command)
end

function EX_CMD.WirelessHD(tParams)
	local input = tInputConnMapByName["WirelessHD"].ID
	SET_INPUT(5001,input)
end

function EX_CMD.ScreenMirroring(tParams)
	local input = tInputConnMapByName["Screen Mirroring"].ID
	SET_INPUT(5001,input)
end
function EX_CMD.HDR10Setting(tParams)
	local Level = tParams["Level"]
	local command =("HDRPQ "..Level)
	LogTrace("command = " .. command)
	PackAndQueueCommand("HDR 10 Setting", command)
	PackAndQueueCommand("HDR 10 Setting Query", "HDRPQ?")
end
function EX_CMD.HDRHLGSetting(tParams)
	local Level = tParams["Level"]
	local command =("HDRHLG "..Level)
	LogTrace("command = " .. command)
	PackAndQueueCommand("HDR HLG Setting", command)
	PackAndQueueCommand("HDR HLG Setting Query", "HDRHLG?")
end

function EX_CMD.AutoIRISSetting(tParams)
	local Mode = tParams["Mode"]
	local command =(AutoIRIS_CMDS[Mode])
	LogTrace("command = " .. command)
	PackAndQueueCommand("Auto IRIS Setting", command)
	PackAndQueueCommand("Auto IRIS Setting Query", "IRIS?")
end

function EX_CMD.LensIRISSetting(tParams)
	local Level = tParams["Level"]
	local command =("LIRIS "..Level)
	LogTrace("command = " .. command)
	PackAndQueueCommand("Lens IRIS Setting", command)
	PackAndQueueCommand("Lens IRIS Setting Query", "LIRIS?")
end
function EX_CMD.Brightness(tParams)
	local Level = tonumber(tParams["Level"])
	local command ="BRIGHT "..Level
	LogTrace("command = " .. command)
	PackAndQueueCommand("Brightness", command)
	PackAndQueueCommand("Brightness Query", "BRIGHT?")
end


function EX_CMD.Contrast(tParams)
	local Level = tonumber(tParams["Level"])
	local command ="CONTRAST "..Level
	LogTrace("command = " .. command)
	PackAndQueueCommand("Contrast", command)
	PackAndQueueCommand("Contrast Query", "CONTRAST?")
end

function EX_CMD.ColorDensity(tParams)
	local Level = tonumber(tParams["Level"])
	local command ="DENSITY "..Level
	LogTrace("command = " .. command)
	PackAndQueueCommand("Density", command)
	PackAndQueueCommand("Density Query", "DENSITY?")
end


function EX_CMD.Tint(tParams)
	local Level = tonumber(tParams["Level"])
	local command ="TINT "..Level
	LogTrace("command = " .. command)
	PackAndQueueCommand("Tint", command)
	PackAndQueueCommand("Tint Query", "TINT?")
end

function EX_CMD.Sharpness(tParams)
	local Level = tonumber(tParams["Level"])
	local command ="SHARP "..Level
	LogTrace("command = " .. command)
	PackAndQueueCommand("Sharpness", command)
	PackAndQueueCommand("Sharpness Query", "SHARP?")
end

function EX_CMD.ColorTemperature(tParams)
	local Level = tonumber(tParams["Level"])
	local command ="CTEMP "..Level
	LogTrace("command = " .. command)
	PackAndQueueCommand("Color Temperature", command)
	PackAndQueueCommand("Color Temperature Query", "CTEMP?")
end



function EX_CMD.HorizontalKeystoneCorrection(tParams)
	local Level = tonumber(tParams["Level"])
	local command ="HKEYSTONE "..Level
	LogTrace("command = " .. command)
	PackAndQueueCommand("Horizontal Keystone Correction", command)
	PackAndQueueCommand("Horizontal Keystone Correction Query", "HKEYSTONE?")
end

function EX_CMD.VerticalKeystoneCorrection(tParams)
	local Level = tonumber(tParams["Level"])
	local command ="VKEYSTONE "..Level
	LogTrace("command = " .. command)
	PackAndQueueCommand("Vertical Keystone Correction", command)
	PackAndQueueCommand("Vertical Keystone Correction Query", "VKEYSTONE?")
end


function EX_CMD.ResetVerticalKeystoneCorrection(tParams)
	local command ="AUTOKEYSTONE OFF"
	LogTrace("command = " .. command)
	PackAndQueueCommand("Turning off Auto Vertical Keystone Correction", command)
	command ="AUTOKEYSTONE ON"
	PackAndQueueCommand("Menu",'KEY 3C')--for some reason, projector did't always react, with MENU command first, it reacts
	PackAndQueueCommand("Turning on Auto Vertical Keystone Correction", command)
	PackAndQueueCommand("Vertical Keystone Correction Query", "VKEYSTONE?")
end

function EX_CMD.ResetHorizontalKeystoneCorrection(tParams)
	local command ="SLIDEKEYSTONE 00"
	LogTrace("command = " .. command)
	PackAndQueueCommand("Turning off Auto Horizontal Keystone Correction", command)
	command ="SLIDEKEYSTONE 01"
	PackAndQueueCommand("Menu",'KEY 3C')--for some reason, projector did't always react, with MENU command first, it reacts
	PackAndQueueCommand("Turning on Auto Horizontal Keystone Correction", command)
	PackAndQueueCommand("Horizontal Keystone Correction Query", "HKEYSTONE?")
end

function EX_CMD.HorizontalImageReverse(tParams)
	local Mode = tostring(tParams["Mode"])
	local command ="HREVERSE ".. Mode
	LogTrace("command = " .. command)
	PackAndQueueCommand("Horizontal Image Reverse", command)
end


function EX_CMD.VerticalImageReverse(tParams)
	local Mode = tostring(tParams["Mode"])
	command ="VREVERSE ".. Mode
	LogTrace("command = " .. command)
	PackAndQueueCommand("Vertical Image Reverse", command)
end

function EX_CMD.QuickCornerCorrection(tParams)
	local TLX = tostring(tParams['Left Top X coord.'])
	local TLY = tostring(tParams['Left Top Y coord.'])

	local TRX = tostring(tParams['Right Top X coord.'])
	local TRY = tostring(tParams['Right Top Y coord.'])

	local BRX = tostring(tParams['Right Bottom X coord.'])
	local BRY = tostring(tParams['Right Bottom Y coord.'])

	local BLX = tostring(tParams['Left Bottom X coord.'])
	local BLY = tostring(tParams['Left Bottom Y coord.'])

	local command = string.format("QC %s %s %s %s %s %s %s %s", TLX, TLY, TRX, TRY, BRX, BRY, BLX, BLY)
	PackAndQueueCommand("Quick Corner Correction", command)
end

function EX_CMD.ResetQuickCornerCorrection(tParams)
	local TLX = tostring(0)
	local TLY = tostring(0)

	local TRX = tostring(1279)
	local TRY = tostring(0)

	local BRX = tostring(1279)
	local BRY = tostring(799)

	local BLX = tostring(0)
	local BLY = tostring(799)

	local command = string.format("QC %s %s %s %s %s %s %s %s", TLX, TLY, TRX, TRY, BRX, BRY, BLX, BLY)
	PackAndQueueCommand("Quick Corner Correction Reset", command)
end

function EX_CMD.SaveGeometryMemory(tParams)
	local memoryNumber = tParams["Mode"]

	local command = 'PUSHGC '.. memoryNumber
	PackAndQueueCommand("Save Geometry Memory", command)
end

function EX_CMD.CallGeometryMemory(tParams)
	local memoryNumber = tParams["Mode"]

	local command = 'POPGC '.. memoryNumber
	PackAndQueueCommand("Call Geometry Memory", command)
end

function EX_CMD.DynamicContrast(tParams)
	local mode = tParams["Mode"]

	local command =  DYNAMIC_CONTRAST_CMDS[mode]
	PackAndQueueCommand("Dynamic Contrast Set", command)

	local command =  "DYNCONT?"
	PackAndQueueCommand("Dynamic Contrast query", command)
end
function EX_CMD.AVOut(tParams)
	local mode = tParams["Mode"]
	local command = AV_OUT_CMDS[mode]
	PackAndQueueCommand("AV Out set",command)

	local command = "AVOUT?"
	PackAndQueueCommand("AV Out query",command)
end
function EX_CMD.BackgroundColor(tParams)
	local color = tParams["Color"]
	local command = BACKGROUND_COLOR_CMDS[color]
	PackAndQueueCommand("Background Color change",command)
end
--[[=============================================================================
    Properties Code

    Copyright 2015 Control4 Corporation. All Rights Reserved.
===============================================================================]]

-- This macro is utilized to identify the version string of the driver template version used.
if (TEMPLATE_VERSION ~= nil) then
	TEMPLATE_VERSION.properties = "2015.03.31"
end

function ON_PROPERTY_CHANGED.NetworkKeepAliveIntervalSeconds(propertyValue)
	gNetworkKeepAliveInterval = tonumber(propertyValue)

end

--[[=============================================================================
    UpdateProperty(propertyName, propertyValue)
  
    Description:
    Sets the value of the given property in the driver
  
    Parameters:
    propertyName(string)  - The name of the property to change
    propertyValue(string) - The value of the property being changed
  
    Returns:
    None
===============================================================================]]
function UpdateProperty(propertyName, propertyValue)

	if (Properties[propertyName] ~= nil) then
		C4:UpdateProperty(propertyName, propertyValue)
	end
end


--[[=============================================================================
    Commands received from the Projector proxy (ReceivedFromProxy)

    Copyright 2015 Control4 Corporation. All Rights Reserved.
===============================================================================]]

-- This macro is utilized to identify the version string of the driver template version used.
if (TEMPLATE_VERSION ~= nil) then
	TEMPLATE_VERSION.device_messages = "2015.03.31"
end

--=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
-- Power Functions
--=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
--[[
	Proxy Command: ON
	Parameters:
--]]

function ON()
    local command_delay = tonumber(Properties["Power On Delay Seconds"])
    local delay_units = "SECONDS"
    local command
    
    gPowerOnTimer:StartTimer()
    
    if (gControlMethod == "IR") then
	   command = CMDS_IR["ON"]
	   LogTrace("command = " .. command)
	   PackAndQueueCommand("ON", command, command_delay, delay_units)
    else							    
	   if (gPowerOffTimer:TimerStarted()) then
		  gPowerOnDelayTimer:StartTimer(120, "SECONDS")	--tonumber(Properties["Power Off Delay Seconds"])
	   else
		  command = CMDS["ON"]
		  LogTrace("command = " .. command)
		  PackAndQueueCommand("ON", command, command_delay, delay_units)
		
		  gPowerQueryDelayTimer:StartTimer(command_delay, delay_units)
		  gQueryVolumeLampHoursDelayTimer:StartTimer(command_delay, delay_units)
	   end
    end
end

function OnPowerQueryDelayTimerExpired()
	LogTrace("Power Query Delay Timer expired")
	gPowerQueryDelayTimer:KillTimer()
	
	GetDevicePowerStatus()
end

function OnPowerOnDelayTimerExpired()
    LogTrace("Power On Delay Timer expired")
    gPowerOnDelayTimer:KillTimer()

    local command_delay = tonumber(Properties["Power On Delay Seconds"])
    local delay_units = "SECONDS"
    local command = CMDS["ON"]
    PackAndQueueCommand("ON", command, command_delay, delay_units)
    gPowerQueryDelayTimer:StartTimer(command_delay, delay_units)
    gQueryVolumeLampHoursDelayTimer:StartTimer(command_delay, delay_units)

end

function OnPowerOffDelayTimerExpired()
    LogTrace("Power Off Delay Timer expired")
    gPowerOffDelayTimer:KillTimer()

    local command_delay = tonumber(Properties["Power Off Delay Seconds"])
    local delay_units = "SECONDS"
    local command = CMDS["OFF"]
    PackAndQueueCommand("OFF", command, command_delay, delay_units)	
    gPowerQueryDelayTimer:StartTimer(command_delay, delay_units)

end

function OnPowerOnTimerExpired()
    LogTrace("Power On Timer expired")
    gPowerOnTimer:KillTimer()
end

function OnPowerOffTimerExpired()
    LogTrace("Power Off Timer expired")
    gPowerOffTimer:KillTimer()
end

--[[
	Proxy Command: OFF
	Parameters:
--]]
function OFF()
    local command_delay = tonumber(Properties["Power Off Delay Seconds"])
    local delay_units = "SECONDS"
    local command 
    
    gPowerOffTimer:StartTimer()
    
    if (gControlMethod == "IR") then
	   command = CMDS_IR["OFF"]
	   LogTrace("command = " .. command)
	   PackAndQueueCommand("OFF", command, command_delay, delay_units)		
    else
	   if (gPowerOnTimer:TimerStarted()) then
		  --hanlding a quick power off after a power on...needs a lot of time...
		  gPowerOffDelayTimer:StartTimer(120, "SECONDS")
	   else
		  command = CMDS["OFF"]
		  LogTrace("command = " .. command)
		  PackAndQueueCommand("OFF", command, command_delay, delay_units)	

		  gPowerQueryDelayTimer:StartTimer(command_delay, delay_units)
	   end
    end  

end

--=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
-- Input Selection Functions
--=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
--[[
	Proxy Command: SET_INPUT
	Parameters:
		idBinding: proxybindingid of proxy bound to input connection
		output: mod 1000 value of Output Connection id	
		input: mod 1000 value of Input Connection id
--]]
function SET_INPUT(idBinding, input)
	local command
	if (gControlMethod == "IR") then		
		command = tInputCommandMap_IR[tInputConnMapByID[input].Name]
		PackAndQueueCommand("SET_INPUT", command)
	else
		if(input >=(MINIAPP_START_BINDING%1000) and input <= (MINIAPP_END_BINDING%1000)) then
			local proxyDeviceId, _ = next (C4:GetBoundConsumerDevices (C4:GetDeviceID (), 5001))

			local appProxyId = C4:GetBoundProviderDevice (proxyDeviceId, input+3000)
			if(appProxyId == 0) then return LogTrace("That mini app connection is not connected to any mini app.") end
			local appDeviceId = C4:GetBoundProviderDevice (appProxyId, 5001)
			-- get the details for the app for this kind of universal-minidriver-compatible type
			local miniAppVars = C4:GetDeviceVariables (appDeviceId)
			local var
			for k, var in pairs (miniAppVars) do

				if (var.name == "UM_EPSON") then
					command = tInputCommandMap[var.value]
				end
			end
			if(not command) then return LogTrace("That mini app is not supported.") end
		end

		command = command or tInputCommandMap[tInputConnMapByID[input].Name] 
		local delay, delay_units
		if (gProjectorProxy._PowerState == "POWER_ON_SEQUENCE") then
			delay = 120
			delay_units = "SECONDS"
		else
			delay = tonumber(Properties["Command Delay Milliseconds"])
			delay_units = "MILLISECONDS"
		end
		gSetInputDelayTimer:StartTimer(delay, delay_units)
		gSetInputDelayTimer._CallbackParam = command
	end

end

function OnSetInputDelayTimerExpired(command)
    LogTrace("OnSetInputDelayTimerExpired, command = " .. command)
    gSetInputDelayTimer:KillTimer()
    PackAndQueueCommand("SET_INPUT", command, 4, "SECONDS")
    PackAndQueueCommand("GetDeviceSource: Input", CMDS["GET_SOURCE"])    
    
end

--=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
-- Volume Control Functions
--=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
--[[
	Proxy Command: MUTE_OFF
	Parameters:
		output: mod 1000 value of Output Connection id	
--]]
function MUTE_OFF()
	local command
	if (gControlMethod == "IR") then		
		-- TODO: create packet/command to send to the device
		command = CMDS_IR["MUTE_OFF"]	
	else
		-- TODO: create packet/command to send to the device
		command = CMDS["MUTE_OFF"]
	end 	
	LogTrace("command = " .. command)
	PackAndQueueCommand("MUTE_OFF", command)
	PackAndQueueCommand("MUTE_QUERY", "MUTE?")
end

--[[
	Proxy Command: MUTE_ON
	Parameters:
		output: mod 1000 value of Output Connection id	
--]]
function MUTE_ON()
	local command
	if (gControlMethod == "IR") then			
		command = CMDS_IR["MUTE_ON"]	
	else
		-- TODO: create packet/command to send to the device
		command = CMDS["MUTE_ON"]
	end 	
	LogTrace("command = " .. command)
	PackAndQueueCommand("MUTE_ON", command)	
	PackAndQueueCommand("MUTE_QUERY", "MUTE?")
end

--[[
	Proxy Command: MUTE_TOGGLE
	Parameters:
		output: mod 1000 value of Output Connection id	
--]]
function MUTE_TOGGLE()
	local command
	if (gControlMethod == "IR") then			
		command = CMDS_IR["MUTE_TOGGLE"]	
	else
		-- TODO: create packet/command to send to the device
		command = CMDS["MUTE_TOGGLE"]
		if (gMuteState == "OFF") then
		  command = CMDS["MUTE_ON"]
		elseif (gMuteState == "ON") then
		  command = CMDS["MUTE_OFF"]
		else
		  command = CMDS["MUTE_ON"]
		end
	end 
	LogTrace("command = " .. command)
	PackAndQueueCommand("MUTE_TOGGLE", command)
	PackAndQueueCommand("MUTE_QUERY", "MUTE?")
end

--[[
	Proxy Command: SET_VOLUME_LEVEL
	Parameters:
		output: mod 1000 value of Output Connection id	
		c4VolumeLevel: volume level to be set represented in Control4 scale (0-100)
--]]
function SET_VOLUME_LEVEL(c4VolumeLevel)
	-- TODO: edit "minDeviceLevel" & "maxDeviceLevel" values, based upon the protocol specification for volume range
	local minDeviceLevel = 0
	local maxDeviceLevel = 255
	local deviceVolumeLevel = ConvertVolumeToDevice(c4VolumeLevel, minDeviceLevel, maxDeviceLevel)
	
	---- TODO: uncomment and edit string padding, if required, based upon the protocol specification
    --deviceVolumeLevel = string.rep("0", 2 - string.len(deviceVolumeLevel)) .. deviceVolumeLevel
	
	LogInfo('deviceVolumeLevel: ' .. deviceVolumeLevel)
  
	-- TODO: create packet/command to send to the device
	local command = CMDS["SET_VOLUME_LEVEL"] .. deviceVolumeLevel
	LogTrace("command = " .. command)
	PackAndQueueCommand("SET_VOLUME_LEVEL", command)
	command = CMDS["GET_VOLUME_LEVEL"]
	PackAndQueueCommand("GET_VOLUME_LEVEL", command)
end

--[[
	Helper Function: SET_VOLUME_LEVEL_DEVICE
	Parameters:
		output: mod 1000 value of Output Connection id	
		deviceVolumeLevel: volume level to be set represented in device scale (as sepcified in the device's control protocol)
--]]
function SET_VOLUME_LEVEL_DEVICE(deviceVolumeLevel)
	--Called from ContinueVolumeRamping()
	
	-- TODO: create packet/command to send to the device
	local command = CMDS["SET_VOLUME_LEVEL"] .. deviceVolumeLevel
	
	LogTrace("command = " .. command)
	local command_delay = tonumber(Properties["Volume Ramp Delay Milliseconds"])
	PackAndQueueCommand("SET_VOLUME_LEVEL_DEVICE", command, command_delay)
	command = CMDS["GET_VOLUME_LEVEL"]
	PackAndQueueCommand("GET_VOLUME_LEVEL", command)	
end

--[[
	Proxy Command: PULSE_VOL_DOWN
	Parameters:
		output: mod 1000 value of Output Connection id	
--]]
function PULSE_VOL_DOWN()
	local command
	if (gControlMethod == "IR") then		
		-- TODO: create packet/command to send to the device
		command = CMDS_IR["PULSE_VOL_DOWN"]	
	else
		-- TODO: create packet/command to send to the device
		command = CMDS["PULSE_VOL_DOWN"]
	end 		
	--LogTrace("command = " .. command)
	local command_delay = tonumber(Properties["Volume Ramp Delay Milliseconds"])
	PackAndQueueCommand("PULSE_VOL_DOWN", command, command_delay)
	command = CMDS["GET_VOLUME_LEVEL"]
	PackAndQueueCommand("GET_VOLUME_LEVEL", command)
end

--[[
	Proxy Command: PULSE_VOL_UP
	Parameters:
		output: mod 1000 value of Output Connection id	
--]]
function PULSE_VOL_UP()
	local command
	if (gControlMethod == "IR") then	
		-- TODO: create packet/command to send to the device
		command = CMDS_IR["PULSE_VOL_UP"]	
	else
		-- TODO: create packet/command to send to the device
		command = CMDS["PULSE_VOL_UP"]
	end 		
	--LogTrace("command = " .. command)
	local command_delay = tonumber(Properties["Volume Ramp Delay Milliseconds"])
	PackAndQueueCommand("PULSE_VOL_UP", command, command_delay)
	command = CMDS["GET_VOLUME_LEVEL"]
	PackAndQueueCommand("GET_VOLUME_LEVEL", command)
end

--=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
-- Helper Functions
--=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=

--[[
	Helper Function: SEND_COMMAND_FROM_COMMAND_TABLE
	Parameters:
		idBinding: proxy id	
		output: mod 1000 value of Output Connection id
		command_name: name of command to be sent
--]]
function SEND_COMMAND_FROM_COMMAND_TABLE(idBinding, command_name)
    LogTrace("SEND_COMMAND_FROM_COMMAND_TABLE(), idBinding=" .. idBinding .. ", command_name=" .. command_name)
	
	-- TODO: create packet/command to send to the device
	local command = GetCommandFromCommandTable(idBinding, command_name)
	
	if (command == nil) then
		LogTrace("command is nil")
	else
		LogTrace("command = " .. command)
	end	
	PackAndQueueCommand(command_name, command)		
end

--[[
	Helper Function: GetCommandFromCommandTable
	Parameters:
		idBinding: proxy id	
		output: mod 1000 value of Output Connection id
		command_name: name of command to be returned
--]]
function GetCommandFromCommandTable(idBinding, command_name)
	LogTrace("GetCommand()")
	local t = {}
	
	-- TODO: select/create appropriate commands table structure
	
	if (gControlMethod == "IR") then
		t = CMDS_IR
	else
		t = CMDS
	end	

	if (t[idBinding][command_name] ~= nil) then
		return t[idBinding][command_name]
	elseif (t[command_name] ~= nil) then
		return t[command_name]
	else
		LogWarn('GetCommandFromCommandTable: command not defined - '.. command_name)
		return nil
	end	
end

function OnQueryVolumeLampHoursDelayTimerExpired()
    LogTrace("OnQueryVolumeLampHoursDelayTimerExpired")
    gQueryVolumeLampHoursDelayTimer:KillTimer()
    
    if (getCapabilityValueFromConfig("has_audio") == "True") then
	   PackAndQueueCommand("GetDeviceVolumeStatus: Volume", CMDS["GET_VOLUME_STATUS"])
	   --PackAndQueueCommand("GetDeviceVolumeStatus: Mute", CMDS["GET_MUTE_STATUS"])	
    end

    PackAndQueueCommand("GetLampHours", CMDS["GET_LAMP_HOURS"])
end

--[[
	Helper Function: GetDeviceVolumeStatus
--]]
function GetDeviceVolumeStatus()
    LogTrace("GetDeviceVolumeStatus()")
	
	-- TODO: verify table entry in CMDS table
	local command = CMDS["GET_VOLUME_STATUS"]
	--LogTrace("command = " .. command)
	PackAndQueueCommand("GetDeviceVolumeStatus: Volume", command)
	
	-- TODO: verify table entry in CMDS table
	command = CMDS["GET_MUTE_STATUS"]
	--LogTrace("command = " .. command)
	PackAndQueueCommand("GetDeviceVolumeStatus: Mute", command)	
end

--[[
	Helper Function: GetLampHours
--]]
function GetLampHours()
    LogTrace("GetLampHours()")
	
	-- TODO: verify table entry in CMDS table
	local command = CMDS["GET_LAMP_HOURS"]
	LogTrace("command = " .. command)
	PackAndQueueCommand("GetLampHours", command)
end

--[[
	Helper Function: GetDevicePowerStatus
--]]
function GetDevicePowerStatus(...)
	local bPriority = select(1, ...) or false
     LogTrace("GetDevicePowerStatus()")
	
	local command = CMDS["GET_POWER_STATUS"]
	LogTrace("command = " .. command)
	if (bPriority) then
		PackAndQueuePriority1Command("GetDevicePowerStatus", command)
	else
		PackAndQueueCommand("GetDevicePowerStatus", command)
	end
end
	
function getCapabilityValueFromConfig(sName)
	local xml = C4:GetDriverConfigInfo("capabilities")
	xml = "<capabilities>" .. xml .. "</capabilities>"
	local t = C4:ParseXml(xml)

	local value
	for j,k in pairs(t.ChildNodes) do 
		if (k.Name == sName) then return k.Value end
	end

	LogWarn("getCapabilityValueFromConfig, " .. sName .. " capability does not exist!")

end

function printCapabilityValues()
	local xml = C4:GetDriverConfigInfo("capabilities")
	xml = "<capabilities>" .. xml .. "</capabilities>"
	local t = C4:ParseXml(xml)

	for j,k in pairs(t.ChildNodes) do 
		print(k.Name, k.Value)
	end
end
--[[=============================================================================
    PROJECTOR Protocol Initialization Functions

    Copyright 2015 Control4 Corporation. All Rights Reserved.
===============================================================================]]

-- This macro is utilized to identify the version string of the driver template version used.
if (TEMPLATE_VERSION ~= nil) then
	TEMPLATE_VERSION.device_messages = "2015.03.31"
end


PROTOCOL_DECLARATIONS = {}

function ON_DRIVER_EARLY_INIT.projector_init()
	MINIAPP_START_BINDING = 3051
	MINIAPP_END_BINDING = 3055

end

function ON_DRIVER_INIT.projector_init()
	--LogTrace("ON_DRIVER_INIT.ProtocolDeclarations()")
	for k,v in pairs(PROTOCOL_DECLARATIONS) do
		if (PROTOCOL_DECLARATIONS[k] ~= nil and type(PROTOCOL_DECLARATIONS[k]) == "function") then
			PROTOCOL_DECLARATIONS[k]()
		end
	end	
end

function ON_DRIVER_LATEINIT.projector_init()
end

function PROTOCOL_DECLARATIONS.SetGlobals()
	LogTrace("PROTOCOL_DECLARATIONS.SetGlobals()")

	--[[
	TODO: Determine projector operating conditions to be exposed for Composer Programming.
		Most projectors will report status, i.e. Warm Up,Cool Down and Error, etc.
		Lamp usage may also be reported. The sample variables below are referenced in the
		DEV_MSG.POWER() function in the device_messages.lua file.
	--]]
	C4:AddVariable('Projector_Operating_Status',0,'STRING',false,false)	
	C4:AddVariable('Lamp_Hours',0,'NUMBER',false,false)
	gProjector_Operating_Status = ""
	gLamp_Hours = ""
end
	
function PROTOCOL_DECLARATIONS.CommandsTableInit_IR()
	LogTrace("PROTOCOL_DECLARATIONS.CommandsTableInit_IR()")

	CMDS_IR = {
		--index:  Proxy Command Name
		--value:  IR Code ID	 
		--Power
		["ON"]             = "",
		["OFF"]            = "",
		
		--Input Toggle
		["INPUT_TOGGLE"]="",
		
		--Menu
		["INFO"] = "",	--Display
		["GUIDE"] = "",
		["MENU"] = "",
		["CANCEL"] = "",
		["UP"] = "",
		["DOWN"] = "",
		["LEFT"] = "",
		["RIGHT"] = "",
		["ENTER"] = "",	
		["RECALL"]         = "",
		["PREV"]           = "",
		
		--Digits
		["NUMBER_0"]       = "", 	
		["NUMBER_1"]       = "",		
		["NUMBER_2"]       = "",	
		["NUMBER_3"]       = "",	
		["NUMBER_4"]       = "",	
		["NUMBER_5"]       = "",	
		["NUMBER_6"]       = "",	
		["NUMBER_7"]       = "",	
		["NUMBER_8"]       = "",	
		["NUMBER_9"]       = "",
		["STAR"]           = "",
		["DOT"]            = "",
		
		--Volume
		["MUTE_ON"]        = "",
		["MUTE_OFF"]       = "",
		["MUTE_TOGGLE"]    = "",		
		["PULSE_VOL_DOWN"] = "",
		["PULSE_VOL_UP"]   = "",		
		["SET_VOLUME_LEVEL"] = "",		
		
	}			
	
	CMDS_IR[PROJECTOR_PROXY_BINDINGID] = {}	
	
end

function PROTOCOL_DECLARATIONS.CommandsTableInit_Serial()
		LogTrace("PROTOCOL_DECLARATIONS.CommandsTableInit_Serial()")

	CMDS = {
		--index:  Proxy Command Name
		--value:  Protocol Command Data
		
		['GET_SOURCE']        = 'SOURCE?',
		['GET_LAMP_HOURS']    = 'LAMP?',
		['GET_VOLUME_STATUS'] = 'VOL?',
		
		--Power
		['ON']         = 'PWR ON',
		['OFF']        = 'PWR OFF',
		['GET_POWER_STATUS'] = "PWR?",

		--Aspect COMMANDS
		['Normal']      = 'ASPECT 00',
		['16:9']      = 'ASPECT 20',
		['16:9 Up']      = 'ASPECT 21',
		['16:9 Low']      = 'ASPECT 22',		
		['Auto']        = 'ASPECT 30',		
		['Native']       = 'ASPECT 60',
		['Full']        = 'ASPECT 40',
		['Zoom']      = 'ASPECT 50',
		['4:3']      = 'ASPECT 10',
		['Anamorphic']        = 'ASPECT 80',
		['H-Squeeze']      = 'ASPECT 90',
		["H-Zoom"]       = "ASPECT 50",
		["V-Zoom"]       = "ASPECT A0",
		["Brightness"] = "BRIGHT",
		["Contrast"] = "CONTRAST",
		["Density"] = "DENSITY",
		["Tint"] = "TINT",
		["Sharpness"] = "SHARP",
		["Color Temerature"] = "CTEMP",

		--Zoom
		['MIN']      = 'ZOOM MIN',
		['MAX']      = 'ZOOM MAX',
		['INC']        = 'ZOOM INC',		
		['DEC']       = 'ZOOM DEC',		
		['Zoom OFF']       = 'ZOOM OFF',
		
		--Volume		
		['PULSE_VOL_UP']       = 'VOL INC',
		['PULSE_VOL_DOWN']     = 'VOL DEC',
		["MUTE_ON"]       = "MUTE ON",
		["MUTE_OFF"]       = "MUTE OFF",
		["MUTE_TOGGLE"]       = "",		
		["SET_VOLUME_LEVEL"] = "VOL ",
		["GET_MUTE_STATUS"] = "",
		["GET_VOLUME_LEVEL"] = "VOL?",

		--Menu
		['GUIDE']      = 'KEY 30',
		['MENU']       = 'KEY 3C',
		['CANCEL']     = 'KEY 3D',
		['UP']         = 'KEY 58',
		['DOWN']       = 'KEY 59',
		['LEFT']       = 'KEY 5A',
		['RIGHT']      = 'KEY 5B',
		['ENTER']      = 'KEY 49',
	}
	LUM_CMDS = {		--Luminance
		['Level 1']    = 'LUMLEVEL 01',
		['Level 2']    = 'LUMLEVEL 02',
		['Level 3']    = 'LUMLEVEL 03',
		['Level 4']    = 'LUMLEVEL 04',
		['Level 5']    = 'LUMLEVEL 05',
		['Normal']    = 'LUMINANCE 00',
		['Eco']    = 'LUMINANCE 01',
		['Temperature-sensitive mode']    = 'LUMINANCE A0',
		['Auto']    = 'LUMINANCE 03',	
		['Quiet']    = 'LUMINANCE 01',
		['Long']    = 'LUMINANCE 04',
		['Custom']    = 'LUMINANCE 05',
		['High']    = 'LUMINANCE 00',
		['Medium']    = 'LUMINANCE 02',
		['Eco2']    = 'LUMINANCE 06',
	}
	Device_CMDS = {	--Projector viewing modes
	     ["Auto"]       = "CMODE 00", 
		["sRGB"]       = "CMODE 01", 	
		["Presentation"]       = "CMODE 04",		
		["Theatre"]       = "CMODE 05",	
		["Dynamic"]       = "CMODE 06",	
		["DICOM SIM"]       = "CMODE 0F",	
		["Photo"]       = "CMODE 14",	
		["Multi Projection"]       = "CMODE 1A",	
		["3D Dynamic"]       = "CMODE 18",	
		["3D Theatre"]       = "CMODE 1B",
		["3D Presentation"]       = "CMODE 1C",
		["3D Multi Projection"]       = "CMODE 1D",
		["3D THX"]       = "CMODE 90",
		["3D Cinema"]       = "CMODE 17",
		["Off"]       = "MNRS 00", -- Mos Noise Reduction OFF
		["On"]       = "MNRS 01", -- Mos Noise Reduction ON
		["White Board"]       = "CMODE 12",	
		["Black Board"]       = "CMODE 11",
		["Sports"]       = "CMODE 08",	
		["Cinema"]       = "CMODE 15",
		["Customized"]       = "CMODE 10",
		["Bright Cinema"]       = "CMODE 0C",
		["THX"]       = "CMODE 13",
		["B W Cinema"]       = "CMODE 20", 
		["Adobe RGB"]       = "CMODE 21",	
		["Digital Cinema"]       = "CMODE 22",	
		["Natural"]       = "CMODE 07",
		["Living"]       = "CMODE 0C",
		["Game"]       = "CMODE OD",
	}
	FI_CMDS = {
		["Off"]       = "MCFI 00",		
		["Low"]       = "MCFI 01",		
		["Normal"]       = "MCFI 02",		
		["High"]       = "MCFI 03",		
	}
	AV_CMDS = {
		["On"]       = "MUTE ON",	
		["Off"]       = "MUTE OFF",
	}
	Freeze_CMDS = {	
		["On"]       = "FREEZE ON",	
		["Off"]       = "FREEZE OFF",	
	}
	ClosedCaption_CMDS ={
		["Off"]       = "CCAP 00",
		["CC1"]       = "CCAP 11",
		["CC2"]       = "CCAP 12",
		["Toggle"]       = "CCAP F0",
	}
	EdgeBlending_CMDS ={
		["Off"]       = "EGBLEND 01 00",	
		["On"]       = "EGBLEND 01 01",
	}
	POPMEM_CMDS = {
		["1"]       = "POPMEM 02 01",	--Call Memory Start
		["2"]       = "POPMEM 02 02",	
		["3"]       = "POPMEM 02 03",	
		["4"]       = "POPMEM 02 04",	
		["5"]       = "POPMEM 02 05",	
		["6"]       = "POPMEM 02 06",	
		["7"]       = "POPMEM 02 07",	
		["8"]       = "POPMEM 02 08",	
		["9"]       = "POPMEM 02 09",
		["10"]       = "POPMEM 02 0A",	
	}
	SplitScreen_CMDS = {
		["Off"]       				= "SPS 01 00",	
		["On"]       				= "SPS 01 01",	
		["Both Same Size"]       	= "SPS 02 00",	
		["Left Bigger"]       		= "SPS 02 01",
		["Right Bigger"]       		= "SPS 02 02",
		["Switch Left and Right"]  	= "SPS 05",	
		["Left COMP"]       		= "SPS 03 B0",	
		["Left PC"]       			= "SPS 03 10",	
		["Left DP"]       			= "SPS 03 70",	
		["Left HDBaseT"]       		= "SPS 03 80",	
		["Left HDMI"]       		= "SPS 03 30",	
		["Left SVID"]       		= "SPS 03 42",	
		["Left SDI"]       			= "SPS 03 70",	
		["Left VID"]       			= "SPS 03 41",	
		["Right COMP"]       		= "SPS 04 B0",	
		["Right PC"]       			= "SPS 04 10",	
		["Right DP"]       			= "SPS 04 70",	
		["Right HDBaseT"]       		= "SPS 04 80",	
		["Right HDMI"]       		= "SPS 04 30",	
		["Right SVID"]       		= "SPS 04 42",	
		["Right SDI"]       		= "SPS 04 70",
		["Right VID"]      	 		= "SPS 04 41",			
	}
	DYNAMIC_CONTRAST_CMDS = {
		["Off"] = 		"DYNCONT 00",
		["Normal"] = 	"DYNCONT 01",
		["High Speed"]=	"DYNCONT 02",
	}
	AV_OUT_CMDS = {
		["On Projection"] = "AVOUT 00",
		["Always"] = "AVOUT 01",
	}
	BACKGROUND_COLOR_CMDS = {
		["Black"] = "MSEL 00",
		["Blue"] = "MSEL 01",
		["Logo"] = "MSEL 02",

	}
	GAMMA_CMDS = {
		["-2"]       = "GAMMA 20",		
		["-1"]       = "GAMMA 21",		
		["0"]       = "GAMMA 22",		
		["1"]       = "GAMMA 23",		
		["2"]       = "GAMMA 24",		
		["Custom"]       = "GAMMA F0",
		["OFF"]       = "OVSCAN 00",		
		["4 Percent"]       = "OVSCAN 02",		
		["8 Percent"]       = "OVSCAN 04",		
		["Auto"]       = "OVSCAN A0",
	}
	STATUS_CMDS = {
		['POWER']       = 'PWR?',
		['SOURCE']      = 'SOURCE?',
		['LAMP_USE']    = 'LAMP?',
		['3_DIMENSION'] = '',
		['VOLUME'] = 'VOL?',
	}
	NR_CMDS = {
		['Off']       = 'NRS 01',
		['NR1']      = 'NRS 02',
		['NR2']    = 'NRS 03',
		['NR3']      = 'NRS 04',
		['NR4']    = 'NRS 05',
	}
	MJPEG_NR_CMDS = {
		['Off']       = 'MPEGNRS 00',
		['Low']      = 'MPEGNRS 01',
		['Normal']    = 'MPEGNRS 02',
		['High']      = 'MPEGNRS 03',
	}	
	GAMUT_CMDS = {
		["Off"]       = "SUPERRES 00",	
		["1"]       = "SUPERRES 01",	
		["2"]       = "SUPERRES 02",	
		["3"]       = "SUPERRES 03",	
		["4"]       = "SUPERRES 04",	
		["5"]       = "SUPERRES 05",	
		["4K 1"]       = "SUPERRES 11",	
		["4K 2"]       = "SUPERRES 12",	
		["4K 3"]       = "SUPERRES 13",
		["4K 4"]       = "SUPERRES 14",	
		["4K 5"]       = "SUPERRES 15", 
		["HDTV"]       = "CGAMUT 01",	
		["EBU"]       = "CGAMUT 02",	
		["SMPTE C"]       = "CGAMUT 03",
	}
	CMDS_3D = {
		["3D Off"]       = "3DIMENSION 01 00",	
		["3D On"]       = "3DIMENSION 01 01",	
		["Off"]       = "3DIMENSION 02 00",	
		["Low"]       = "3DIMENSION 02 01",	
		["Middle"]       = "3DIMENSION 02 02",	
		["High"]       = "3DIMENSION 02 03",	
		["Auto"]       = "3DIMENSION 03 00",	
		["2D"]       = "3DIMENSION 03 01",	
		["Side by Side"]       = "3DIMENSION 03 02",
		["Top and Bottom"]       = "3DIMENSION 03 03",	
		["-10"]       = "3DIMENSION 04 F6", 
		["-9"]       = "3DIMENSION 04 F5",
		["-8"]       = "3DIMENSION 04 F4", 
		["-7"]       = "3DIMENSION 04 F3",
		["-6"]       = "3DIMENSION 04 F2", 
		["-5"]       = "3DIMENSION 04 F1",
		["-4"]       = "3DIMENSION 04 FA", 
		["-3"]       = "3DIMENSION 04 FB",
		["-2"]       = "3DIMENSION 04 FC", 
		["-1"]       = "3DIMENSION 04 FD",
		["0"]       = "3DIMENSION 04 00", 
		["1"]       = "3DIMENSION 04 01",
		["2"]       = "3DIMENSION 04 02", 
		["3"]       = "3DIMENSION 04 03",
		["4"]       = "3DIMENSION 04 04", 
		["5"]       = "3DIMENSION 04 05",
		["6"]       = "3DIMENSION 04 06", 
		["7"]       = "3DIMENSION 04 07",
		["8"]       = "3DIMENSION 04 08", 
		["9"]       = "3DIMENSION 04 09",
		["10"]       = "3DIMENSION 04 0A", 
		["Viewing On"]       = "3DIMENSION 07 00",
		["Viewing Off"]       = "3DIMENSION 07 01",		
	}	
	POPLP_CMDS = {
		["1"]       = "POPLP 01",  --Call Lens Position memory
		["2"]       = "POPLP 02", 
		["3"]       = "POPLP 03",
		["4"]       = "POPLP 04", 
		["5"]       = "POPLP 05",
		["6"]       = "POPLP 06", 
		["7"]       = "POPLP 07",
		["8"]       = "POPLP 08", 
		["9"]       = "POPLP 09",
		["10"]       = "POPLP 0A", 
	}
	IMGPRESET_CMDS = {
		["Off"]       = "IMGPRESET 00",  --Call Lens Position memory
		["Preset 1"]       = "IMGPRESET 01", 
		["Preset 2"]       = "IMGPRESET 02",
		["Preset 3"]       = "IMGPRESET 03", 
		["Preset 4"]       = "IMGPRESET 04",
		["Preset 5"]       = "IMGPRESET 05", 
	}
	Enhance4K_CMDS = {
		['Off']       = '4KENHANCE 00',
		['Full HD']      = '4KENHANCE 01',
		['WUXGA+']    = '4KENHANCE 02',
		['On']    = '4KENHANCE 01',
	}	
	HDRDynamicRange_CMDS = {
		['Auto']       = 'DYNRANGE 00',
		['SDR']      = 'DYNRANGE 01',
		['HDR Mode 1']    = 'DYNRANGE 02',
		['HDR Mode 2']    = 'DYNRANGE 03',
		['HDR Mode 3']    = 'DYNRANGE 04',
		['HDR Mode 4']    = 'DYNRANGE 05',
		['HDR10']    = 'DYNRANGE 20',
		['HLG']    = 'DYNRANGE 30',
	}
	HDRColorSpace_CMDS = {
		['Auto']       = 'CLRSPACE 00',
		['BT.709']      = 'CLRSPACE 01',
		['BT.2020']    = 'CLRSPACE 02',
	}
	AutoIRIS_CMDS = {
		['Off'] = "IRIS 00",
		['Normal'] = "IRIS 01",
		["High"] = "IRIS 02",
	}
	LANGUAGE = {
		["English"] = "01",
		["Japanese"] = "02",
		["French"] = "03",
		["German"] = "04",
		["Italian"] = "05",
		["Spanish"] = "06",
		["Portuguese"] = "07",
		["Simplified Chinese"] = "08",
		["Korean"] = "09",
		["Traditional Chinese"] = "16",
		["Thai"] = "1B",
		["Indonesian"]="1C",
		["Malay"] = "20",
		["Vietnamese"] = "21",
		["Hindi"] = "23",
	}
	CMDS[PROJECTOR_PROXY_BINDINGID] = {}
	
end

function PROTOCOL_DECLARATIONS.InputOutputTableInit()
	LogTrace("PROTOCOL_DECLARATIONS.InputOutputTableInit()")
	----------------------------------------- [*COMMAND/RESPONSE HELPER TABLES*] -----------------------------------------
	tInputCommandMap = {
		["BNC"] = "SOURCE B0", 
		["Computer"] = "SOURCE 10",
		["Display Port"] = "SOURCE 70",		
		["HDBaseT"] = "SOURCE 80",		
		["HDMI"] = "SOURCE 30",		
		["S-Video"] = "SOURCE 42",
		["SDI"] = "SOURCE 60",		
		["Video"] = "SOURCE 41",
		["PC"] = "SOURCE 20",
		["Computer2"] = "SOURCE 20",
		["USB Display"] = "SOURCE 51",
		["USB"] = "SOURCE 52",
		["LAN"] = "SOURCE 53",
		["USB 2"] = "SOURCE 54",
		["White Board"] = "SOURCE 55",
		["Screen Mirroring"] = "SOURCE 56",
		["Screen Mirroring 2"] = "SOURCE 59",
		["HDMI 2"] = "SOURCE A0",
		["HDMI 3"] = "SOURCE C0",
		["WirelessHD"] = "SOURCE D0",
	}

    tInputResponseMap = {
		["SOURCE B0"] = "BNC", 
		["SOURCE B1"] = "BNC",	--RGB(Analog)
		["SOURCE B4"] = "BNC",	--Component
		["SOURCE BF"] = "BNC",	--Auto
		
		["SOURCE 10"] = "Computer",
		["SOURCE 11"] = "Computer",	--RGB(Analog)
		["SOURCE 14"] = "Computer",	--Component
		["SOURCE 1F"] = "Computer",	--Auto
		
		["SOURCE 20"] = "Computer2",
		["SOURCE 21"] = "Computer2",	--RGB(Analog)
		
		["SOURCE 70"] = "Display Port",		
		
		["SOURCE 80"] = "HDBaseT",	
		["SOURCE 81"] = "HDBaseT",		--(Digital-RGB)
		["SOURCE 83"] = "HDBaseT",		--(RGB-Video)
		["SOURCE 84"] = "HDBaseT",		--(YCbCr)
		["SOURCE 85"] = "HDBaseT",		--(YPbPr)
		["SOURCE 8F"] = "HDBaseT",		--(Reply to "SOURCE?" command)
		
		
		["SOURCE 30"] = "HDMI",	
		["SOURCE 31"] = "HDMI",		--(Digital-RGB)
		["SOURCE 33"] = "HDMI",		--(RGB-Video)
		["SOURCE 34"] = "HDMI",		--(YCbCr)
		["SOURCE 35"] = "HDMI",		--(YPbPr)
		["SOURCE 3F"] = "HDMI",		--(Reply to "SOURCE?" command)
	   
		["SOURCE 42"] = "S-Video",
		
		["SOURCE 60"] = "SDI",	
		["SOURCE 64"] = "SDI",		--(YCbCr)
		["SOURCE 65"] = "SDI",		--(YPbPr)
		["SOURCE 6F"] = "SDI",		--(Reply to "SOURCE?" command)
		
		["SOURCE 41"] = "Video",
		["SOURCE 51"] = "USB Display",
		["SOURCE 52"] = "USB",
		["SOURCE 53"] = "LAN",
		["SOURCE 54"] = "USB 2",
		["SOURCE 55"] = "White Board",
		["SOURCE 56"] = "Screen Mirroring",
		
		["SOURCE A0"] = "HDMI 2",
		["SOURCE A1"] = "HDMI 2",		--(Digital-RGB)
		["SOURCE A3"] = "HDMI 2",		--(RGB-Video)
		["SOURCE AF"] = "HDMI 2",		--(Reply to "SOURCE?" command)
		
		["SOURCE C0"] = "HDMI 3",
		["SOURCE D0"] = "WirelessHD",
	}
	--tInputResponseMap = ReverseTable(tInputCommandMap)	-- Reverses the tInputCommandMap table	
	
	tInputCommandMap_IR = { 
		--index:  Connection Name
		--value:  IR Code ID
		--["ADAPTER PORT"] = "51362",
		[""] = "",	
	}

	----------------------------------------- [*I/O HELPER TABLES*] -----------------------------------------
	tOutputConnMap = {
		--index:  mod 1000 value of Output Connection id
		--value:  Output Connection Name
		[0] = "Main Output",
	}

	tInputConnMapByID = {
		--index:  mod 1000 value of Input Connection id
		--[0] = {Name = "INPUT HDMI 1",BindingID = PROJECTOR_PROXY_BINDINGID,},
		[1] = {Name = "Computer",BindingID = PROJECTOR_PROXY_BINDINGID,AVFlag = true},
		[21] = {Name = "HDMI",BindingID = PROJECTOR_PROXY_BINDINGID,},
		[5] = {Name = "Video",BindingID = PROJECTOR_PROXY_BINDINGID,AVFlag = true},
		[999] = {Name = "S-Video",BindingID = PROJECTOR_PROXY_BINDINGID,},
		[999] = {Name = "Display Port",BindingID = PROJECTOR_PROXY_BINDINGID,},
		[85] = {Name = "HDBaseT",BindingID = PROJECTOR_PROXY_BINDINGID,},
		[999] = {Name = "BNC",BindingID = PROJECTOR_PROXY_BINDINGID,},
		[999] = {Name = "SDI",BindingID = PROJECTOR_PROXY_BINDINGID,},
		[2] = {Name = "Computer2",BindingID = PROJECTOR_PROXY_BINDINGID,AVFlag = true},
		[22] = {Name = "HDMI 2",BindingID = PROJECTOR_PROXY_BINDINGID,},
		[95] = {Name = "USB",BindingID = PROJECTOR_PROXY_BINDINGID,},
		[88] = {Name = "LAN",BindingID = PROJECTOR_PROXY_BINDINGID,},
		[96] = {Name = "USB 2",BindingID = PROJECTOR_PROXY_BINDINGID,},
		[89] = {Name = "USB Display",BindingID = PROJECTOR_PROXY_BINDINGID,},
		[26] = {Name = "White Board",BindingID = PROJECTOR_PROXY_BINDINGID,},
		[86] = {Name = "Screen Mirroring",BindingID = PROJECTOR_PROXY_BINDINGID,},
		[23] = {Name = "HDMI 3",BindingID = PROJECTOR_PROXY_BINDINGID,},
		[27] = {Name = "WirelessHD",BindingID = PROJECTOR_PROXY_BINDINGID,},
	}
	
	tInputConnMapByName = {
		--index:  Input Connection Name
		--ID: mod 1000 value of Input Connection id
		--["INPUT HDMI 1"] = {ID = 0,BindingID = PROJECTOR_PROXY_BINDINGID,},
		["Computer"] = {ID = 1,BindingID = PROJECTOR_PROXY_BINDINGID,},
		["HDMI"] = {ID = 21,BindingID = PROJECTOR_PROXY_BINDINGID,},
		["Video"] = {ID = 5,BindingID = PROJECTOR_PROXY_BINDINGID,},
		["S-Video"] = {ID = 999,BindingID = PROJECTOR_PROXY_BINDINGID,},
		["Display Port"] = {ID = 999,BindingID = PROJECTOR_PROXY_BINDINGID,},
		["HDBaseT"] = {ID = 85,BindingID = PROJECTOR_PROXY_BINDINGID,},
		["BNC"] = {ID = 999,BindingID = PROJECTOR_PROXY_BINDINGID,},
		["SDI"] = {ID = 999,BindingID = PROJECTOR_PROXY_BINDINGID,},
		["Computer2"] = {ID = 2,BindingID = PROJECTOR_PROXY_BINDINGID,},
		["HDMI 2"] = {ID = 22,BindingID = PROJECTOR_PROXY_BINDINGID,},
		["USB"] = {ID =95,BindingID = PROJECTOR_PROXY_BINDINGID,},
		["LAN"] = {ID = 88,BindingID = PROJECTOR_PROXY_BINDINGID,},
		["USB 2"] = {ID = 96,BindingID = PROJECTOR_PROXY_BINDINGID,},
		["USB Display"] = {ID = 89,BindingID = PROJECTOR_PROXY_BINDINGID,},
		["White Board"] = {ID = 26,BindingID = PROJECTOR_PROXY_BINDINGID,},
		["Screen Mirroring"] = {ID = 86,BindingID = PROJECTOR_PROXY_BINDINGID,},
		["HDMI 3"] = {ID = 23,BindingID = PROJECTOR_PROXY_BINDINGID,},
		["WirelessHD"] = {ID = 27,BindingID = PROJECTOR_PROXY_BINDINGID,},
	}

end	

function PROTOCOL_DECLARATIONS.PowerCommandsTableInit_Serial()
	LogTrace("PROTOCOL_DECLARATIONS.PowerCommandsTableInit_Serial()")
	
end

function PROTOCOL_DECLARATIONS.VolumeCommandsTableInit_Serial()
	LogTrace("PROTOCOL_DECLARATIONS.VolumeCommandsTableInit_Serial()")
	
	--[[
	TODO: Update tVolumeCurve table based upon devices volume range and optimal ramping steps
		tVolumeCurve is used to plot values that will be used
		to build volume commands during volume ramping.
		Specifically, they are used in GetNextVolumeCurveValue() which is 
		called from the ContinueVolumeRamping() function. 
		The values entered here should correlate to the volume parameters as defined in the device protocol.
		Below is an example, the table for your device may use different values.
		Also the number and size of steps may differ for your device. Adjust values to get 
		a smooth volume ramping from low to high volume.
		NOTE: values must be in decimal (i.e. not hexadecimal...)
	--]]
	tVolumeCurve = {
		"0", "12", "24", "36", "48", "60", "73", "85", "97", "109", 
		"121", "134", "146", "158", "170", "182", "195", "207", "219", "231", "243",                           
	}
	
end

function ReverseTable(a)
	local b = {}
	for k,v in pairs(a) do b[v] = k end
	return b
end

function SendConnectRequest()
    LogInfo ("SendConnectRequest()")
	local header = string.pack("A", "ESC/VP.net")

	local packet = string.pack("bb>Hbb", 0x10, 0x03, 0x0000, 0x00, 0x00)

	local command = header .. packet

	C4:SendToNetwork(NETWORK_BINDING_ID, NETWORK_PORT, command .. "\r")

end

--[[=============================================================================
    Projector Proxy Class Code

    Copyright 2015 Control4 Corporation. All Rights Reserved.
===============================================================================]]

-- This macro is utilized to identify the version string of the driver template version used.
if (TEMPLATE_VERSION ~= nil) then
	TEMPLATE_VERSION.properties = "2015.03.31"
end

DEFAULT_POWER_STATUS_QUERY_INTERVAL = 3         	-- Amount of time to wait BEFORE sending the next power status query, in seconds

ProjectorProxy = inheritsFrom(nil)

function ProjectorProxy:construct(projectorBindingID, bProcessesDeviceMessages, tVolumeRampingTracking, bUsePulseCommandsForVolumeRamping)
	-- member variables
	self._ProjectorBindingID = projectorBindingID
	self._PowerState = ""							--Valid Values: "ON", "OFF", "POWER_ON_SEQUENCE", "POWER_OFF_SEQUENCE"
	self._VolumeIsRamping = false
	self._VolumeRamping = tVolumeRampingTracking	--{["state"] = false, ["mode"] = "",} ||	"state" is boolean, "mode" values: "VOLUME_UP" & "VOLUME_DOWN"
	self._LastVolumeStatusValue = 0	
	self._MenuNavigationInProgress = false
     self._MenuNavigationMode = ""
	self._MenuNavigationProxyID = ""	
	self._CurrentlySelectedInput = ""
	self._ProcessesDeviceMessages = bProcessesDeviceMessages
	self._ControlMethod = ""						--Valid Values: "NETWORK", "SERIAL", "IR" 
	self._PowerStatusQueryDelayS = DEFAULT_COMMAND_RESPONSE_INTERVAL
	
	self._UsePulseCommandsForVolumeRamping = bUsePulseCommandsForVolumeRamping
end

function ProjectorProxy:Initialize(PowerStatusQueryDelayInterval, CallbackParam)
	if (PowerStatusQueryDelayInterval ~= nil) then
		self._PowerStatusQueryDelayS = PowerStatusQueryDelayInterval
	end
	self._PowerStatusQueryDelayTimer = c4_timer:new("PowerStatusQueryDelay", self._PowerStatusQueryDelayS, "SECONDS", ProjectorProxy.OnPowerStatusQueryDelayTimerExpired, false, self)
end

-- Note the '.' instead of the ':'
function ProjectorProxy.OnPowerStatusQueryDelayTimerExpired(Instance)
	LogTrace("Power Status Query Timer expired")
	Instance._PowerStatusQueryDelayTimer:KillTimer()
	
	if (GetDevicePowerStatus ~= nil and type(GetDevicePowerStatus) == "function") then
		GetDevicePowerStatus(true)  --"true" sends command as a priority command
	end
end

------------------------------------------------------------------------
-- Projector Proxy Commands (PRX_CMD)
------------------------------------------------------------------------
--=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
-- Power Functions
--=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
function ProjectorProxy:prx_ON(tParams)
	if (self._PowerState ~= nil) then
		if (self._PowerState == 'ON') or (self._PowerState == 'POWER_ON_SEQUENCE') then 
			LogInfo("Power State is '" .. self._PowerState .. "', exiting ON")
			return 
		end	
	end	
	if (self._ProcessesDeviceMessages == false) then
		self._PowerState = 'ON'
	else	
		self._PowerState = 'POWER_ON_SEQUENCE'
	end	
    if (gControlMethod == "IR") then			
		NOTIFY.ON()			
    end		
	ON()
end

function ProjectorProxy:prx_OFF(tParams)
    if (self._ProcessesDeviceMessages == false) then
	   self._PowerState = 'OFF'
    else	
	   self._PowerState = 'POWER_OFF_SEQUENCE'
    end	
    if (gControlMethod == "IR") then			
	   NOTIFY.OFF()			
    end 	
    OFF()
  
    self._CurrentlySelectedInput = -1
    C4:SendToProxy(self._ProjectorBindingID, 'INPUT_CHANGED', {INPUT = -1})
    C4:SendToProxy(self._ProjectorBindingID, 'INPUT_CHANGED', {INPUT = -1})
end

--=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
-- Input Selection Functions
--=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
function ProjectorProxy:prx_SET_INPUT(idBinding, tParams)
	local input = tonumber(tParams["INPUT"] % 1000)
	self._CurrentlySelectedInput = input
	if (gControlMethod == "IR") then			
		NOTIFY.INPUT_CHANGED(self._ProjectorBindingID, input)				
	end 		

	SET_INPUT(idBinding, input)
end

function ProjectorProxy:prx_TV_VIDEO(idBinding, tParams)
	SEND_COMMAND_FROM_COMMAND_TABLE(idBinding, "TV_VIDEO")
end

--=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
-- Volume Control Functions
--=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
function ProjectorProxy:prx_MUTE_OFF(tParams)
	if (gControlMethod == "IR") then			
		NOTIFY.MUTE_CHANGED(self._ProjectorBindingID, "False")
	end 		
	MUTE_OFF()
end

function ProjectorProxy:prx_MUTE_ON(tParams)
	if (gControlMethod == "IR") then			
		NOTIFY.MUTE_CHANGED(self._ProjectorBindingID, "True")
	end 		
	MUTE_ON()
end

function ProjectorProxy:prx_MUTE_TOGGLE(tParams)
	MUTE_TOGGLE()
end

function ProjectorProxy:prx_SET_VOLUME_LEVEL(tParams)
	local c4VolumeLevel = tonumber(tParams['LEVEL'])
	SET_VOLUME_LEVEL(c4VolumeLevel)
end

function ProjectorProxy:prx_PULSE_VOL_DOWN(tParams)
	PULSE_VOL_DOWN()
end

function ProjectorProxy:prx_PULSE_VOL_UP(tParams)
	PULSE_VOL_UP()
end

function ProjectorProxy:prx_START_VOL_DOWN(tParams)
	self:ChangeVolume("START_VOL_DOWN")
end

function ProjectorProxy:prx_START_VOL_UP(tParams)
	self:ChangeVolume("START_VOL_UP")
end

function ProjectorProxy:prx_STOP_VOL_DOWN(tParams)
	self:ChangeVolume("STOP_VOL_DOWN")
end

function ProjectorProxy:prx_STOP_VOL_UP(tParams)
	self:ChangeVolume("STOP_VOL_UP")
end

---------------------- Volume Helper Functions ----------------------
function ProjectorProxy:ChangeVolume(command_name)
	if (command_name == "STOP_VOL_UP") or (command_name == "STOP_VOL_DOWN") then
		self._VolumeIsRamping = false
		self._VolumeRamping.state = false
		self._VolumeRamping.mode = ""
	elseif (command_name == "START_VOL_UP") then 
		self._VolumeIsRamping = true
		self._VolumeRamping.state = true
		self._VolumeRamping.mode = "VOLUME_UP" 
		PULSE_VOL_UP()	
	elseif (command_name == "START_VOL_DOWN") then 	
		self._VolumeIsRamping = true
		self._VolumeRamping.state = true
		self._VolumeRamping.mode = "VOLUME_DOWN"	
		PULSE_VOL_DOWN()		
	else
		LogWarn(command_name .. " not handled in ChangeVolume()")
	end
end

function ProjectorProxy:ContinueVolumeRamping()
	local command
	if (gControlMethod == "IR" or self._UsePulseCommandsForVolumeRamping == true) then   
		if (self._VolumeRamping.mode == "VOLUME_UP") then
			PULSE_VOL_UP()	
		elseif (self._VolumeRamping.mode == "VOLUME_DOWN") then
			PULSE_VOL_DOWN()	
		else
			LogWarn("ContinueVolumeRamping() ramping mode is not valid.")
		end	
	else
		local volume = self._LastVolumeStatusValue
		local deviceVolumeLevel = self:GetNextVolumeCurveValue(volume)
		if (deviceVolumeLevel ~= nil) then
			self._LastVolumeStatusValue = deviceVolumeLevel
			SET_VOLUME_LEVEL_DEVICE(deviceVolumeLevel)                                 
		else
			LogWarn("ContinueVolumeRamping() next value is nil")
			return
		end
	end 
end

function ProjectorProxy:GetNextVolumeCurveValue(volume)
	local i, point
	volume=tonumber(volume)
	if (self._VolumeRamping.mode == "VOLUME_UP") then
		for i=1,table.maxn(tVolumeCurve) do
			point=tonumber(tVolumeCurve[i])
			if point > volume then		
				return tVolumeCurve[i]
			end
		end
	elseif (self._VolumeRamping.mode == "VOLUME_DOWN") then
		for i=table.maxn(tVolumeCurve),1,-1 do
			point=tonumber(tVolumeCurve[i])
			if point < volume then
				return tVolumeCurve[i]
			end
		end
	else
		LogWarn("Volume Ramping Mode not set for "  .. tOutputConnMap[output])
		return nil
	end 
end

function ConvertVolumeToC4(volume, minDeviceLevel, maxDeviceLevel)
	--to be used when converting a volume level from a device to a 
	--percentage value that can be used by C4 proxies
	--"volume" is the volume value from the device
	--"minDeviceLevel" & "maxDeviceLevel" are the minimum and maximum volume levels
	--as specified in the device protocol documentation
	return ProcessVolumeLevel(volume, minDeviceLevel, maxDeviceLevel, 0, 100)
end

function ConvertVolumeToDevice(volume, minDeviceLevel, maxDeviceLevel)
	--to be used when converting a volume level from a C4 proxy to a 
	--value that can be used by the device 
	--"volume" is the volume value from the C4 proxy
	--"minDeviceLevel" & "maxDeviceLevel" are the minimum and maximum volume levels
	--as specified in the device protocol documentation
	return ProcessVolumeLevel(volume, 0, 100, minDeviceLevel, maxDeviceLevel)
end

function ProcessVolumeLevel(volLevel, minVolLevel, maxVolLevel, minDeviceLevel, maxDeviceLevel)
	  local level = (volLevel-minVolLevel)/(maxVolLevel-minVolLevel)
	  --LogInfo("level = " .. level)
	  local vl=(level*(maxDeviceLevel-minDeviceLevel))+minDeviceLevel
	  --LogInfo("vl = " .. vl)
	  vl= tonumber_loc(("%.".."0".."f"):format(vl))
	  --LogInfo("vl new = " .. vl)
	  LogInfo("ProcessVolumeLevel(level in=" .. volLevel .. ", level out=" .. vl .. ")")
	  return vl
end

--=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
-- Tuner Proxy Functions
--=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
function ProjectorProxy:prx_PULSE_CH_UP(idBinding, tParams)
	SEND_COMMAND_FROM_COMMAND_TABLE(idBinding, "PULSE_CH_UP")
end

function ProjectorProxy:prx_START_CH_UP(idBinding, tParams)
	SEND_COMMAND_FROM_COMMAND_TABLE(idBinding, "PULSE_CH_UP")
end

function ProjectorProxy:prx_PULSE_CH_DOWN(idBinding, tParams)
	SEND_COMMAND_FROM_COMMAND_TABLE(idBinding, "PULSE_CH_DOWN")
end

function ProjectorProxy:prx_START_CH_DOWN(idBinding, tParams)
	SEND_COMMAND_FROM_COMMAND_TABLE(idBinding, "PULSE_CH_DOWN")
end

function ProjectorProxy:prx_PRESET_UP(idBinding, tParams)
	SEND_COMMAND_FROM_COMMAND_TABLE(idBinding, "PRESET_UP")
end

function ProjectorProxy:prx_PRESET_DOWN(idBinding, tParams)
	SEND_COMMAND_FROM_COMMAND_TABLE(idBinding, "PRESET_DOWN")
end

function ProjectorProxy:prx_SET_CHANNEL(idBinding, tParams)
    local channel= tParams['CHANNEL']
	SET_CHANNEL(idBinding, channel)
end


--=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
-- Menu Functions
--=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
function ProjectorProxy:prx_INFO(idBinding, tParams)
	SEND_COMMAND_FROM_COMMAND_TABLE(idBinding, "INFO")
end

function ProjectorProxy:prx_GUIDE(idBinding, tParams)
	SEND_COMMAND_FROM_COMMAND_TABLE(idBinding, "GUIDE")
end

function ProjectorProxy:prx_MENU(idBinding, tParams)
	SEND_COMMAND_FROM_COMMAND_TABLE(idBinding, "MENU")
end

function ProjectorProxy:prx_CANCEL(idBinding, tParams)
	SEND_COMMAND_FROM_COMMAND_TABLE(idBinding, "CANCEL")
end

function ProjectorProxy:prx_UP(idBinding, tParams)
	SEND_COMMAND_FROM_COMMAND_TABLE(idBinding, "UP")
end

function ProjectorProxy:prx_DOWN(idBinding, tParams)
	SEND_COMMAND_FROM_COMMAND_TABLE(idBinding, "DOWN")
end

function ProjectorProxy:prx_LEFT(idBinding, tParams)
	SEND_COMMAND_FROM_COMMAND_TABLE(idBinding, "LEFT")
end

function ProjectorProxy:prx_RIGHT(idBinding, tParams)
	SEND_COMMAND_FROM_COMMAND_TABLE(idBinding, "RIGHT")
end

function ProjectorProxy:prx_START_DOWN(idBinding, tParams)
	self:NavigateMenu(idBinding, "START_DOWN")
end

function ProjectorProxy:prx_START_UP(idBinding, tParams)
	self:NavigateMenu(idBinding, "START_UP")
end

function ProjectorProxy:prx_START_LEFT(idBinding, tParams)
	self:NavigateMenu(idBinding, "START_LEFT")
end

function ProjectorProxy:prx_START_RIGHT(idBinding, tParams)
	self:NavigateMenu(idBinding, "START_RIGHT")
end

function ProjectorProxy:prx_STOP_DOWN(idBinding, tParams)
	self:NavigateMenu(idBinding, "STOP_DOWN")
end

function ProjectorProxy:prx_STOP_UP(idBinding, tParams)
	self:NavigateMenu(idBinding, "STOP_UP")
end

function ProjectorProxy:prx_STOP_LEFT(idBinding, tParams)
	self:NavigateMenu(idBinding, "STOP_LEFT")
end

function ProjectorProxy:prx_STOP_RIGHT(idBinding, tParams)
	self:NavigateMenu(idBinding, "STOP_RIGHT")
end

function ProjectorProxy:prx_ENTER(idBinding, tParams)
	SEND_COMMAND_FROM_COMMAND_TABLE(idBinding, "ENTER")
end

function ProjectorProxy:prx_RECALL(idBinding, tParams)
	SEND_COMMAND_FROM_COMMAND_TABLE(idBinding, "RECALL")
end

function ProjectorProxy:prx_OPEN_CLOSE(idBinding, tParams)
	SEND_COMMAND_FROM_COMMAND_TABLE(idBinding, "OPEN_CLOSE")
end

function ProjectorProxy:prx_PROGRAM_A(idBinding, tParams)
	SEND_COMMAND_FROM_COMMAND_TABLE(idBinding, "PROGRAM_A")
end

function ProjectorProxy:prx_PROGRAM_B(idBinding, tParams)
	SEND_COMMAND_FROM_COMMAND_TABLE(idBinding, "PROGRAM_B")
end

function ProjectorProxy:prx_PROGRAM_C(idBinding, tParams)
	SEND_COMMAND_FROM_COMMAND_TABLE(idBinding, "PROGRAM_C")
end

function ProjectorProxy:prx_PROGRAM_D(idBinding, tParams)
	SEND_COMMAND_FROM_COMMAND_TABLE(idBinding, "PROGRAM_D")
end

---------------------- Menu Navigation Helper Functions ----------------------
function ProjectorProxy:NavigateMenu(idBinding, command_name)
	if (command_name == "STOP_UP") 
				or (command_name == "STOP_DOWN") 
				or (command_name == "STOP_LEFT") 
				or (command_name == "STOP_RIGHT") then
		self._MenuNavigationInProgress = false
		self._MenuNavigationMode = ""
		self._MenuNavigationProxyID = ""
		return
	elseif (command_name == "START_UP") then 
		self._MenuNavigationMode = "UP"
	elseif (command_name == "START_DOWN") then 	
		self._MenuNavigationMode = "DOWN"	
	elseif (command_name == "START_LEFT") then 
		self._MenuNavigationMode = "LEFT"	
     elseif (command_name == "START_RIGHT") then 
		self._MenuNavigationMode = "RIGHT"
	else
		LogWarn(command_name .. " not handled in NavigateMenu()")
	end
	
	SEND_COMMAND_FROM_COMMAND_TABLE(idBinding, self._MenuNavigationMode)
	self._MenuNavigationInProgress = true
	self._MenuNavigationProxyID = idBinding
	
end

--=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
-- Digit Functions
--=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
function ProjectorProxy:prx_NUMBER_0(idBinding, tParams)
	SEND_COMMAND_FROM_COMMAND_TABLE(idBinding, "NUMBER_0")
end

function ProjectorProxy:prx_NUMBER_1(idBinding, tParams)
	SEND_COMMAND_FROM_COMMAND_TABLE(idBinding, "NUMBER_1")
end

function ProjectorProxy:prx_NUMBER_2(idBinding, tParams)
	SEND_COMMAND_FROM_COMMAND_TABLE(idBinding, "NUMBER_2")
end

function ProjectorProxy:prx_NUMBER_3(idBinding, tParams)
	SEND_COMMAND_FROM_COMMAND_TABLE(idBinding, "NUMBER_3")
end

function ProjectorProxy:prx_NUMBER_4(idBinding, tParams)
	SEND_COMMAND_FROM_COMMAND_TABLE(idBinding, "NUMBER_4")
end

function ProjectorProxy:prx_NUMBER_5(idBinding, tParams)
	SEND_COMMAND_FROM_COMMAND_TABLE(idBinding, "NUMBER_5")
end

function ProjectorProxy:prx_NUMBER_6(idBinding, tParams)
	SEND_COMMAND_FROM_COMMAND_TABLE(idBinding, "NUMBER_6")
end

function ProjectorProxy:prx_NUMBER_7(idBinding, tParams)
	SEND_COMMAND_FROM_COMMAND_TABLE(idBinding, "NUMBER_7")
end

function ProjectorProxy:prx_NUMBER_8(idBinding, tParams)
	SEND_COMMAND_FROM_COMMAND_TABLE(idBinding, "NUMBER_8")
end

function ProjectorProxy:prx_NUMBER_9(idBinding, tParams)
	SEND_COMMAND_FROM_COMMAND_TABLE(idBinding, "NUMBER_9")
end

function ProjectorProxy:prx_STAR(idBinding, tParams)
	SEND_COMMAND_FROM_COMMAND_TABLE(idBinding, "STAR")
end

function ProjectorProxy:prx_POUND(idBinding, tParams)
	SEND_COMMAND_FROM_COMMAND_TABLE(idBinding, "POUND")
end

------------------------------------------------------------------------
-- Projector Proxy Notifies
------------------------------------------------------------------------

function ProjectorProxy:dev_InputChanged(bindingID, input)
	NOTIFY.INPUT_CHANGED(bindingID, input)
end

function ProjectorProxy:dev_PowerOn()
	self._PowerState = "ON"
	NOTIFY.ON()	
end

function ProjectorProxy:dev_PowerOff()
	self._PowerState = "OFF"
	NOTIFY.OFF()
end

function ProjectorProxy:dev_VolumeLevelChanged(output, c4Level, deviceLevel)
	NOTIFY.VOLUME_LEVEL_CHANGED(self._ProjectorBindingID, output, c4Level)	
	
	if (self._VolumeIsRamping) then
		--do nothing
		--during volume ramping, LastVolumeStatusValue is set in ContinueVolumeRamping()
	else
		self._LastVolumeStatusValue = deviceLevel
	end	
end

function ProjectorProxy:dev_MuteChanged(output, state)
	NOTIFY.MUTE_CHANGED(self._ProjectorBindingID, output, state)
end		

function ProjectorProxy:dev_BassLevelChanged(output, level)
	NOTIFY.BASS_LEVEL_CHANGED(self._ProjectorBindingID, output, level)
end	

function ProjectorProxy:dev_TrebleLevelChanged(output, level)
	NOTIFY.TREBLE_LEVEL_CHANGED(self._ProjectorBindingID, output, level)
end	

function ProjectorProxy:dev_BalanceLevelChanged(output, level)
	NOTIFY.BALANCE_LEVEL_CHANGED(self._ProjectorBindingID, output, level)
end	

function ProjectorProxy:dev_LoudnessChanged(output, state)
	NOTIFY.LOUDNESS_CHANGED(self._ProjectorBindingID, output, state)
end

function ProjectorProxy:dev_ChannelStatus(channel)
	NOTIFY.CHANNEL_CHANGED(self._ProjectorBindingID, channel)
end

function ProjectorProxy:prx_PASSTHROUGH(idBinding, tParams)
	ReceivedFromProxy(idBinding, tParams.PASSTHROUGH_COMMAND)
end
--[[=============================================================================
    ReceivedFromProxy Code for the Projector Proxy

    Copyright 2015 Control4 Corporation. All Rights Reserved.
===============================================================================]]

-- This macro is utilized to identify the version string of the driver template version used.
if (TEMPLATE_VERSION ~= nil) then
	TEMPLATE_VERSION.properties = "2015.03.31"
end

--=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
-- Power Functions
--=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
function PRX_CMD.ON(idBinding, tParams)
	gProjectorProxy:prx_ON(tParams)
end

function PRX_CMD.OFF(idBinding, tParams)
	gProjectorProxy:prx_OFF(tParams)
end

--=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
-- Input Selection Functions
--=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
function PRX_CMD.SET_INPUT(idBinding, tParams)
	gProjectorProxy:prx_SET_INPUT(idBinding, tParams)
end

function PRX_CMD.TV_VIDEO(idBinding, tParams)
	gProjectorProxy:prx_TV_VIDEO(idBinding, tParams)
end

--=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
-- Volume Control Functions
--=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
function PRX_CMD.MUTE_OFF(idBinding, tParams)
	gProjectorProxy:prx_MUTE_OFF(tParams)
end

function PRX_CMD.MUTE_ON(idBinding, tParams)
	gProjectorProxy:prx_MUTE_ON(tParams)	
end

function PRX_CMD.MUTE_TOGGLE(idBinding, tParams)
	gProjectorProxy:prx_MUTE_TOGGLE(tParams)
end

function PRX_CMD.SET_VOLUME_LEVEL(idBinding, tParams)
	gProjectorProxy:prx_SET_VOLUME_LEVEL(tParams)
end

function PRX_CMD.PULSE_VOL_DOWN(idBinding, tParams)
	gProjectorProxy:prx_PULSE_VOL_DOWN(tParams)
end

function PRX_CMD.PULSE_VOL_UP(idBinding, tParams)
	gProjectorProxy:prx_PULSE_VOL_UP(tParams)
end

function PRX_CMD.START_VOL_DOWN(idBinding, tParams)
	gProjectorProxy:prx_START_VOL_DOWN(tParams)
end

function PRX_CMD.START_VOL_UP(idBinding, tParams)
	gProjectorProxy:prx_START_VOL_UP(tParams)
end

function PRX_CMD.STOP_VOL_DOWN(idBinding, tParams)
	gProjectorProxy:prx_STOP_VOL_DOWN(tParams)
end

function PRX_CMD.STOP_VOL_UP(idBinding, tParams)
	gProjectorProxy:prx_STOP_VOL_UP(tParams)
end

--=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
-- Menu Functions
--=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
function PRX_CMD.INFO(idBinding, tParams)
	gProjectorProxy:prx_INFO(idBinding, tParams)
end

function PRX_CMD.GUIDE(idBinding, tParams)
	gProjectorProxy:prx_GUIDE(idBinding, tParams)
end

function PRX_CMD.MENU(idBinding, tParams)
	gProjectorProxy:prx_MENU(idBinding, tParams)
end

function PRX_CMD.CANCEL(idBinding, tParams)
	gProjectorProxy:prx_CANCEL(idBinding, tParams)
end

function PRX_CMD.UP(idBinding, tParams)
	gProjectorProxy:prx_UP(idBinding, tParams)
end

function PRX_CMD.DOWN(idBinding, tParams)
	gProjectorProxy:prx_DOWN(idBinding, tParams)
end

function PRX_CMD.LEFT(idBinding, tParams)
	gProjectorProxy:prx_LEFT(idBinding, tParams)
end

function PRX_CMD.RIGHT(idBinding, tParams)
	gProjectorProxy:prx_RIGHT(idBinding, tParams)
end

function PRX_CMD.START_DOWN(idBinding, tParams)
	gProjectorProxy:prx_START_DOWN(idBinding, tParams)
end

function PRX_CMD.START_UP(idBinding, tParams)
	gProjectorProxy:prx_START_UP(idBinding, tParams)
end

function PRX_CMD.START_LEFT(idBinding, tParams)
	gProjectorProxy:prx_START_LEFT(idBinding, tParams)
end

function PRX_CMD.START_RIGHT(idBinding, tParams)
	gProjectorProxy:prx_START_RIGHT(idBinding, tParams)
end

function PRX_CMD.STOP_DOWN(idBinding, tParams)
	gProjectorProxy:prx_STOP_DOWN(idBinding, tParams)
end

function PRX_CMD.STOP_UP(idBinding, tParams)
	gProjectorProxy:prx_STOP_UP(idBinding, tParams)
end

function PRX_CMD.STOP_LEFT(idBinding, tParams)
	gProjectorProxy:prx_STOP_LEFT(idBinding, tParams)
end

function PRX_CMD.STOP_RIGHT(idBinding, tParams)
	gProjectorProxy:prx_STOP_RIGHT(idBinding, tParams)
end

function PRX_CMD.ENTER(idBinding, tParams)
	gProjectorProxy:prx_ENTER(idBinding, tParams)
end

function PRX_CMD.RECALL(idBinding, tParams)
	gProjectorProxy:prx_RECALL(idBinding, tParams)
end

function PRX_CMD.OPEN_CLOSE(idBinding, tParams)
	gProjectorProxy:prx_OPEN_CLOSE(idBinding, tParams)
end

function PRX_CMD.PROGRAM_A(idBinding, tParams)
	gProjectorProxy:prx_PROGRAM_A(idBinding, tParams)
end

function PRX_CMD.PROGRAM_B(idBinding, tParams)
	gProjectorProxy:prx_PROGRAM_B(idBinding, tParams)
end

function PRX_CMD.PROGRAM_C(idBinding, tParams)
	gProjectorProxy:prx_PROGRAM_C(idBinding, tParams)
end

function PRX_CMD.PROGRAM_D(idBinding, tParams)
	gProjectorProxy:prx_PROGRAM_D(idBinding, tParams)
end

--=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
-- Digit Functions
--=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
function PRX_CMD.NUMBER_0(idBinding, tParams)
	gProjectorProxy:prx_NUMBER_0(idBinding, tParams)
end

function PRX_CMD.NUMBER_1(idBinding, tParams)
	gProjectorProxy:prx_NUMBER_1(idBinding, tParams)
end

function PRX_CMD.NUMBER_2(idBinding, tParams)
	gProjectorProxy:prx_NUMBER_2(idBinding, tParams)
end

function PRX_CMD.NUMBER_3(idBinding, tParams)
	gProjectorProxy:prx_NUMBER_3(idBinding, tParams)
end

function PRX_CMD.NUMBER_4(idBinding, tParams)
	gProjectorProxy:prx_NUMBER_4(idBinding, tParams)
end

function PRX_CMD.NUMBER_5(idBinding, tParams)
	gProjectorProxy:prx_NUMBER_5(idBinding, tParams)
end

function PRX_CMD.NUMBER_6(idBinding, tParams)
	gProjectorProxy:prx_NUMBER_6(idBinding, tParams)
end

function PRX_CMD.NUMBER_7(idBinding, tParams)
	gProjectorProxy:prx_NUMBER_7(idBinding, tParams)
end

function PRX_CMD.NUMBER_8(idBinding, tParams)
	gProjectorProxy:prx_NUMBER_8(idBinding, tParams)
end

function PRX_CMD.NUMBER_9(idBinding, tParams)
	gProjectorProxy:prx_NUMBER_9(idBinding, tParams)
end

function PRX_CMD.STAR(idBinding, tParams)
	gProjectorProxy:prx_STAR(idBinding, tParams)
end

function PRX_CMD.POUND(idBinding, tParams)
	gProjectorProxy:prx_POUND(idBinding, tParams)
end

function PRX_CMD.PASSTHROUGH(idBinding, tParams)
	gProjectorProxy:prx_PASSTHROUGH(idBinding, tParams)
end

--[[=============================================================================
    Projector Proxy Notification Code

    Copyright 2015 Control4 Corporation. All Rights Reserved.
===============================================================================]]

-- This macro is utilized to identify the version string of the driver template version used.
if (TEMPLATE_VERSION ~= nil) then
	TEMPLATE_VERSION.properties = "2015.03.31"
end

--[[
	Notify: INPUT_CHANGED
	Sent to the proxy to indicate the currently selected Input has changed
	Parameters
		bindingID - proxy id of proxy bound to input connection
		input - mod 1000 value of Input Connection ID
--]]
function NOTIFY.INPUT_CHANGED(bindingID, input)
    local tParams = {INPUT = input}
    SendNotify("INPUT_CHANGED", tParams, bindingID)
end
------------------------------ Power Notification Functions ------------------------------
function NOTIFY.ON()
	SendSimpleNotify("ON")
end

function NOTIFY.OFF()
	SendSimpleNotify("OFF")
end
------------------------------ Volume Notification Code ------------------------------
--[[
	Notify: VOLUME_LEVEL_CHANGED
	Sent to the Projector proxy to indicate that the Volume level has changed on the specified Output
	Parameters
		bindingID - proxy id of Projector proxy
		output - mod 1000 value of Output Connection ID
		level - C4 level uses a percentage scale: 0 - 100
--]]
function NOTIFY.VOLUME_LEVEL_CHANGED(bindingID, output, level)
    local tParams = {LEVEL = level, OUTPUT = output + 4000}
    SendNotify("VOLUME_LEVEL_CHANGED", tParams, bindingID)	
end

--[[
	Notify: BASS_LEVEL_CHANGED
	Sent to the Projector proxy to indicate that the Bass level has changed on the specified Output
	Parameters
		bindingID - proxy id of Projector proxy
		output - mod 1000 value of Output Connection ID
		level - C4 volume level uses a percentage scale: 0 - 100
--]]
function NOTIFY.BASS_LEVEL_CHANGED(bindingID, output, level)
    local tParams = {LEVEL = level, OUTPUT = output + 4000}
    SendNotify("BASS_LEVEL_CHANGED", tParams, bindingID)		
end

--[[
	Notify: TREBLE_LEVEL_CHANGED
	Sent to the Projector proxy to indicate that the Treble level has changed on the specified Output
	Parameters
		bindingID - proxy id of Projector proxy
		output - mod 1000 value of Output Connection ID
		level - C4 volume level uses a percentage scale: 0 - 100
--]]
function NOTIFY.TREBLE_LEVEL_CHANGED(bindingID, output, level)
    local tParams = {LEVEL = level, OUTPUT = output + 4000}
    SendNotify("TREBLE_LEVEL_CHANGED", tParams, bindingID)	
end

--[[
	Notify: BALANCE_LEVEL_CHANGED
	Sent to the Projector proxy to indicate that the Balance level has changed on the specified Output
	Parameters
		bindingID - proxy id of Projector proxy
		output - mod 1000 value of Output Connection ID
		level - C4 volume level uses a percentage scale: 0 - 100
--]]
function NOTIFY.BALANCE_LEVEL_CHANGED(bindingID, output, level)
    local tParams = {LEVEL = level, OUTPUT = output + 4000}
    SendNotify("BALANCE_LEVEL_CHANGED", tParams, bindingID)		
end

--[[
	Notify: LOUNDENSS_CHANGED
	Sent to the Projector proxy to indicate that the Loudness state has changed on the specified Output
	Parameters
		bindingID - proxy id of Projector proxy
		output - mod 1000 value of Output Connection ID
		state - represented as "True" or "False" (literal string, not boolean)
--]]
function NOTIFY.LOUDNESS_CHANGED(bindingID, output, state)
    local tParams = {LOUDNESS = state, OUTPUT = output + 4000}
    SendNotify("LOUDNESS_CHANGED", tParams, bindingID)		
end

--[[
	Notify: MUTE_LEVEL_CHANGED
	Sent to the Projector proxy to indicate that the Mute state has changed on the specified Output
	Parameters
		bindingID - proxy id of Projector proxy
		output - mod 1000 value of Output Connection ID
		state - represented as "True" or "False" (literal string, not boolean)
--]]
function NOTIFY.MUTE_CHANGED(bindingID, output, state)
    local tParams = {MUTE = state, OUTPUT = output + 4000}
    SendNotify("MUTE_CHANGED", tParams, bindingID)		
end


------------------------------------------------------------

-- This macro is utilized to identify the version string of the driver template version used.
if (TEMPLATE_VERSION ~= nil) then
	TEMPLATE_VERSION.driver = "2015.03.31"
end

--[[=============================================================================
    Constants
===============================================================================]]
DRIVER_NAME = Properties["Epson Projector"]
PROJECTOR_PROXY_BINDINGID = 5001


--[[=============================================================================
    Initialization Code
===============================================================================]]
function ON_DRIVER_EARLY_INIT.main()

end

function ON_DRIVER_INIT.main()
	SetLogName(DRIVER_NAME)

	-- TODO: If cloud based driver then uncomment the following line
	--ConnectURL()
end

function ON_DRIVER_LATEINIT.main()
    gControlMethod = C4:PersistGetValue("gControlMethod")
    gProjector_Operating_Status = C4:PersistGetValue("gProjector_Operating_Status")
    SetControlMethodFlags()
    PowerStatusPollingTimer = c4_timer:new("Polling Power Status Timer", 30, "SECONDS", OnPowerStatusPollingTimer, true)
    PowerStatusPollingTimer:StartTimer()
end
function OnPowerStatusPollingTimer()
    LogTrace("OnPowerStatusPollingTimer()")
    LogTrace("Checking Power Status")

    --TODO: set keep alive command for the network connected system if required.
    local command = "PWR?"
    LogTrace("command = " .. command)
    PackAndQueueCommand("Power Sstatus Query", command)
end
function ON_DRIVER_EARLY_INIT.projector_driver()

end

function ON_DRIVER_INIT.projector_driver()
    -- TODO: Modify tVolumeRamping to have on entry per Output connection
    --index is mod 1000 value of output connection
    local tVolumeRamping = {["state"] = false, ["mode"] = ""}
    gMuteState = "" --ON, OFF

    -- Create an instance of the PROJECTOR Proxy class
    -- TODO: Change bProcessesDeviceMessages to false if Device Messages will not be processes
    local bProcessesDeviceMessages = true
    local bUsePulseCommandsForVolumeRamping = true
    gProjectorProxy = ProjectorProxy:new(PROJECTOR_PROXY_BINDINGID,  bProcessesDeviceMessages, tVolumeRamping, bUsePulseCommandsForVolumeRamping)
    gProjectorProxy:Initialize()
    
    gPowerOnTimer = c4_timer:new("PowerOnTimer", tonumber(Properties["Power On Delay Seconds"]), "SECONDS", OnPowerOnTimerExpired)
    gPowerOffTimer = c4_timer:new("PowerOffTimer", tonumber(Properties["Power Off Delay Seconds"]), "SECONDS", OnPowerOffTimerExpired)
    gPowerOnDelayTimer = c4_timer:new("PowerOnDelayTimer", tonumber(Properties["Power Off Delay Seconds"]), "SECONDS", OnPowerOnDelayTimerExpired)
    gPowerOffDelayTimer = c4_timer:new("PowerOffDelayTimer", tonumber(Properties["Power On Delay Seconds"]), "SECONDS", OnPowerOffDelayTimerExpired)
    gPowerQueryDelayTimer = c4_timer:new("PowerQueryDelayTimer", tonumber(Properties["Power On Delay Seconds"]), "SECONDS", OnPowerQueryDelayTimerExpired)
    gDoEventsTimer = c4_timer:new("DoEventsTimer", 500, "MILLISECONDS", OnDoEventsTimerExpired)
    gQueryVolumeLampHoursDelayTimer = c4_timer:new("QueryVolumeLampHoursDelayTimer", 4, "SECONDS", OnQueryVolumeLampHoursDelayTimerExpired)
    gSetInputDelayTimer = c4_timer:new("SetInputDelayTimer", 2, "SECONDS", OnSetInputDelayTimerExpired, command)
end

function ON_DRIVER_LATEINIT.projector_driver()
end

--[[=============================================================================
    Driver Code
===============================================================================]]
function PackAndQueueCommand(...)
    local command_name = select(1, ...) or ""
    local command = select(2, ...) or ""
    local command_delay = select(3, ...) or tonumber(Properties["Command Delay Milliseconds"])
    local delay_units = select(4, ...) or "MILLISECONDS"
    LogTrace("PackAndQueueCommand(), command_name = " .. command_name .. ", command delay set to " .. command_delay .. " " .. delay_units)
    if (command == "") then
	   LogWarn("PackAndQueueCommand(), command_name = " .. command_name .. ", command string is empty - exiting PackAndQueueCommand()")
	   return
    end
	
	-- TODO: pack command with any any required starting or ending characters
    local cmd, stx, etx
    if (gControlMethod == "Network") then
		-- TODO: define any required starting or ending characters. 
		stx = ""
		etx = "\r"
		cmd = stx .. command .. etx
    elseif (gControlMethod == "Serial") then
		-- TODO: define any required starting or ending characters. 
		stx = ""
		etx = "\r"
		cmd = stx .. command .. etx
    elseif (gControlMethod == "IR") then
		cmd = command
    else
		LogWarn("PackAndQueueCommand(): gControlMethod is not valid, ".. gControlMethod)
		return
    end
    gCon:QueueCommand(cmd, command_delay, delay_units, command_name)
	
end

function PackAndQueuePriority1Command(...)
    local command_name = select(1, ...) or ""
    local command = select(2, ...) or ""
    local command_delay = select(3, ...) or tonumber(Properties["Command Delay Milliseconds"])
    local delay_units = select(4, ...) or "MILLISECONDS"
    LogTrace("PackAndQueueCommand(), command_name = " .. command_name .. ", command delay set to " .. command_delay .. " " .. delay_units)
    if (command == "") then
	   LogWarn("PackAndQueueCommand(), command_name = " .. command_name .. ", command string is empty - exiting PackAndQueueCommand()")
	   return
    end
	
	-- TODO: pack command with any any required starting or ending characters
    local cmd, stx, etx
    if (gControlMethod == "Network") then
		-- TODO: define any required starting or ending characters. 
		stx = ""
		etx = "\r"
		cmd = stx .. command .. etx
    elseif (gControlMethod == "Serial") then
		-- TODO: define any required starting or ending characters. 
		stx = ""
		etx = "\r"
		cmd = stx .. command .. etx
    elseif (gControlMethod == "IR") then
		cmd = command
    else
		LogWarn("PackAndQueueCommand(): gControlMethod is not valid, ".. gControlMethod)
		return
    end
    gCon:QueuePriority1Command(cmd, command_delay, delay_units, command_name)	
	
end
